/*
 *  The Syncro Soft SRL License
 *
 *  Copyright (c) 1998-2012 Syncro Soft SRL, Romania.  All rights
 *  reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistribution of source or in binary form is allowed only with
 *  the prior written permission of Syncro Soft SRL.
 *
 *  2. Redistributions of source code must retain the above copyright
 *  notice, this list of conditions and the following disclaimer.
 *
 *  3. Redistributions in binary form must reproduce the above copyright
 *  notice, this list of conditions and the following disclaimer in
 *  the documentation and/or other materials provided with the
 *  distribution.
 *
 *  4. The end-user documentation included with the redistribution,
 *  if any, must include the following acknowledgment:
 *  "This product includes software developed by the
 *  Syncro Soft SRL (http://www.sync.ro/)."
 *  Alternately, this acknowledgment may appear in the software itself,
 *  if and wherever such third-party acknowledgments normally appear.
 *
 *  5. The names "Oxygen" and "Syncro Soft SRL" must
 *  not be used to endorse or promote products derived from this
 *  software without prior written permission. For written
 *  permission, please contact support@oxygenxml.com.
 *
 *  6. Products derived from this software may not be called "Oxygen",
 *  nor may "Oxygen" appear in their name, without prior written
 *  permission of the Syncro Soft SRL.
 *
 *  THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 *  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *  OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *  DISCLAIMED.  IN NO EVENT SHALL THE SYNCRO SOFT SRL OR
 *  ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 *  USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 *  ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 *  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 *  OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */
package ro.sync.ecss.extensions.docbook;

import java.net.URL;
import java.util.HashMap;
import java.util.Map;

import org.apache.log4j.Logger;

import ro.sync.annotations.api.API;
import ro.sync.annotations.api.APIType;
import ro.sync.annotations.api.SourceType;
import ro.sync.exml.workspace.api.node.customizer.BasicRenderingInformation;
import ro.sync.exml.workspace.api.node.customizer.NodeRendererCustomizerContext;
import ro.sync.exml.workspace.api.node.customizer.XMLNodeRendererCustomizer;

/**
 * Class used to customize the way an Docbook node is rendered in the UI. 
 * A node represents an entry from Author outline, Author bread crumb,
 * Text page outline, content completion proposals window or Elements view.
 * 
 * @author alina_iordache
 */
@API(type = APIType.INTERNAL, src = SourceType.PUBLIC)
public class DocbookNodeRendererCustomizer extends XMLNodeRendererCustomizer {
  /**
   * Logger for logging.
   */
  private static final Logger logger = Logger.getLogger(DocbookNodeRendererCustomizer.class
      .getName());
  /**
   * Mapping between name and icon path
   */
  private static final Map<String, String> nameToIconPath = new HashMap<String, String>();

  // Initialize mapping
  static {
    // ROOT
    String docbookRoot = getImageURL("/images/node-customizer/ElementDocbook16.png");
    nameToIconPath.put("article", docbookRoot);
    nameToIconPath.put("book", docbookRoot);
    
    // link
    nameToIconPath.put("link", getImageURL("/images/node-customizer/ElementLink16.png"));

    // figure
    nameToIconPath.put("figure", getImageURL("/images/node-customizer/ElementImage16.png"));

    // list item
    nameToIconPath.put("listitem", getImageURL("/images/node-customizer/ElementLi16.png"));

    // ordered list
    nameToIconPath.put("orderedlist", getImageURL("/images/node-customizer/ElementOl16.png"));

    // ordered list
    nameToIconPath.put("itemizedlist", getImageURL("/images/node-customizer/ElementUl16.png"));

    // paragraph
    nameToIconPath.put("para", getImageURL("/images/node-customizer/ElementPara16.png"));

    // table
    nameToIconPath.put("table", getImageURL("/images/node-customizer/ElementTable16.png"));

    // table body
    nameToIconPath.put("tbody", getImageURL("/images/node-customizer/ElementTBody16.png"));

    String tableCellIcon = getImageURL("/images/node-customizer/ElementTd16.png");

    // table cell (CALS model)
    nameToIconPath.put("entry", tableCellIcon);
    
    // table cell (HTML model)
    nameToIconPath.put("td", tableCellIcon);
    
    // table foot
    nameToIconPath.put("tfoot", getImageURL("/images/node-customizer/ElementTFoot16.png"));

    // table head
    nameToIconPath.put("thead", getImageURL("/images/node-customizer/ElementTHead16.png"));

    String titleIcon = getImageURL("/images/node-customizer/ElementTitle16.png");

    // title
    nameToIconPath.put("title", titleIcon);
    
    // caption
    nameToIconPath.put("caption", titleIcon);

    String tableRowIcon = getImageURL("/images/node-customizer/ElementTr16.png");

    // table row (CALS model)
    nameToIconPath.put("row", tableRowIcon);
    // table row (HTML model)
    nameToIconPath.put("tr", tableRowIcon);

    // sect1
    nameToIconPath.put("sect1", getImageURL("/images/node-customizer/ElementH116.png"));

    // sect2
    nameToIconPath.put("sect2", getImageURL("/images/node-customizer/ElementH216.png"));

    // sect3
    nameToIconPath.put("sect3", getImageURL("/images/node-customizer/ElementH316.png"));

    // sect4
    nameToIconPath.put("sect4", getImageURL("/images/node-customizer/ElementH416.png"));

    // sect5
    nameToIconPath.put("sect5", getImageURL("/images/node-customizer/ElementH516.png"));

    // indexterm
    nameToIconPath.put("indexterm", getImageURL("/images/node-customizer/ElementIndexterm16.png"));

    // Some images that depend both on element name and some attribute value
    // The key is like: elementName @ attributeValue
    
    // bold
    nameToIconPath.put("emphasis@bold", getImageURL("/images/node-customizer/ElementBold16.png"));

    // italic
    nameToIconPath.put("emphasis@italic", getImageURL("/images/node-customizer/ElementItalic16.png"));

    // underline
    nameToIconPath.put("emphasis@underline", getImageURL("/images/node-customizer/ElementUnderline16.png"));
    
    // Audio and video
    nameToIconPath.put("audiodata", getImageURL("/images/node-customizer/ElementAudio16.png"));
    nameToIconPath.put("audioobject", getImageURL("/images/node-customizer/ElementAudio16.png"));
    
    nameToIconPath.put("videodata", getImageURL("/images/node-customizer/ElementVideo16.png"));
    nameToIconPath.put("videoobject", getImageURL("/images/node-customizer/ElementVideo16.png"));
    
    nameToIconPath.put("mediaobject", getImageURL("/images/node-customizer/ElementMedia16.png"));
  }

  /**
   * Searches the class-path for the image with the given path 
   * and if found it returns the string representation of its URL, otherwise it returns <code>null</code>.  
   * 
   * @param path The image path to search for.
   * 
   * @return The string representation of the image URL or <code>null</code> if the image is not found.
   */
  private static String getImageURL(String path) {
    URL imageURL = DocbookNodeRendererCustomizer.class.getResource(path);
    if (imageURL != null) {
      return imageURL.toExternalForm();
    } else {
      logger.error(DocbookNodeRendererCustomizer.class.getName() + " - Image not found: " + path);
      return null;
    }
  }

  @Override
  public BasicRenderingInformation getRenderingInformation(NodeRendererCustomizerContext context) {
    BasicRenderingInformation renderingInfo = new BasicRenderingInformation();
    String nodeName = context.getNodeName();
    if (nodeName != null) {
      // Get icon path from map
      String imageKey = nodeName;
      // Some images that depend both on element name and some attribute value
      // The key is like: elementName @ attributeValue
      if ("emphasis".equals(nodeName)) {
        String roleValue = context.getAttributeValue("role");
        if ("bold".equals(roleValue)) {
          imageKey += "@bold";
        } else if ("underline".equals(roleValue)) {
          imageKey += "@underline";
        } else {
          imageKey += "@italic";
        }
      }
      String iconPath = nameToIconPath.get(imageKey);
      renderingInfo.setIconPath(iconPath);
    }
    return renderingInfo;
  }

  @Override
  public String getDescription() {
    return "Docbook Node Renderer Customizer";
  }
}