/*
 *  The Syncro Soft SRL License
 *
 *  Copyright (c) 1998-2012 Syncro Soft SRL, Romania.  All rights
 *  reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistribution of source or in binary form is allowed only with
 *  the prior written permission of Syncro Soft SRL.
 *
 *  2. Redistributions of source code must retain the above copyright
 *  notice, this list of conditions and the following disclaimer.
 *
 *  3. Redistributions in binary form must reproduce the above copyright
 *  notice, this list of conditions and the following disclaimer in
 *  the documentation and/or other materials provided with the
 *  distribution.
 *
 *  4. The end-user documentation included with the redistribution,
 *  if any, must include the following acknowledgment:
 *  "This product includes software developed by the
 *  Syncro Soft SRL (http://www.sync.ro/)."
 *  Alternately, this acknowledgment may appear in the software itself,
 *  if and wherever such third-party acknowledgments normally appear.
 *
 *  5. The names "Oxygen" and "Syncro Soft SRL" must
 *  not be used to endorse or promote products derived from this
 *  software without prior written permission. For written
 *  permission, please contact support@oxygenxml.com.
 *
 *  6. Products derived from this software may not be called "Oxygen",
 *  nor may "Oxygen" appear in their name, without prior written
 *  permission of the Syncro Soft SRL.
 *
 *  THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 *  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *  OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *  DISCLAIMED.  IN NO EVENT SHALL THE SYNCRO SOFT SRL OR
 *  ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 *  USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 *  ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 *  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 *  OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */
package ro.sync.ecss.extensions.schematron;

import java.net.URL;
import java.util.HashMap;
import java.util.Map;

import org.apache.log4j.Logger;

import ro.sync.annotations.api.API;
import ro.sync.annotations.api.APIType;
import ro.sync.annotations.api.SourceType;
import ro.sync.exml.workspace.api.node.customizer.BasicRenderingInformation;
import ro.sync.exml.workspace.api.node.customizer.NodeRendererCustomizerContext;
import ro.sync.exml.workspace.api.node.customizer.XMLNodeRendererCustomizer;

/**
 * Class used to customize the way a Schematron node is rendered in the UI. 
 * A node represents an entry from Author outline, Author bread crumb,
 * Text page outline, content completion proposals window or Elements view.
 * 
 * @author alina_iordache
 */
@API(type = APIType.INTERNAL, src = SourceType.PUBLIC)
public class SchematronNodeRendererCustomizer extends XMLNodeRendererCustomizer {
  /**
   * Logger for logging.
   */
  private static final Logger logger = Logger.getLogger(SchematronNodeRendererCustomizer.class.getName());
  /**
   * Mapping between name and icon path
   */
  private static final Map<String, String> nameToIconPath = new HashMap<String, String>();

  // Initialize mapping
  static {
    // assert
    nameToIconPath.put("assert", getImageURL("/images/node-customizer/SchAssert16.png"));

    // report
    nameToIconPath.put("report", getImageURL("/images/node-customizer/SchReport16.png"));

    // rule
    nameToIconPath.put("rule", getImageURL("/images/node-customizer/SchRule16.png"));

    // pattern
    nameToIconPath.put("pattern", getImageURL("/images/node-customizer/SchPattern16.png"));

    // phase
    nameToIconPath.put("phase", getImageURL("/images/node-customizer/SchPhase16.png"));

    // active
    nameToIconPath.put("active", getImageURL("/images/node-customizer/SchActive16.png"));

    // diagnostics
    nameToIconPath.put("diagnostics", getImageURL("/images/node-customizer/SchDiagnostics16.png"));

    // diagnostic
    nameToIconPath.put("diagnostic", getImageURL("/images/node-customizer/SchDiagnostic16.png"));

    // schema
    nameToIconPath.put("schema", getImageURL("/images/node-customizer/SchSchema16.png"));

    // namespace
    nameToIconPath.put("ns", getImageURL("/images/node-customizer/SchNameSpace16.png"));
    
    // let
    nameToIconPath.put("let", getImageURL("/images/node-customizer/SchLet16.png"));
    
    // include
    nameToIconPath.put("include", getImageURL("/images/node-customizer/Include16.png"));
    
    // title
    nameToIconPath.put("title", getImageURL("/images/node-customizer/ElementTitle16.png"));
    
    // p
    nameToIconPath.put("p", getImageURL("/images/node-customizer/ElementPara16.png"));
    
    // span
    nameToIconPath.put("span", getImageURL("/images/node-customizer/ElementStylesheet16.png"));
  }

  /**
   * Searches the class-path for the image with the given path 
   * and if found it returns the string representation of its URL, otherwise it returns <code>null</code>.  
   * 
   * @param path The image path to search for.
   * 
   * @return The string representation of the image URL or <code>null</code> if the image is not found.
   */
  private static String getImageURL(String path) {
    URL imageURL = SchematronNodeRendererCustomizer.class.getResource(path);
    if (imageURL != null) {
      return imageURL.toExternalForm();
    } else {
      logger.error(SchematronNodeRendererCustomizer.class.getName() + " - Image not found: " + path);
      return null;
    }
  }
  
  /**
   * @see ro.sync.exml.workspace.api.node.customizer.XMLNodeRendererCustomizer#getRenderingInformation(ro.sync.exml.workspace.api.node.customizer.NodeRendererCustomizerContext)
   */
  @Override
  public BasicRenderingInformation getRenderingInformation(NodeRendererCustomizerContext context) {
    BasicRenderingInformation renderingInfo = new BasicRenderingInformation();
    String nodeName = context.getNodeName();
    if (nodeName != null) {
      String nodeNS = context.getNodeNamespace();
      String iconPath = null;
      if ("http://www.ascc.net/xml/schematron".equals(nodeNS) ||
          "http://purl.oclc.org/dsdl/schematron".equals(nodeNS)) {
        String imageKey = nodeName;
        int index = nodeName.indexOf(":");
        if (index == -1) {
          imageKey = nodeName;
        } else {
          imageKey = nodeName.substring(index + 1);
        }
        // Get icon path from map.
        iconPath = nameToIconPath.get(imageKey);
        renderingInfo.setIconPath(iconPath);
      } 
    }
    return renderingInfo;
  }

  /**
   * @see ro.sync.ecss.extensions.api.Extension#getDescription()
   */
  @Override
  public String getDescription() {
    return "Schematron Node Renderer Customizer";
  }
}