/*
 *  The Syncro Soft SRL License
 *
 *  Copyright (c) 1998-2012 Syncro Soft SRL, Romania.  All rights
 *  reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistribution of source or in binary form is allowed only with
 *  the prior written permission of Syncro Soft SRL.
 *
 *  2. Redistributions of source code must retain the above copyright
 *  notice, this list of conditions and the following disclaimer.
 *
 *  3. Redistributions in binary form must reproduce the above copyright
 *  notice, this list of conditions and the following disclaimer in
 *  the documentation and/or other materials provided with the
 *  distribution.
 *
 *  4. The end-user documentation included with the redistribution,
 *  if any, must include the following acknowledgment:
 *  "This product includes software developed by the
 *  Syncro Soft SRL (http://www.sync.ro/)."
 *  Alternately, this acknowledgment may appear in the software itself,
 *  if and wherever such third-party acknowledgments normally appear.
 *
 *  5. The names "Oxygen" and "Syncro Soft SRL" must
 *  not be used to endorse or promote products derived from this
 *  software without prior written permission. For written
 *  permission, please contact support@oxygenxml.com.
 *
 *  6. Products derived from this software may not be called "Oxygen",
 *  nor may "Oxygen" appear in their name, without prior written
 *  permission of the Syncro Soft SRL.
 *
 *  THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 *  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *  OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *  DISCLAIMED.  IN NO EVENT SHALL THE SYNCRO SOFT SRL OR
 *  ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 *  USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 *  ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 *  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 *  OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */
package ro.sync.ecss.extensions.wsdl;

import java.net.URL;
import java.util.HashMap;
import java.util.Map;

import org.apache.log4j.Logger;

import ro.sync.annotations.api.API;
import ro.sync.annotations.api.APIType;
import ro.sync.annotations.api.SourceType;
import ro.sync.ecss.extensions.xsd.XSDNodeRendererCustomizer;
import ro.sync.exml.workspace.api.node.customizer.BasicRenderingInformation;
import ro.sync.exml.workspace.api.node.customizer.NodeRendererCustomizerContext;
import ro.sync.exml.workspace.api.node.customizer.XMLNodeRendererCustomizer;

/**
 * Class used to customize the way an WSDL node is rendered in the UI. 
 * A node represents an entry from Author outline, Author bread crumb,
 * Text page outline, content completion proposals window or Elements view.
 * 
 * @author alina_iordache
 * @author mircea
 */
@API(type = APIType.INTERNAL, src = SourceType.PUBLIC)
public class WSDLNodeRendererCustomizer extends XMLNodeRendererCustomizer {
  /**
   * Logger for logging.
   */
  private static final Logger logger = Logger.getLogger(WSDLNodeRendererCustomizer.class.getName());
  /**
   * Mapping between name and icon path
   */
  private static final Map<String, String> nameToIconPath = new HashMap<String, String>();
  /**
   * Renderer customizer used for XSD nodes.
   */
  private final XSDNodeRendererCustomizer xsdNodeRendererCustomizer = new XSDNodeRendererCustomizer();

  // Initialize mapping
  static {
    // message
    nameToIconPath.put("message", getImageURL("/images/node-customizer/wsdlMessage16.png"));

    // part
    nameToIconPath.put("part", getImageURL("/images/node-customizer/wsdlMessagePart16.png"));

    // portType
    nameToIconPath.put("portType", getImageURL("/images/node-customizer/wsdlPortType16.png"));

    // portType/operation
    nameToIconPath.put("portType/operation", getImageURL("/images/node-customizer/wsdlPortTypeOperation16.png"));

    // binding
    nameToIconPath.put("binding", getImageURL("/images/node-customizer/wsdlBinding16.png"));

    // binding/operation
    nameToIconPath.put("binding/operation", getImageURL("/images/node-customizer/wsdlBindingOperation16.png"));

    // service
    nameToIconPath.put("service", getImageURL("/images/node-customizer/wsdlService16.png"));

    // port
    nameToIconPath.put("port", getImageURL("/images/node-customizer/wsdlServicePort16.png"));

    // input
    nameToIconPath.put("input", getImageURL("/images/node-customizer/wsdlOperationInput16.png"));

    // output
    nameToIconPath.put("output", getImageURL("/images/node-customizer/wsdlOperationOutput16.png"));

    // fault
    nameToIconPath.put("fault", getImageURL("/images/node-customizer/wsdlOperationFault16.png"));
    
    // types
    nameToIconPath.put("types", getImageURL("/images/node-customizer/wsdlTypesGroup16.png"));
  }

  /**
   * Searches the class-path for the image with the given path 
   * and if found it returns the string representation of its URL, otherwise it returns <code>null</code>.  
   * 
   * @param path The image path to search for.
   * 
   * @return The string representation of the image URL or <code>null</code> if the image is not found.
   */
  private static String getImageURL(String path) {
    URL imageURL = WSDLNodeRendererCustomizer.class.getResource(path);
    if (imageURL != null) {
      return imageURL.toExternalForm();
    } else {
      logger.error(WSDLNodeRendererCustomizer.class.getName() + " - Image not found: " + path);
      return null;
    }
  }
  
  /**
   * @see ro.sync.exml.workspace.api.node.customizer.XMLNodeRendererCustomizer#getRenderingInformation(ro.sync.exml.workspace.api.node.customizer.NodeRendererCustomizerContext)
   */
  @Override
  public BasicRenderingInformation getRenderingInformation(NodeRendererCustomizerContext context) {
    BasicRenderingInformation renderingInfo = new BasicRenderingInformation();
    String nodeName = context.getNodeName();
    if (nodeName != null) {
      String nodeNS = context.getNodeNamespace();
      String iconPath = null;
      if ("http://schemas.xmlsoap.org/wsdl/".equals(nodeNS)) {
        String imageKey;
        int index = nodeName.indexOf(":");
        if (index == -1) {
          imageKey = nodeName;
        } else {
          imageKey = nodeName.substring(index + 1);
        }
        // Some images that depend both on element name and some attribute value
        // The key is like: elementName @ attributeValue
        if ("operation".equals(imageKey)) {
          String parentElement = context.getParentNodeName();
          if (parentElement != null) {
            index = parentElement.indexOf(":");
            if (index != -1) {
              parentElement = parentElement.substring(index + 1);
            }
            imageKey = parentElement + "/" + imageKey;
          }
        }
        // Get icon path from map.
        iconPath = nameToIconPath.get(imageKey);
        renderingInfo.setIconPath(iconPath);
      } else if ("http://www.w3.org/2001/XMLSchema".equals(nodeNS)) {
        return xsdNodeRendererCustomizer.getRenderingInformation(context);
      }
    }
    return renderingInfo;
  }

  /**
   * @see ro.sync.ecss.extensions.api.Extension#getDescription()
   */
  @Override
  public String getDescription() {
    return "XSLT Node Renderer Customizer";
  }
}