/*
 *  The Syncro Soft SRL License
 *
 *
 *  Copyright (c) 1998-2007 Syncro Soft SRL, Romania.  All rights
 *  reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistribution of source or in binary form is allowed only with
 *  the prior written permission of Syncro Soft SRL.
 *
 *  2. Redistributions of source code must retain the above copyright
 *  notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *  notice, this list of conditions and the following disclaimer in
 *  the documentation and/or other materials provided with the
 *  distribution.
 *
 *  3. The end-user documentation included with the redistribution,
 *  if any, must include the following acknowledgment:
 *  "This product includes software developed by the
 *  Syncro Soft SRL (http://www.sync.ro/)."
 *  Alternately, this acknowledgment may appear in the software itself,
 *  if and wherever such third-party acknowledgments normally appear.
 *
 *  4. The names "Oxygen" and "Syncro Soft SRL" must
 *  not be used to endorse or promote products derived from this
 *  software without prior written permission. For written
 *  permission, please contact support@oxygenxml.com.
 *
 *  5. Products derived from this software may not be called "Oxygen",
 *  nor may "Oxygen" appear in their name, without prior written
 *  permission of the Syncro Soft SRL.
 *
 *  THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 *  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *  OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *  DISCLAIMED.  IN NO EVENT SHALL THE SYNCRO SOFT SRL OR
 *  ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 *  USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 *  ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 *  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 *  OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */
package ro.sync.document;

import java.net.MalformedURLException;
import java.net.URL;
import java.util.HashMap;
import java.util.Map;

import javax.swing.text.Position;

import org.apache.log4j.Logger;

import ro.sync.annotations.api.API;
import ro.sync.annotations.api.APIType;
import ro.sync.annotations.api.SourceType;
import ro.sync.annotations.obfuscate.SkipLevel;
import ro.sync.annotations.obfuscate.SkipObfuscate;
import ro.sync.exml.EXMLResourceBoundle;
import ro.sync.exml.MessageBundle;
import ro.sync.exml.Tags;
import ro.sync.exml.editor.Anchor;
import ro.sync.exml.editor.EditorPageConstants;
import ro.sync.exml.options.Options;
import ro.sync.ui.application.HelpPageProvider;
import ro.sync.util.Equaler;
import ro.sync.util.URLUtil;
import ro.sync.xml.XmlUtil;

/**
 *  This class holds information related to the document, refering to some
 *  errors, or find results. These informations hold the position in the
 *  document, the message and the systemId of the document in which they appear.
 *  If the systemId is null, then the semantics is that the info belongs to the
 *  currently edited file.
 *
 * @author dan
 * @author mircea
 * @author tavy
 * @since Jan 31, 2006, it holds severity level as well.
 */
@SkipObfuscate(classes = SkipLevel.NOT_SPECIFIED, fields = SkipLevel.PUBLIC, methods = SkipLevel.PUBLIC)
@API(type=APIType.EXTENDABLE, src=SourceType.PUBLIC)
public class DocumentPositionedInfo {
  /**
   * I18N
   */
  private static final MessageBundle messages = EXMLResourceBoundle.getResourceBundleInstance();
  /**
   * Logger for logging.
   */
  private static Logger logger = Logger.getLogger(DocumentPositionedInfo.class);
  /**
   * Information message.
   */
  public static final int SEVERITY_INFO = 0;
  /**
   * Warning message.
   */
  public static final int SEVERITY_WARN = 1;
  /**
   * Error message.
   */
  public static final int SEVERITY_ERROR = 2;
  /**
   * Fatal error message.
   */
  public static final int SEVERITY_FATAL = 3;
  /**
   * If one of the components is not known, use this constant for it.
   */
  public static final int NOT_KNOWN = -1;
  /**
   * The message CSS style.
   */
  public static final String MESSAGE_CSS_STYLE = 
      "        table {\n" + 
      "            width: 100%;\n" +
      "            padding: 2px 2px 2px 2px;\n" + 
      "        }\n" + 
      "        th, td {\n" + 
      "            border-width: 1px;\n" +
      "            border-style: solid;\n" + 
      "            border-color: #E8E8E8;\n" + 
      "            text-align: left;\n" +
      "            font-size: 11pt;\n" +
      "        }\n" +
      "        th {\n" +
      "            white-space: nowrap;\n" +
      "        }\n" +
      "        td {\n" +
      "            width: 100%;\n" +
      "        }\n" +
      "        pre {\n" +
      "            font-family: inherit;\n" +
      "            font-size: 11pt;\n" +
      "            text-align: left;\n" +
      "            margin: 0px 0px 0px 0px;\n" +
      "        }\n" ;
  /**
   * The severity level.
   */
  private int severity = SEVERITY_INFO;
  /**
   * Column where the information is placed in the document.
   */
  private int column = NOT_KNOWN;
  /**
   * Line where the information is placed in the document. 1 based.
   */
  private int line = NOT_KNOWN;
  /**
   * Length of the message.
   */
  private int length = NOT_KNOWN;
  /**
   * Provider of the message.
   */
  private MessageProvider messageProvider;
  /**
   * Message's system ID.
   */
  private String systemID;
  /**
   * The URL from which the user can retrieve additional info about the error.
   */
  private URL additionalInfo;
  /**
   * The offset in the document at which the info is located.
   * This is used specially for the find operations. Note that the XML parsers
   * do not offer this kind of information, only the line and column of the event,
   * but we can encapsulate this in the case of an XPath.
   */
  private int offset = NOT_KNOWN;
  /**
   * True if the opener must highlight the entire text from the 1 column to 
   * the column number. This is useful in case of error checks, like validations 
   * or well-formed.
   */  
  private boolean highlightToColumn;
  /**
   * The line in the text where the highlighted region ends.
   */
  private int endLine = NOT_KNOWN;
  /**
   * The column in the text where the highlighted region ends.
   */
  private int endColumn = NOT_KNOWN;
  /**
   * Custom untyped data.
   */
  private DPIData data = null;
  /**
   * The name of the engine who provided the error.
   * Used to be presented in the message.
   */
  private String engineName = null;
  
  /**
   * The page in which to open a new editor if DPI is from another file.
   * If null, use the default page.
   */
  private String imposedInitialPage;
  
  /**
   * Anchor for the error, if available.
   */
  private Anchor anchor;
  
  /**
   * The offset used to highlight the match inside the message in the results panel.
   */
  private int messageHighlightOffset = NOT_KNOWN;
  
  /**
   * The description of an operation (that is an operation which is applied over multiple resources)
   * that generated this DPI.
   */
  private OperationDescription operationDescription;
  
  /**
   * Additional info about DITA completeness check.
   */
  private DITAAdditionalInfo ditaAdditionalInfo;
  /**
   * Map between a page Id and the start and end position of the DPI. 
   * Temporary positions added by an editor page, until the page is closed.
   */
  private Map<String, Position[]> startEndPositionsMap;
  
  /**
   * Used to obtain the severity level to be used when sorting the DPIs by severity.
   * This is done to allow natural sorting by severity, having the <code>SEVERITY_FATAL</code> to be
   * the small value and the <code>SEVERITY_INFO</code> the big value. 
   * 
   * @param severityToFlip The severity of a DPI to be translated into natural ordering.
   * @return The natural ordering value to be used for a DPI' severity level.
   */
  public static int flipSeverity(int severityToFlip) {
    switch (severityToFlip) {
      case SEVERITY_INFO: return SEVERITY_FATAL;
      case SEVERITY_WARN: return SEVERITY_ERROR;
      case SEVERITY_ERROR: return SEVERITY_WARN;
      case SEVERITY_FATAL: return SEVERITY_INFO;
      default: return severityToFlip;
    }
  }

  /**
   *  Constructor for a DPI used only to hold the current position in the editor.
   *  
   * @param offset Start of selection.
   */
  public DocumentPositionedInfo(int offset) {
    this(SEVERITY_INFO, "", "", NOT_KNOWN, NOT_KNOWN, 0, offset, null, false);
  }
  
  /**
   *  Constructor.
   *
   *@param  severity  the severity level of the message.
   *@param  message   the error message.
   *@param  systemID  the system ID.
   *@param  line      the line on which the error occurred in the document.
   *@param  column    the column on which the error occurred.
   *@param  length    the length of the text to be selected.
   */
  public DocumentPositionedInfo(
      int severity, String message, String systemID, int line, int column, int length) {
    this(severity, message, systemID, line, column, length, NOT_KNOWN, null, false);
  }

  /**
   *  Constructor.
   *
   *@param  severity  the severity level of the message.
   *@param  message   the error message.
   *@param  systemID  the system ID.
   *@param  line      the line on which the error occurred in the document.
   *@param  column    the column on which the error occurred.
   *@param  length    the length of the text to be selected.
   *@param  offset    the offset in the document
   *@param  additionalInfo the URL from which the user can retrieve additional info about the error.
   *@param  highlightToColumn true if the opener must highlight the entire text from the 1 column to 
   * the column number. This is useful in case of error checks, like validations or well-formed.
   */
  public DocumentPositionedInfo(
      int severity,
      String message,
      String systemID,
      int line,
      int column,
      int length,
      int offset,
      URL additionalInfo,
      boolean highlightToColumn) {
    this(
        severity,
        new SimpleMessageProvider(message),
        systemID,
        line,
        column,
        NOT_KNOWN,
        NOT_KNOWN,
        length,
        offset,
        additionalInfo,
        highlightToColumn);
  }
  
  /**
   * Constructor. Uses a message provider for the cases in which composing the message takes a lot
   * of time (e.g. XPATH).
   *
   *@param  severity  the severity level of the message.
   *@param  messageProvider   the error message provider.
   *@param  systemID  the system ID.
   *@param  line      the line on which the error occurred in the document.
   *@param  column    the column on which the error occurred.
   *@param  endLine   the highlighted text must end in this line.
   *@param  endColumn   the highlighted text must end in this column.
   *@param  length    the length of the text to be selected.
   *@param  offset    the offset in the document
   *@param  additionalInfo the URL from which the user can retrieve additional info about the error.
   *@param  highlightToColumn true if the opener must highlight the entire text from the 1 column to 
   * the column number. This is useful in case of error checks, like validations or well-formed.
   */
  public DocumentPositionedInfo(int severity, 
      MessageProvider messageProvider, 
      String systemID, 
      int line, 
      int column, 
      int endLine,
      int endColumn,
      int length,
      int offset,
      URL additionalInfo,
      boolean highlightToColumn){
    this.severity = severity;
    this.systemID = systemID;
    this.messageProvider = messageProvider;

    // The line must be 1 based or NOT_KNOWN. 
    if (line > 0) {
      this.line = line;
    } else {
      this.line = NOT_KNOWN;
    }
    this.column = column;
    
    
    // The line must be 1 based or NOT_KNOWN. 
    if (endLine > 0) {
      this.endLine = endLine;
    } else {
      this.endLine = NOT_KNOWN;
    }        
    this.endColumn = endColumn;
    
    this.length = length;
    this.offset = offset;
    this.additionalInfo = additionalInfo;
    this.highlightToColumn = highlightToColumn;
  }

  /**
   *  Constructor.
   *
   *@param  severity  the severity level of the message.
   *@param  message   the error message.
   *@param  systemID  the system ID.
   *@param  line      the line on which the error occurred in the document.
   *@param  column    the column on which the error occurred.
   */
  public DocumentPositionedInfo(int severity, String message, String systemID, int line, int column) {
    this(severity, message, systemID, line, column, NOT_KNOWN);
  }

  /**
   *  Constructor.
   *
   *@param  severity  the severity level of the message.
   *@param  message   the error message.
   *@param  systemID  the system ID.
   */
  public DocumentPositionedInfo(int severity, String message, String systemID) {
    this(severity, message, systemID, NOT_KNOWN, NOT_KNOWN);
  }

  /**
   *  Constructor.
   *
   *@param  severity  the severity level of the message.
   *@param  message  the error message.
   */
  public DocumentPositionedInfo(int severity, String message) {
    this(severity, message, null);
  }
  
  /**
   * Indicates whether some other document positioned information  is "equal to" this one.
   * 
   * @see java.lang.Object#equals(java.lang.Object)
   */
  @Override
  public boolean equals(Object obj) {
    boolean toReturn = false;
    if (obj instanceof DocumentPositionedInfo) {
      DocumentPositionedInfo other = (DocumentPositionedInfo) obj;
      toReturn = 
        Equaler.verifyEquals(systemID, other.getSystemID()) &&
        severity == other.severity &&
        offset == other.offset &&
        length == other.length &&
        line == other.line &&
        endLine == other.endLine &&
        column == other.column &&
        endColumn == other.endColumn &&
        Equaler.verifyEquals(getMessage(), other.getMessage()) &&
        Equaler.verifyEquals(engineName, other.engineName) && 
        Equaler.verifyEquals(operationDescription, other.operationDescription) &&
        Equaler.verifyEquals(ditaAdditionalInfo, other.ditaAdditionalInfo);
    }
    return toReturn;
  }
  
  /**
   * EXM-17723 - hashcode() trebuie sa fie consistent cu equals(), altfel tzapim map-urile
   * 
   * @see java.lang.Object#hashCode()
   */
  @Override
  public int hashCode() {
    int hashCode = line ^ column ^ offset;

    if (systemID != null) {
      hashCode ^= systemID.hashCode();
    }
    
    if (messageProvider != null) {
      String mess = messageProvider.getMessage();
      if(mess != null) {
        hashCode ^= mess.hashCode();
      }
    }
    
    return hashCode;
  }

  /**
   *  Sets the error severity.
   *
   *@param  severity  The new severity value. One of: SEVERITY_INFO, SEVERITY_WARN, SEVERITY_ERROR
   * or SEVERITY_FATAL
   */
  public void setSeverity(int severity) {
    this.severity = severity;
  }

  /**
   *  Sets the column of the event, or <code>NOT_KNOWN</code> if the column is not available.
   *
   *@param  column  The column
   */
  public void setColumn(int column) {
    this.column = column;
  }

  /**
   *  Sets the line of the event, or <code>NOT_KNOWN</code> if the line is not available.
   *
   *@param  line  The new line value, 1 based.
   */
  public void setLine(int line) {
    // The line must be 1 based or NOT_KNOWN. 
    if (line > 0) {
      this.line = line;
    } else {
      this.line = NOT_KNOWN;
    }
  }

  /**
   *  Sets the message. It will clear any user info if there is present into the message.
   *
   *@param  message  The new message
   */
  public void setMessage(String message) {
    messageProvider.setMessage(message);
  }
  
  /**
   * Set mask passwords in URLs, <code>true</code> by default.
   * 
   * @param maskPasswordsInURLs <code>true</code> to mask passwords in URLs which are part of the message.
   */
  public void setMaskPasswordsInURLs(boolean maskPasswordsInURLs) {
    messageProvider.setMaskPasswordsInURLs(maskPasswordsInURLs);
  }

  /**
   *  Sets the length of the selected text.
   *
   *@param  length  The length of the selected text.
   */
  public void setLength(int length) {
    this.length = length;
  }

  /**
   *  Sets the systemID of the event.
   *
   *@param  systemID  The new systemID value
   */
  public void setSystemID(String systemID) {
    this.systemID = systemID;
  }
  
  /**
   *  Gets the length of the text that will be selected in the editor.
   *
   *@return    The length of the text to be selected.
   */
  public int getLength() {
    return length;
  }

  /**
   *  Gets the column of the event.
   *
   *@return    The column value, 1 based. Can be NOT_KNOWN.
   */
  public int getColumn() {
    return column;
  }

  /**
   *  Gets the severity level. One of: SEVERITY_INFO, SEVERITY_WARN, SEVERITY_ERROR
   * or SEVERITY_FATAL
   *
   *@return    The severity value
   */
  public int getSeverity() {
    return severity;
  }
  
  /**
   *  Gets the severity level as string
   *
   *@return    The severity value string
   */
  public String getSeverityAsString() {
    String severityLevel = "info";
    switch (severity) {
      case SEVERITY_WARN:
        severityLevel = "warning";
        break;
      case SEVERITY_ERROR:
        severityLevel = "error";
        break;
      case SEVERITY_FATAL:
        severityLevel = "fatal error";
    }
    return severityLevel;
  }

  /**
   *  Gets the line attribute of the event
   *
   *@return    The line value, 1 based. Can be NOT_KNOWN.
   */
  public int getLine() {
    return line;
  }

  /**
   *  Gets the message
   *
   *@return The message value
   */
  public String getMessage() {
    return messageProvider.getMessage();
  }  
  
  /**
   * The provider of messages from the current document position info.
   * 
   * @return Returns the message provider.
   */
  public MessageProvider getMessageProvider() {
    return messageProvider;
  }
  
  /**
   *  Gets the message prepended with the engine name if available.
  *
  *@return The message value
  */
 public String getMessageWithEngine() {
   String preferredEngine = engineName; 
   if (ditaAdditionalInfo != null && ditaAdditionalInfo.getProblemType() != null) {
     preferredEngine = ditaAdditionalInfo.getProblemType();
   }
   return (preferredEngine != null ? ("[" + preferredEngine + "] " + getMessage()) : getMessage());
 }
  
  /**
   * Get the message with severity info as a char in the front of the message (W, E, F).
   * 
   * @return The message prefixed with the severity char. For INFO we do not present such a char.
   */
  public String getMessageWithSeverity() {
    String severityLevel = "";
    switch (severity) {
      case SEVERITY_WARN:
        severityLevel = "W ";
        break;
      case SEVERITY_ERROR:
        severityLevel = "E ";
        break;
      case SEVERITY_FATAL: 
        severityLevel = "F ";
    }
    
    String res = null;
    String messageWithEngine = getMessageWithEngine();
    if(messageWithEngine != null) {
      res = severityLevel + messageWithEngine;
    }
    return res;
  }

  /**
   *  Gets the systemID attribute of the DocumentPositionedInfo object
   *
   *@return    The systemID value
   */
  public String getSystemID() {
    return systemID;
  }

  /**
   * Sets the event's offset. can be NOT_KNOWN
   *
   * @param offset the new offset.
   */
  public void setOffset(int offset) {
    this.offset = offset;
  }

  /**
   * Get the event's offset. Can be NOT_KNOWN
   * 
   * @return the offset
   */
  public int getOffset() {
    return offset;
  }

  /**
   *  Gets the string representation.
   *
   *@return    the string representation.
   */
  @Override
  public String toString() {
    // TOString is used sometimes by the renderers
    // If the message provider takes a long time to compute the message, do not put its value here.
    final String msg =
      (messageProvider instanceof SimpleMessageProvider) ? messageProvider.getMessage()
                                                         : " not available";

    final StringBuilder buffer = new StringBuilder("[");
    buffer.append("SystemID: ").append(systemID);
    buffer.append("\nSeverity: ").append(getSeverityAsString());
    buffer.append("\nLine: ").append(line);
    buffer.append("\nColumn: ").append(column);
    buffer.append("\nEndLine: ").append(endLine);
    buffer.append("\nEndColumn: ").append(endColumn);
    buffer.append("\nLength: ").append(length);
    buffer.append("\nOffset: ").append(offset);
    buffer.append("\nMessage: ").append(msg);
    buffer.append(additionalInfo == null ? "" : "\nAdditional: " + additionalInfo);
    String mainValidatedFile = null; 
    if (operationDescription != null) {
      mainValidatedFile = operationDescription.getMainFileURL();
    }
    buffer.append(mainValidatedFile == null ? "" : "\nMain validated file: " + mainValidatedFile);
    buffer.append("]");
    return buffer.toString();
  }

  /**
   * The URL for additional information if available.
   * 
   * @return The additional information URL, or <code>null</code>.
   */
  public URL getAdditionalInfo() {
    return additionalInfo;
  }

  /**
   * Set the additional information URL if available.
   * 
   * @param url The URL at which the user can find additional information.
   */
  public void setAdditionalInfo(URL url) {
    additionalInfo = url;
  }

  /**
   * Check if the highlight must be made from column 0 to the specified column.
   *  
   * @return True if the highlight must be made from column 0 to the specified column.
   */
  public boolean isHighlightToColumn() {
    return highlightToColumn;
  }
  
  /**
   * Set highlight strategy for the document positioned information visual presentation.
   * 
   * @param how <code>true</code> if the visual part should highlight from column 0 to specified column
   * <code>false</code> otherwise. 
   */
  public void setHighlightToColumn(boolean how) {
    highlightToColumn = how;
  }

  /**
   * Get the highlight end column, or NOT_KNOWN if not available.
   * 
   * @return the highlight end column, or NOT_KNOWN.
   */
  public int getEndColumn() {
    return endColumn;
  }

  /**
   * Get the highlight end line, or NOT_KNOWN if not available.
   * 
   * @return the highlight end line or NOT_KNOWN.
   */
  public int getEndLine() {
    return endLine;
  }

  /**
   * Sets the end line of the highlight.
   * 
   * @param endLine The end line of the highlight.
   */
  public void setEndLine(int endLine) {
    this.endLine = endLine;
  }  
  
  /**
   * Sets the end column of the highlight.
   * 
   * @param endColumn The end column of the highlight.
   */
  public void setEndColumn(int endColumn) {
    this.endColumn = endColumn;
  }  
  
  /**
   * Setter for the user managed data.
   * 
   * @param data The user managed data.
   */
  public void setData(DPIData data) {
    this.data = data;
  }
  
  /**
   * Getter for the user managed data.
   * 
   * @return The user managed data.
   */
  public DPIData getData() {
    return data;
  }
  
  /**
   * Set the name of the engine who provided the error.
   * 
   * @param engineName The source engine name who provided the error.
   */
  public void setEngineName(String engineName) {
    this.engineName = engineName;
  }
  
  /**
   * Get the name of the engine who provided the error.
   * 
   * @return Returns the source engine name who provided the error.
   * <code>null</code> if no was engine set.
   */
  public String getEngineName() {
    return engineName;
  }
  
  /**
   * Build the XML representation of this DPI.
   * 
   * @return The XML representation of the DPI.
   */
  public String asXML() {
    return asXML(true, true, true, true, true, true);
  }
  
  /**
   * Build the XML representation of this DPI.
   * 
   * @param includeSeverity <code>true</code> if severity details should be included. 
   * @param includeAdditionalInfo <code>true</code> if additional info details should be included.
   * @param includeDescription  <code>true</code> if description details should be included.
   * @param includeSystemID  <code>true</code> if system ID details should be included.
   * @param includeLocation  <code>true</code> if location details should be included.
   * @param includeOperationDescription  <code>true</code> if operation description details should be included.
   * 
   * @return The XML representation of the DPI.
   */
  public String asXML(
      boolean includeSeverity,
      boolean includeAdditionalInfo,
      boolean includeDescription,
      boolean includeSystemID,
      boolean includeLocation,
      boolean includeOperationDescription) {
    
    String lineSeparator = 
      Options.getInstance().getStringProperty(Options.EDITOR_LINE_SEPARATOR);

    // Informations from the DPI.
    String message = getMessage();

    // The buffer where to put the result
    StringBuilder buf = new StringBuilder();

    // Test to see if the DPI has description or system ID informations.
    // Else we don't have what informations to put in the <incident> tag. 
    if ((message != null && message.length() > 0)
        || (systemID != null && systemID.length() > 0)) {

      // Begin of a new DPI
      buf.append("<incident>");
      buf.append(lineSeparator);

      // Test if the engine name is OK
      if (engineName != null && engineName.length() > 0) {
        buf.append("<engine>").append(engineName).append("</engine>");
        buf.append(lineSeparator);
      }

      if (includeSeverity) {
        // Problem severity
        buf.append("<severity>");
        switch (severity) {
          case SEVERITY_INFO:
            buf.append("info");
            break;
          case SEVERITY_WARN:
            buf.append("warning");
            break;
          case SEVERITY_ERROR:
            buf.append("error");
            break;
          case SEVERITY_FATAL:
            buf.append("fatal");
            break;
        }
        buf.append("</severity>");
        buf.append(lineSeparator);
      }

      if (includeDescription) {
        // Test if description is null or the empty string
        if (message != null && message.length() > 0) {
          String originalMessage = message;
          message = XmlUtil.escape(message);
          buf.append("<description>").append(message).append("</description>");
          buf.append(lineSeparator);

          //Print match text.
          //Guard some overflows
          if(messageHighlightOffset >= 0 && length > 0 
              && length < originalMessage.length() 
              && messageHighlightOffset < originalMessage.length() ) {
            buf.append("<match>").append(XmlUtil.escape(originalMessage.substring(messageHighlightOffset, 
                messageHighlightOffset + length))).append("</match>");
            buf.append(lineSeparator);
          }
        }
      }
      
      if (includeSystemID) {
        // Test if systemID is null or the empty string
        if (systemID != null && systemID.length() > 0) {
          try {
            String sysID = URLUtil.getDescription(new URL(systemID));
            sysID = XmlUtil.escape(sysID);
            buf.append("<systemID>").append(sysID).append("</systemID>");
            buf.append(lineSeparator);
          } catch (MalformedURLException e) {
            if (logger.isDebugEnabled()){
              logger.debug("'" + systemID + "'" + " is not a URL, remains unchanged");
            }
          }
        }
      }
      
      if (includeOperationDescription) {
        // Append operation description if not null.
        if (operationDescription != null) {
          buf.append(operationDescription.asXML());
          buf.append(lineSeparator);
        }
      }

      if (ditaAdditionalInfo != null) {
        buf.append(ditaAdditionalInfo.asXML());
        buf.append(lineSeparator);
      }
      
      if (includeLocation) {
        // Add location only if start line is known.
        if (line != DocumentPositionedInfo.NOT_KNOWN) {
          // We have location
          buf.append("<location>");
          buf.append(lineSeparator);

          // Start line and column
          buf.append("<start>");
          buf.append(lineSeparator);

          // Add line
          buf.append("<line>").append(line).append("</line>");
          buf.append(lineSeparator);

          // Add column
          buf.append("<column>");
          // Test if column is not -1
          if(column != DocumentPositionedInfo.NOT_KNOWN){
            buf.append(column);
          } else {
            // If it is, then we put 0 (zero) as value
            buf.append('0');
          }
          buf.append("</column>");
          buf.append(lineSeparator);

          // End start line and column
          buf.append("</start>");
          buf.append(lineSeparator);

          if (endLine != DocumentPositionedInfo.NOT_KNOWN) {
            // Add end line and column
            buf.append("<end>");
            buf.append(lineSeparator);

            // Add endline
            buf.append("<line>").append(endLine).append("</line>");
            buf.append(lineSeparator);

            // Add endcolumn if we have it
            if (endColumn != DocumentPositionedInfo.NOT_KNOWN) {
              buf.append("<column>").append(endColumn).append("</column>");
              buf.append(lineSeparator);
            }

            // The end
            buf.append("</end>");
            buf.append(lineSeparator);
          }

          buf.append("</location>");
          buf.append(lineSeparator);
        }

        // Add the offset info.
        if (offset != NOT_KNOWN) {
          buf.append("<offset>").append(offset).append("</offset>");
          buf.append(lineSeparator);
        }

        // Add the length info.
        if (length > 0) {
          buf.append("<length>").append(length).append("</length>");
          buf.append(lineSeparator);
        }
      }

      if (includeAdditionalInfo) {
        // Add the additional informations where you can find out more based on the error message.
        if (additionalInfo != null 
        		// EXM-28644 An internal link, it should not appear in the report
        		&& !additionalInfo.toString().contains(HelpPageProvider.INTERNAL_HELP_PREFIX)) {
          buf.append("<additionalInfo>");
          buf.append(XmlUtil.escape(additionalInfo.toString()));
          buf.append("</additionalInfo>");
          buf.append(lineSeparator);
        }
      }

      buf.append("</incident>");
    }

    return buf.toString();
  }
  
  /**
   * Build the text representation of this DPI. DIFFERENT from toString() !!!
   * 
   * @return The text representation of this DPI.
   */
  public String asText() {
    return asText(true, true, true, true, true, true);
  }
  
  /**
   * Build the text representation of this DPI. DIFFERENT from toString() !!!
   * 
   * @param includeSeverity <code>true</code> if severity details should be included. 
   * @param includeAdditionalInfo <code>true</code> if additional info details should be included.
   * @param includeDescription  <code>true</code> if description details should be included.
   * @param includeSystemID  <code>true</code> if system ID details should be included.
   * @param includeLocation  <code>true</code> if location details should be included.
   * @param includeOperationDescription  <code>true</code> if operation description details should be included.
   * 
   * @return The text representation of this DPI.
   */
  public String asText(
      boolean includeSeverity,
      boolean includeAdditionalInfo,
      boolean includeDescription,
      boolean includeSystemID,
      boolean includeLocation,
      boolean includeOperationDescription) {
    final String lineSeparator =
      Options.getInstance().getStringProperty(Options.EDITOR_LINE_SEPARATOR);

    String systemID = this.systemID;
    try {
      systemID = URLUtil.getDescription(new URL(systemID));
    } catch (MalformedURLException e) {
      if (logger.isDebugEnabled()){
        logger.debug("'" + systemID + "'" + " is not a URL, remains unchanged");
      }
    }

    String message = getMessage();
    StringBuilder buf = new StringBuilder();

    if ((message != null && message.length() > 0)
        || (systemID != null && systemID.length() > 0)) {

      //Add the systemID.
      if (includeSystemID) {
        if (systemID != null && systemID.length() > 0) {
          buf.append(messages.getString(Tags.SYSTEM_ID));
          buf.append(": ").append(systemID);
          buf.append(lineSeparator);
        }
      }
      
      if (includeOperationDescription) {
        //Add the main validate file.
        if (operationDescription != null) {
          buf.append(operationDescription.asText());
          buf.append(lineSeparator);
        }
      }

      // Add the engine name.
      if (engineName != null && engineName.length() > 0) {
        buf.append(messages.getString(Tags.ENGINE_NAME)).append(": ").append(engineName);
        buf.append(lineSeparator);
      }

      if (includeSeverity) {
        // Problem severity
        buf.append(messages.getString(Tags.SEVERITY)).append(": ");
        switch (severity) {
          case SEVERITY_INFO:
            buf.append("info");
            break;
          case SEVERITY_WARN:
            buf.append("warning");
            break;
          case SEVERITY_ERROR:
            buf.append("error");
            break;
          case SEVERITY_FATAL:
            buf.append("fatal");
            break;
        }
        buf.append(lineSeparator);
      }

      if (includeDescription) {
        // Add the message
        if (message != null) {
          buf.append(messages.getString(Tags.DESCRIPTION)).append(": ").append(message);
          buf.append(lineSeparator);
          
          //Print match text.
          //Guard some overflows
          if(messageHighlightOffset >= 0 && length > 0 
              && length < message.length() 
              && messageHighlightOffset < message.length() ) {
            buf.append(messages.getString(Tags.MATCH)).append(": ").append(message.substring(messageHighlightOffset, messageHighlightOffset + length));
            buf.append(lineSeparator);
          }
        }
      }
      
      if (ditaAdditionalInfo != null) {
        buf.append(ditaAdditionalInfo.asText());
        buf.append(lineSeparator);
      }

      if (includeLocation) {
        // Add locations
        if (line != NOT_KNOWN) {
          // Add the start location
          buf.append(messages.getString(Tags.START_LOCATION)).append(": ").append(line).append(":");
          if (column != NOT_KNOWN) {
            buf.append(column);
          } else {
            buf.append('0');        
          }
          buf.append(lineSeparator);

          // Add the end location
          if (endLine != NOT_KNOWN) {
            buf.append(messages.getString(Tags.END_LOCATION)).append(": ").append(endLine).append(":");
            if (endColumn != NOT_KNOWN) {
              buf.append(endColumn);
            } else {
              buf.append('0');
            }
            buf.append(lineSeparator);
          }
        }

        // Add the offset
        if (offset != NOT_KNOWN) {
          buf.append(messages.getString(Tags.OFFSET)).append(": ").append(offset);
          buf.append(lineSeparator);
        }

        // Add the length info
        if (length > 0) {
          buf.append(messages.getString(Tags.LENGTH)).append(": ").append(length);
          buf.append(lineSeparator);
        }
      }

      if (includeAdditionalInfo) {
        // Add additional info
        if (additionalInfo != null 
        		// EXM-28644 An internal link, it should not appear in the report
        		&& !additionalInfo.toString().contains(HelpPageProvider.INTERNAL_HELP_PREFIX)) {
          buf.append("URL: ").append(additionalInfo);
          buf.append(lineSeparator);
        }
      }
    }

    return buf.toString();
  }
  
  /**
   * Get the imposed initial page if a new editor will be opened.
   * 
   * @return the imposed initial page
   */
  public String getImposedInitialPage() {
    return imposedInitialPage;
  }

  /**
   * Set the imposed initial page in which to open the DPI for a new editor.
   * 
   * @param imposedInitialPage The imposed initial page.
   */
  public void setImposedInitialPage(String imposedInitialPage) {
    this.imposedInitialPage = imposedInitialPage;
  }
  
  /**
   * Get the error anchor if any.
   * 
   * @return the error anchor if any.
   */
  public Anchor getAnchor() {
    return anchor;
  }
  
  /**
   * Set the error anchor. 
   * 
   * @param anchor The anchor to set.
   */
  public void setAnchor(Anchor anchor) {
    this.anchor = anchor;
  }
  
  /**
   * If specified, indicates the matched part of the description.
   * It is composed of a start and end offset from the description text.
   */
  private int[] matchRange;

  /**
   * Get the match range if any.
   * 
   * @return The match range if possible.
   */
  public int[] getMatchRange() {
    return matchRange;
  }

  /**
   * Set the match range.
   * 
   * @param matchRange The match range.
   */
  public void setMatchRange(int[] matchRange) {
    this.matchRange = matchRange;
  }
  
  /**
   * Return the offset used to highlight the match inside the message in the results panel.
   * 
   * @return the message highlight offset.
   */
  public int getMessageHighlightOffset() {
    return messageHighlightOffset;
  }

  /**
   * Set the offset used to highlight the match inside the message in the results panel.
   * 
   * @param messageHighlightOffset The offset to be set.
   */
  public void setMessageHighlightOffset(int messageHighlightOffset) {
    this.messageHighlightOffset = messageHighlightOffset;
  }

  /**
   * Sets the description of an operation (that is an operation which is applied over multiple resources)
   * that generated this DPI. Can be <code>null</code>.
   *
   * @param operationDescription the description of an operation 
   * (that is an operation which is applied over multiple resources)
   * that generated this DPI. Can be <code>null</code>.
   */
  public void setOperationDescription(OperationDescription operationDescription) {
    this.operationDescription = operationDescription;
  }

  /** 
   * The description of an operation (that is an operation which is applied over multiple resources)
   * that generated this DPI. Can be <code>null</code>.
   * @return the description of an operation (that is an operation which is applied over multiple resources)
   * that generated this DPI. Can be <code>null</code>.
   */
  public OperationDescription getOperationDescription() {
    return operationDescription;
  }

  /**
   * Additional info about DITA completeness check.
   * 
   * @return Returns additional info.
   */
  public DITAAdditionalInfo getDITAAdditionalInfo() {
    return ditaAdditionalInfo;
  }

  /**
   * Additional info about DITA completeness check.
   * 
   * @param additionalInfo Additional info about DITA completeness check.
   */
  public void setDITAAdditionalInfo(DITAAdditionalInfo additionalInfo) {
    this.ditaAdditionalInfo = additionalInfo;
  }

  /**
   * Set the temporary start and end positions used for keeping the DPI synchronized with the 
   * highlight range from the page.
   * 
   * @param startPosition The start position.
   * @param endPosition   The end position.
   * @param pageId        The id of the page for which the positions were created,
   *                      one of {@link EditorPageConstants} constants.
   */
  public void setTemporaryPositions(Position startPosition, Position endPosition, String pageId) {
    if (startEndPositionsMap == null) {
      startEndPositionsMap = new HashMap<String, Position[]>();
    }
    if (startPosition == null && endPosition == null) {
      startEndPositionsMap.remove(pageId);
    } else {
      startEndPositionsMap.put(pageId, new Position[] {startPosition, endPosition});
    }
  }

  /**
   * Get the temporary start and end positions for the current page.
   * 
   * @param pageId The id of the page for which we want the positions,
   *               one of {@link EditorPageConstants} constants.
   *
   * @return An arrays with two positions, representing the start and end positions of the DPI.
   *         Can be <code>null</code> if the page is closed, or the DPI is not synchronized with 
   *         the modifications from editor. 
   */
  public Position[] getStartEndPositions(String pageId) {
    return startEndPositionsMap != null ? startEndPositionsMap.get(pageId) : null;
  }
  
  /**
   * Build the HTML representation of this DPI.
   * 
   * @return The HTML representation of this DPI.
   */
  public String asHTML() {
    return asHTML(true, true, true, true, true, true);
  }

  /**
   * Build the HTML representation of this DPI.
   * 
   * @param includeSeverity <code>true</code> if severity details should be included. 
   * @param includeAdditionalInfo <code>true</code> if additional info details should be included.
   * @param includeDescription  <code>true</code> if description details should be included.
   * @param includeSystemID  <code>true</code> if system ID details should be included.
   * @param includeLocation  <code>true</code> if location details should be included.
   * @param includeOperationDescription  <code>true</code> if operation description details should be included.
   * 
   * @return The HTML representation of this DPI.
   */
  public String asHTML(
      boolean includeSeverity,
      boolean includeAdditionalInfo,
      boolean includeDescription,
      boolean includeSystemID,
      boolean includeLocation,
      boolean includeOperationDescription) {
    String message = getMessage();
    StringBuilder buf = new StringBuilder();
  
    buf.append("<html>").append("\n");
    buf.append(
        "<head>\n").append("    <title>Message</title>\n");
    buf.append("    <style type=\"text/css\">\n");
    buf.append(MESSAGE_CSS_STYLE).append("    </style>\n</head>");
    buf.append("<body>").append("\n");
    buf.append("<table>").append("\n");
    buf.append("<tbody>").append("\n");
    
    if (includeDescription) {
      // Add the message
      if (message != null) {
        buf.append("<tr>").append("\n").append("<th valign=\"top\">");
        buf.append(messages.getString(Tags.DESCRIPTION));
        buf.append("</th>").append("\n");
        buf.append("<td>");
        
        message = message.trim();
        boolean usePre = message.indexOf('\n') != -1;
        if (usePre) {
          buf.append("<pre>");
        }
        buf.append(XmlUtil.escape(message));
        if (usePre) {
          buf.append("</pre>");
        }
        buf.append("</td>").append("\n");
        buf.append("</tr>").append("\n");
        
        //Keep the original message for the match range. 
        String originalMessage = getMessage();
        //Print match text.
        //Guard some overflows
        if(messageHighlightOffset >= 0 && length > 0 
            && length < originalMessage.length() 
            && messageHighlightOffset < originalMessage.length() ) {
          buf.append("<tr>").append("\n").append("<th valign=\"top\">");
          buf.append(messages.getString(Tags.MATCH));
          buf.append("</th>").append("\n");
          buf.append("<td>");
          
          usePre = originalMessage.indexOf('\n') != -1;
          if (usePre) {
            buf.append("<pre>");
          }
          buf.append(XmlUtil.escape(originalMessage.substring(messageHighlightOffset, messageHighlightOffset + length)));
          if (usePre) {
            buf.append("</pre>");
          }
          buf.append("</td>").append("\n");
          buf.append("</tr>").append("\n");
        }
      }
    }
    
    String additionalHtmlContent = getAdditionalHtmlContent();
    if (additionalHtmlContent != null) {
      buf.append(additionalHtmlContent);
    }
    
    if (includeSeverity) {
      // Problem severity
      buf.append("<tr>").append("\n").append("<th valign=\"top\">");
      buf.append(messages.getString(Tags.SEVERITY));
      buf.append("</th>").append("\n").append("<td>");
      switch (severity) {
        case SEVERITY_INFO:
          buf.append("Info");
          break;
        case SEVERITY_WARN:
          buf.append("Warning");
          break;
        case SEVERITY_ERROR:
          buf.append("Error");
          break;
        case SEVERITY_FATAL:
          buf.append("Fatal");
          break;
      }
      buf.append("</td>").append("\n");
      buf.append("</tr>").append("\n");
    }
    
    
    String systemID = this.systemID;
    try {
      systemID = URLUtil.getDescription(new URL(systemID));
    } catch (MalformedURLException e) {
      if (logger.isDebugEnabled()){
        logger.debug("'" + systemID + "'" + " is not a URL, remains unchanged");
      }
    }
  
    
    if ((message != null && message.length() > 0)
        || (systemID != null && systemID.length() > 0)) {
  
      //Add the systemID.
      if (includeSystemID) {
        if (systemID != null && systemID.length() > 0) {
          buf.append("<tr>").append("\n").append("<th valign=\"top\">");
          buf.append(messages.getString(Tags.SYSTEM_ID));
          buf.append("</th>").append("\n").append("<td>").append("\n");
          buf.append("<a href=\"").append(XmlUtil.escapeAttributeValue(systemID)).append("\">");
          buf.append(XmlUtil.escape(systemID));
          buf.append("</a>").append("\n");
          buf.append("</td>").append("\n");
          buf.append("</tr>").append("\n");
        }
      }
      
      if (includeOperationDescription) {
        //Add the main validate file.
        if (operationDescription != null) {
          buf.append(operationDescription.asHTML());
        }
      }
  
      // Add the engine name.
      if (engineName != null && engineName.length() > 0) {
        buf.append("<tr>").append("\n").append("<th valign=\"top\">");
        buf.append(messages.getString(Tags.ENGINE_NAME));
        buf.append("</th>").append("\n").append("<td>");
        buf.append(XmlUtil.escape(engineName));
        buf.append("</td>").append("\n");
        buf.append("</tr>").append("\n");
      }
  
      if (ditaAdditionalInfo != null) {
        buf.append(ditaAdditionalInfo.asHTML());
      }
  
      if (includeLocation) {
        // Add locations
        if (line != NOT_KNOWN) {
          // Add the start location
          buf.append("<tr>").append("\n").append("<th valign=\"top\">");
          buf.append(messages.getString(Tags.START_LOCATION));
          buf.append("</th>").append("\n").append("<td>");
          buf.append("line: ").append(line).append(", ").append("column: ");
          if (column != NOT_KNOWN) {
            buf.append(column);
          } else {
            buf.append('0');        
          }
          buf.append("</td>").append("\n");
          buf.append("</tr>").append("\n");
  
          // Add the end location
          if (endLine != NOT_KNOWN) {
            buf.append("<tr>").append("\n").append("<th valign=\"top\">");
            buf.append(messages.getString(Tags.END_LOCATION));
            buf.append("</th>").append("\n").append("<td>");
            buf.append("line: ").append(endLine).append(", ").append("column: ");
            if (endColumn != NOT_KNOWN) {
              buf.append(endColumn);
            } else {
              buf.append('0');
            }
            buf.append("</td>").append("\n");
            buf.append("</tr>").append("\n");
          }
        }
  
        // Add the offset
        if (offset != NOT_KNOWN) {
          buf.append("<tr>").append("\n").append("<th valign=\"top\">");
          buf.append(messages.getString(Tags.OFFSET));
          buf.append("</th>").append("\n").append("<td>");
          buf.append(offset);
          buf.append("</td>").append("\n");
          buf.append("</tr>").append("\n");
        }
  
        // Add the length info
        if (length > 0) {
          buf.append("<tr>").append("\n").append("<th valign=\"top\">");
          buf.append(messages.getString(Tags.LENGTH));
          buf.append("</th>").append("\n").append("<td>");
          buf.append(length);
          buf.append("</td>").append("\n");
          buf.append("</tr>").append("\n");
        }
      }
  
      if (includeAdditionalInfo) {
        // Add additional info
        if (additionalInfo != null) {
          buf.append("<tr>").append("\n").append("<th valign=\"top\">");
          buf.append(messages.getString(Tags.ADDITIONAL_INFO));
          buf.append("</th>").append("\n").append("<td>");
          String infoString = additionalInfo.toString();
          buf.append("<a href=\"").append(XmlUtil.escapeAttributeValue(infoString)).append("\">");
          if (infoString.contains(HelpPageProvider.INTERNAL_HELP_PREFIX)) {
             // EXM-27334 Internal link, do not show the real link
            buf.append(messages.getString(Tags.MORE_DETAILS)).append("</a>");
          } else {
            buf.append(XmlUtil.escape(infoString)).append("</a>");
          }
          buf.append("</td>").append("\n");
          buf.append("</tr>").append("\n");
        }
      }
    }
    
    buf.append("</tbody").append("\n");
    buf.append("</table>").append("\n");
    buf.append("</body>").append("\n");
    buf.append("</html>");
  
    return buf.toString();
  }
  
  /**
   * Return the additional HTML content, if any.
   * 
   * @return the additional HTML content, if any.
   */
  protected String getAdditionalHtmlContent() {
    return null;
  }
}