/*
 *  The Syncro Soft SRL License
 *
 *  Copyright (c) 1998-2012 Syncro Soft SRL, Romania.  All rights
 *  reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistribution of source or in binary form is allowed only with
 *  the prior written permission of Syncro Soft SRL.
 *
 *  2. Redistributions of source code must retain the above copyright
 *  notice, this list of conditions and the following disclaimer.
 *
 *  3. Redistributions in binary form must reproduce the above copyright
 *  notice, this list of conditions and the following disclaimer in
 *  the documentation and/or other materials provided with the
 *  distribution.
 *
 *  4. The end-user documentation included with the redistribution,
 *  if any, must include the following acknowledgment:
 *  "This product includes software developed by the
 *  Syncro Soft SRL (http://www.sync.ro/)."
 *  Alternately, this acknowledgment may appear in the software itself,
 *  if and wherever such third-party acknowledgments normally appear.
 *
 *  5. The names "Oxygen" and "Syncro Soft SRL" must
 *  not be used to endorse or promote products derived from this
 *  software without prior written permission. For written
 *  permission, please contact support@oxygenxml.com.
 *
 *  6. Products derived from this software may not be called "Oxygen",
 *  nor may "Oxygen" appear in their name, without prior written
 *  permission of the Syncro Soft SRL.
 *
 *  THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 *  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *  OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *  DISCLAIMED.  IN NO EVENT SHALL THE SYNCRO SOFT SRL OR
 *  ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 *  USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 *  ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 *  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 *  OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */
package ro.sync.ecss.component.editor;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyAdapter;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;

import javax.swing.AbstractButton;
import javax.swing.Icon;
import javax.swing.JButton;
import javax.swing.border.Border;

import ro.sync.annotations.api.API;
import ro.sync.annotations.api.APIType;
import ro.sync.annotations.api.SourceType;
import ro.sync.ecss.extensions.api.CursorType;
import ro.sync.ecss.extensions.api.editor.AbstractInplaceEditor;
import ro.sync.ecss.extensions.api.editor.AuthorInplaceContext;
import ro.sync.ecss.extensions.api.editor.EditingEvent;
import ro.sync.ecss.extensions.api.editor.IAuthorExtensionAction;
import ro.sync.ecss.extensions.api.editor.InplaceEditorArgumentKeys;
import ro.sync.ecss.extensions.api.editor.InplaceRenderer;
import ro.sync.ecss.extensions.api.editor.RendererLayoutInfo;
import ro.sync.exml.EXMLResourceBoundle;
import ro.sync.exml.IDEAccess;
import ro.sync.exml.MessageBundle;
import ro.sync.exml.Tags;
import ro.sync.exml.view.graphics.Dimension;
import ro.sync.exml.view.graphics.Font;
import ro.sync.exml.view.graphics.Point;
import ro.sync.exml.view.graphics.Rectangle;
import ro.sync.ui.ImageUtil;
import ro.sync.ui.application.ApplicationBorderFactory;
import ro.sync.ui.tooltip.MultiLineToolTip;
import ro.sync.util.PlatformDetector;

/**
 * A button that can be used to invoke an author extension action.
 * 
 * @author alex_jitianu
 */
@API(type=APIType.INTERNAL, src=SourceType.PUBLIC)
public class ButtonEditor extends AbstractInplaceEditor implements InplaceRenderer {
  /** 
   * The messages resource bundle. 
   */
  protected static final MessageBundle messages = EXMLResourceBoundle.getResourceBundleInstance();
  
  /**
   * Button used to render and edit.
   */
  JButton button;
  /**
   * The action listener set on the button.
   */
  private ActionListener actionListener;
  /**
   * To handle ESC.
   */
  private KeyListener keyListener;
  
  /**
   * The default font.
   */
  private final java.awt.Font defaultFont;
  
  /**
   * The default border
   */
  private Border defaultBorder;
  
  /**
   * The default foreground color.
   */
  private Color defaultForegroundColor;
  
  /**
   * Constructor.
   */
  public ButtonEditor() {
    button = new JButton() {
      @Override
      public javax.swing.JToolTip createToolTip() {
        MultiLineToolTip tooltip = new MultiLineToolTip();
        tooltip.setComponent(this);
        return tooltip;
      }
      
      @Override
      protected void paintComponent(Graphics g) {
        if ( 
            // Only if it's flat.
            !isContentAreaFilled() 
            && getModel().isRollover()) {
          // A kind of yellow.
          paintRolloverBox(g, this.getWidth(), this.getHeight());
        }
        
        super.paintComponent(g);
        
        if ( 
            // Only if it's flat.
            !isContentAreaFilled() && isFocusOwner()) {
            // Draw a rectangle to highlight the focus
            ((Graphics2D)g).setStroke(new BasicStroke(1, BasicStroke.CAP_BUTT, BasicStroke.JOIN_BEVEL, 0, new float[]{1}, 0));
            g.drawRect(0, 0, this.getWidth() - 1, this.getHeight() - 1);
        }
      }
    };
    
    defaultFont = button.getFont();
    defaultForegroundColor = button.getForeground();
    defaultBorder = button.getBorder();
  }
  
  
  /**
   * Paints a colored box to simulate the roll-over effect on buttons.
   * 
   * @param g the graphic object to paint on.
   * @param width the width of the box.
   * @param height the height of the box.
   */
  static void paintRolloverBox(Graphics g, int width, int height) {
    g.setColor(new Color(255, 232, 118));
    g.fillRect(0, 0, width - 1, height - 1);
    g.setColor(Color.GRAY);
    g.drawRect(0, 0, width - 1, height - 1);
  }
  
  /**
   * Decorates the JButton to be used as an inplace editor.
   * 
   * @param button the button to be decorated.
   * @param context the context in which the button should be decorated.
   * @param defaultBorder the default border.
   */
  static void decorateJButton(AbstractButton button, AuthorInplaceContext context, Border defaultBorder) {
    // If the mouse is over the button paint the highlight.
    button.getModel().setRollover(context.getRelativeMouseLocation() != null);

    boolean isSA = true;
    if (context.getAuthorAccess() != null 
        && context.getAuthorAccess().getWorkspaceAccess() != null) {
      isSA = context.getAuthorAccess().getWorkspaceAccess().isStandalone();
    }
    
    boolean flatApperance = 
        Boolean.TRUE.equals(context.getArguments().get(InplaceEditorArgumentKeys.PROPERTY_TRANSPARENT));
    
    boolean eclipseOnMac = PlatformDetector.isMacOS() && !isSA;
    // Get the 'drawBorder' property for the button.
    if (eclipseOnMac) {
      // On Mac the SWT tool item is not focusable.
      flatApperance = false;
    }
    
    // Pe L&F Windows classic, tot butonul devine transparent. 
    boolean opaque = !flatApperance;
    if (PlatformDetector.isMacOS()) {
      opaque = false;
    }
    button.setOpaque(opaque);
    
    //EXM-29114 EXM-29209 Use the default border if the button is not transparent
    // as using an empty border may cause the button to become transparent on 
    // different transform.
    Border border = defaultBorder;
    if (flatApperance) {
      if (PlatformDetector.isWin32()) {
        if (button.getIcon() == null) {
          border = ApplicationBorderFactory.createEmptyBorder(3, 4, 3, 4);
        } else {
          border = ApplicationBorderFactory.createEmptyBorder(1, 2, 1, 2);
        }
      } else if (PlatformDetector.isMacOS()) {
        border = ApplicationBorderFactory.createEmptyBorder(1, 1, 1, 1);
      }
      
      button.setFocusPainted(false);
      button.setContentAreaFilled(false);
      if (!PlatformDetector.isMacOS()) {
        button.setBorderPainted(false);
      }
    } else {
      button.setFocusPainted(true);
      button.setContentAreaFilled(true);
      if (!PlatformDetector.isMacOS()) {
        button.setBorderPainted(true);
      }
    }
    button.setBorder(border);
  }
  
  /**
   * Initializes the button for the given context. 
   */
  void prepare(AuthorInplaceContext context) {
    Font font = (Font) context.getArguments().get(InplaceEditorArgumentKeys.FONT);
    if (font != null) {
      button.setFont(new java.awt.Font(font.getName(), font.getStyle(), font.getSize()));
    } else {
      button.setFont(defaultFont);
    }
    
    ro.sync.exml.view.graphics.Color color =
        (ro.sync.exml.view.graphics.Color) context.getArguments().get(InplaceEditorArgumentKeys.PROPERTY_COLOR);
    if (color != null) {
      button.setForeground(new Color(color.getRed(), color.getGreen(), color.getBlue(), color.getAlpha()));
    } else {
      button.setForeground(defaultForegroundColor);
    }
    
    final Object actionID = context.getArguments().get(InplaceEditorArgumentKeys.PROPERTY_ACTION_ID);
    IAuthorExtensionAction act = null;
    if (actionID instanceof IAuthorExtensionAction) {
      act = (IAuthorExtensionAction) actionID;
    }
    
    final IAuthorExtensionAction extAction = act;
    if (actionListener != null) {
      button.removeActionListener(actionListener);
    }
    
    if (keyListener != null) {
      button.removeKeyListener(keyListener);
    }

    if (extAction != null) {
      // By using the actionCore we could use this code even in a SWT environment.

      String actionDescription = (String) extAction.getValue(IAuthorExtensionAction.DESCRIPTION);

      // Deals with the icon.
      Icon icon = null;
      Boolean showIcon = getBoolean(context, InplaceEditorArgumentKeys.PROPERTY_SHOW_ICON);
      if (showIcon == null || showIcon == true) {
        Icon largeIcon = ImageUtil.loadToolbarIcon(
            (String) extAction.getValue(IAuthorExtensionAction.LARGE_ICON_PATH)); 
        Icon smallIcon = ImageUtil.loadMenuIcon(
            (String) extAction.getValue(IAuthorExtensionAction.SMALL_ICON_PATH), false);
        
        icon = smallIcon != null ? smallIcon : largeIcon;
      } 
      button.setIcon(icon);
      
      // Deals with the text.
      Boolean showText = getBoolean(context, InplaceEditorArgumentKeys.PROPERTY_SHOW_TEXT);
      String text = null;      
      if (icon == null || Boolean.TRUE.equals(showText)) {
        text = (String) extAction.getValue(IAuthorExtensionAction.ACTION_NAME);
      }
      button.setText(text);
      
      button.setToolTipText(actionDescription);
      actionListener = new ActionListener() {
        @Override
        public void actionPerformed(ActionEvent e) {
          // Request the invocation of the action. The accepted approach is to let
          // the helper do all the committing instead of just invoking the action
          // ourselves.
          // Fire an editing occurred. Otherwise the edited value will not be committed.
          fireEditingStopped(new EditingEvent(extAction));
        }
      };

    } else {
      button.setIcon(null);
      button.setText(messages.getString(Tags.UNKNOWN));
      button.setToolTipText("Unknown action");
      actionListener = new ActionListener() {
        @Override
        public void actionPerformed(ActionEvent e) {
          IDEAccess.getInstance().showErrorMessage(
              "Unknown action ID: " + String.valueOf(actionID) +
          		". The document must have an associated Document Type in which an action with the given ID must be declared.");
         fireEditingStopped(new EditingEvent((String) null));
        }
      };
    }
    button.addActionListener(actionListener);
    
    keyListener = new KeyAdapter() {
      @Override
      public void keyPressed(KeyEvent e) {
        if (e.getKeyCode() == KeyEvent.VK_ESCAPE) {
          e.consume();
          fireEditingCanceled();
        } else if (e.getKeyCode() == KeyEvent.VK_ENTER) {
          e.consume();
          // Fire an editing occurred. Otherwise the edited value will not be committed.
          fireEditingStopped(new EditingEvent(extAction));
        }
      }
    };
    button.addKeyListener(keyListener);
    
    decorateJButton(button, context, defaultBorder);
  }

  /**
   * @see ro.sync.ecss.extensions.api.editor.InplaceRenderer#getRendererComponent(ro.sync.ecss.extensions.api.editor.AuthorInplaceContext)
   */
  @Override
  public Object getRendererComponent(AuthorInplaceContext context) {
    prepare(context);
    
    return button;
  }

  /**
   * Gets rendering info for a button that solves all platform-dependence issues.
   * 
   * @param button the button to be rendered.
   * @param context the context in which we render.
   * @return the render layout information.
   */
  static RendererLayoutInfo getButtonRenderingInfo(AbstractButton button, 
      AuthorInplaceContext context) {
    // Obtain the size from the current context.
    boolean isSA = true;
    if (context.getAuthorAccess() != null 
        && context.getAuthorAccess().getWorkspaceAccess() != null) {
      isSA = context.getAuthorAccess().getWorkspaceAccess().isStandalone();
    }
    
    int correction = 0;
    if (!isSA && PlatformDetector.isMacOS() && button.getIcon() != null) {
      // When using the renderer for Eclipse, MAC OS with just an icon,
      // the SWING button is smaller than the SWT and when imposing the size to 
      // the SWT one the SWT button looks bad.
      correction = 5;
    }
    
    // Obtain the baseline
    final java.awt.Dimension preferredSize = button.getPreferredSize();
    
    return new RendererLayoutInfo(
        button.getBaseline(preferredSize.width, preferredSize.height),
        new Dimension(preferredSize.width + correction, preferredSize.height + correction));
  }
  
  /**
   * @see ro.sync.ecss.extensions.api.editor.InplaceRenderer#getRenderingInfo(ro.sync.ecss.extensions.api.editor.AuthorInplaceContext)
   */
  @Override
  public RendererLayoutInfo getRenderingInfo(AuthorInplaceContext context) {
    prepare(context);
    return getButtonRenderingInfo(button, context);
  }

  /**
   * @see ro.sync.ecss.extensions.api.editor.InplaceEditor#getEditorComponent(ro.sync.ecss.extensions.api.editor.AuthorInplaceContext, ro.sync.exml.view.graphics.Rectangle, ro.sync.exml.view.graphics.Point)
   */
  @Override
  public Object getEditorComponent(AuthorInplaceContext context, Rectangle allocation, Point mouseLocation) {
    prepare(context);
    
    return button;
  }

  /**
   * @see ro.sync.ecss.extensions.api.editor.InplaceEditor#requestFocus()
   */
  @Override
  public void requestFocus() {
    button.requestFocus();
  }

  /**
   * @see ro.sync.ecss.extensions.api.editor.InplaceEditor#getValue()
   */
  @Override
  public Object getValue() {
    // We don't have a value. perhaps we could return the action to invoke...
    return null;
  }

  /**
   * @see ro.sync.ecss.extensions.api.editor.InplaceEditor#stopEditing()
   */
  @Override
  public void stopEditing() {
    // Normally there shouldn't be any need for this. The button automatically 
    // ends editing when clicked. If the button wasn't clicked we will not 
    // execute it's action.
    fireEditingStopped(new EditingEvent((String) null));
  }

  /**
   * @see ro.sync.ecss.extensions.api.editor.InplaceEditor#cancelEditing()
   */
  @Override
  public void cancelEditing() {
    // Normally it shouldn't be any need for this event. The button automatically 
    // ends editing when clicked so there is no cancel involved.
    fireEditingCanceled();
  }

  /**
   * @see ro.sync.ecss.extensions.api.Extension#getDescription()
   */
  @Override
  public String getDescription() {
    return "A button linked to an author extension action";
  }

  /**
   * @see ro.sync.ecss.extensions.api.editor.InplaceEditor#getScrollRectangle()
   */
  @Override
  public Rectangle getScrollRectangle() {
    return null;
  }
  
  /**
   * @see ro.sync.ecss.extensions.api.editor.InplaceRenderer#getTooltipText(ro.sync.ecss.extensions.api.editor.AuthorInplaceContext, int, int)
   */
  @Override
  public String getTooltipText(AuthorInplaceContext context, int x, int y) {
    prepare(context);
    return button.getToolTipText();
  }

  /**
   * @see ro.sync.ecss.extensions.api.editor.InplaceRenderer#getCursorType(ro.sync.ecss.extensions.api.editor.AuthorInplaceContext, int, int)
   */
  @Override
  public CursorType getCursorType(AuthorInplaceContext context, int x, int y) {
    return CursorType.CURSOR_NORMAL;
  }
  
  /**
   * @see ro.sync.ecss.extensions.api.editor.InplaceRenderer#getCursorType(int, int)
   */
  @Override
  public CursorType getCursorType(int x, int y) {
    return null;
  }
}