/*
 *  The Syncro Soft SRL License
 *
 *  Copyright (c) 1998-2012 Syncro Soft SRL, Romania.  All rights
 *  reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistribution of source or in binary form is allowed only with
 *  the prior written permission of Syncro Soft SRL.
 *
 *  2. Redistributions of source code must retain the above copyright
 *  notice, this list of conditions and the following disclaimer.
 *
 *  3. Redistributions in binary form must reproduce the above copyright
 *  notice, this list of conditions and the following disclaimer in
 *  the documentation and/or other materials provided with the
 *  distribution.
 *
 *  4. The end-user documentation included with the redistribution,
 *  if any, must include the following acknowledgment:
 *  "This product includes software developed by the
 *  Syncro Soft SRL (http://www.sync.ro/)."
 *  Alternately, this acknowledgment may appear in the software itself,
 *  if and wherever such third-party acknowledgments normally appear.
 *
 *  5. The names "Oxygen" and "Syncro Soft SRL" must
 *  not be used to endorse or promote products derived from this
 *  software without prior written permission. For written
 *  permission, please contact support@oxygenxml.com.
 *
 *  6. Products derived from this software may not be called "Oxygen",
 *  nor may "Oxygen" appear in their name, without prior written
 *  permission of the Syncro Soft SRL.
 *
 *  THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 *  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *  OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *  DISCLAIMED.  IN NO EVENT SHALL THE SYNCRO SOFT SRL OR
 *  ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 *  USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 *  ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 *  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 *  OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */
package ro.sync.ecss.component.editor;

import java.awt.Color;
import java.awt.Component;
import java.awt.Point;
import java.awt.event.FocusAdapter;
import java.awt.event.FocusEvent;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.awt.event.KeyAdapter;
import java.awt.event.KeyEvent;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import javax.swing.JLabel;
import javax.swing.JList;
import javax.swing.SwingUtilities;
import javax.swing.event.DocumentEvent;
import javax.swing.event.DocumentListener;
import javax.swing.text.BadLocationException;
import javax.swing.text.JTextComponent;

import org.apache.log4j.Logger;

import ro.sync.annotations.api.API;
import ro.sync.annotations.api.APIType;
import ro.sync.annotations.api.SourceType;
import ro.sync.contentcompletion.xml.CIValue;
import ro.sync.ecss.extensions.api.editor.AbstractInplaceEditor;
import ro.sync.ecss.extensions.api.editor.AuthorInplaceContext;
import ro.sync.ecss.extensions.api.editor.EditingEvent;
import ro.sync.ecss.extensions.api.editor.InplaceEditorArgumentKeys;
import ro.sync.ecss.extensions.commons.editor.InplaceEditorUtil;
import ro.sync.exml.editor.helpers.attributes.AttrValueComboBoxEditor;
import ro.sync.exml.view.graphics.Font;
import ro.sync.exml.view.graphics.Rectangle;
import ro.sync.ui.UiUtil;
import ro.sync.ui.application.ApplicationListCellRendererWrapper;
import ro.sync.ui.tooltip.MultiLineToolTip;

/**
 * Combo box value editor.
 */
@API(type=APIType.INTERNAL, src=SourceType.PUBLIC)
public class ComboBoxEditor extends AbstractInplaceEditor {
  
  /**
   * Logger for logging.
   */
  private static final Logger logger = Logger.getLogger(ComboBoxEditor.class.getName());
  
  /**
   * Flag that indicates that the combo is during lookup.
   */
  private boolean isSearching = false;
  /**
   * The editor combo box
   */
  private final AttrValueComboBoxEditor comboBox = new AttrValueComboBoxEditor(false, true) {
    @Override
    public javax.swing.JToolTip createToolTip() {
      MultiLineToolTip tooltip = new MultiLineToolTip();
      tooltip.setComponent(this);
      return tooltip;
    }
    
    @Override
    protected void selectItemWhenSearching(Object item) {
      isSearching = true;
      try {
        super.selectItemWhenSearching(item);
      } finally {
        isSearching = false;
      }
    }
  };
  /**
   * Flag that keeps track when the editing is stopped by request.
   */
  private boolean editingStopped;
  /**
   * A map to identify the label for each value.
   */
  private final HashMap<String, String> valuesToLabels = new HashMap<String, String>();
  /**
   * Focus listener that stops the edititng on focus lost.
   */
  private FocusAdapter focusListener;
  /**
   * Key listener to stop the editing.
   */
  private KeyAdapter keyListener;
  /**
   * Item listener to fire edit notifications and auto commit.
   */
  private ItemListener itemListener;
  /**
   * Document listener to fire edit notifications.
   */
  private DocumentListener documentListener;
  
  /**
   * Constructor.
   */
  public ComboBoxEditor() {
    documentListener = new DocumentListener() {
      @Override
      public void removeUpdate(DocumentEvent e) {
        fireEditingOccured();
      }
      @Override
      public void insertUpdate(DocumentEvent e) {
        fireEditingOccured();             
      }
      @Override
      public void changedUpdate(DocumentEvent e) {
        fireEditingOccured();             
      }
    };

    itemListener = new ItemListener() {
      @Override
      public void itemStateChanged(ItemEvent e) {
        fireEditingOccured();
        
        if (!isSearching && e.getStateChange() == ItemEvent.SELECTED) {
          Point location = comboBox.getLocation();
          fireCommitValue(new EditingEvent((String) getValue()));
          // On Metal if we reposition, the popUp remains at the old position.
          if (!location.equals(comboBox.getLocation())) {
            comboBox.hidePopup();
          }
        }
      }
    };
    // Add focus listener to commit the the selected value on focus lost
    focusListener = new FocusAdapter() {
      @Override
      public void focusLost(FocusEvent e) {
        if (logger.isDebugEnabled()) {
          logger.debug("Focus lost " + e.getOppositeComponent());
        }
        boolean descends = false;
        if (e.getOppositeComponent() != null) {
          descends = SwingUtilities.isDescendingFrom(e.getOppositeComponent(), comboBox);
        }
        if (!descends && !editingStopped) {
          // Commit but request focus only if there is no opposite component.
          stopEditing(e.getOppositeComponent() == null, false);
        }
      }
    };

    keyListener = new KeyAdapter() {
      @Override
      public void keyPressed(KeyEvent e) {
        if (!editingStopped) {
          if (e.getKeyCode() == KeyEvent.VK_ESCAPE && !comboBox.isPopupVisible()) {
            // An Escape while the popUp is not visible must cancel the edit.
            cancelEditing();
            e.consume();
          } else if (e.getKeyCode() == KeyEvent.VK_ENTER && !comboBox.isPopupVisible()) {
            // An ENTER commits the value.
            stopEditing(false, true);
            e.consume();
          }
        }
      }
    };
    
    comboBox.setRenderer(new ApplicationListCellRendererWrapper(comboBox.getRenderer()) {
      @Override
      public Component getListCellRendererComponent(JList list, Object value, int index,
          boolean isSelected, boolean cellHasFocus) {
        JLabel comp = (JLabel) super.getListCellRendererComponent(list, value, index, isSelected, cellHasFocus);
        if (value != null && !comboBox.isEditable()) {
          // If the combo is not edit-able we can render the labels instead of the 
          // real values.
          String label = valuesToLabels.get(((CIValue)value).getInsertString());
          if (label != null) {
            comp.setText(label);
          }
        }
        comp.setForeground(comboBox.getForeground());
        return comp;
      }
    });
  }
  
  /**
   * Prepare editor.
   * 
   * @param context The context where the editor will be used.
   * @param forEditing <code>true</code> if prepare editor for editing.
   */
  private void prepareEditor(AuthorInplaceContext context) {
    // Remove all listeners.
    comboBox.getEditor().getEditorComponent().removeFocusListener(focusListener);
    comboBox.getEditor().getEditorComponent().removeKeyListener(keyListener);
    comboBox.removeKeyListener(keyListener);
    comboBox.removeItemListener(itemListener);
    if (comboBox.isEditable()) {
      ((JTextComponent) comboBox.getEditor().getEditorComponent()).getDocument().removeDocumentListener(documentListener);
    }
    
    comboBox.enableLookup(false);
    try {
      comboBox.removeAllItems();
      // We reuse the editor so we must be sure it doesn't have a preferred size set.
      // Otherwise the combo won't compute it's preferred size. In JActionComboBox there
      // is a fix the imposes the preferred size so we must bypass it.
      comboBox.setPreferredSize(null);
      comboBox.getEditor().getEditorComponent().removeFocusListener(focusListener);
      comboBox.getEditor().getEditorComponent().removeKeyListener(keyListener);
      comboBox.removeKeyListener(keyListener);
      
      Font font = (Font) context.getArguments().get(InplaceEditorArgumentKeys.FONT);
      if (font != null) {
        comboBox.setFont(new java.awt.Font(font.getName(), font.getStyle(), font.getSize()));
      }
      
      ro.sync.exml.view.graphics.Color color =
          (ro.sync.exml.view.graphics.Color) context.getArguments().get(InplaceEditorArgumentKeys.PROPERTY_COLOR);
      if (color != null) {
        comboBox.setForeground(new Color(color.getRed(), color.getGreen(), color.getBlue(), color.getAlpha()));
      } 
      
      if (comboBox.isEditable()) {
        comboBox.getEditor().getEditorComponent().setForeground(comboBox.getForeground());
      }
      
      editingStopped = false;
      boolean editable = true;
      Object editableProp = context.getArguments().get(InplaceEditorArgumentKeys.PROPERTY_EDITABLE);
      if (editableProp != null) {
        editable = (Boolean) editableProp;
      }
      comboBox.setEditable(editable);
      
      comboBox.setDoubleBuffered(false);
      List<CIValue> values = (List<CIValue>) context.getArguments().get(InplaceEditorArgumentKeys.PROPERTY_VALUES);
      List<String> labels = (List<String>) context.getArguments().get(InplaceEditorArgumentKeys.PROPERTY_LABELS);
      String currentValue = (String) context.getArguments().get(InplaceEditorArgumentKeys.INITIAL_VALUE);
      CIValue currentCIValue = null;
      
      
      if (values != null) {
        valuesToLabels.clear();
        
        // Avoids duplicates.
        Set<CIValue> visited = new HashSet<CIValue>();
        for (int i = 0; i < values.size(); i++) {
          CIValue val = values.get(i);
          
          if (!visited.contains(val)) {
            comboBox.addItem(val);
            visited.add(val);
            if (labels != null && !labels.isEmpty()) {
              valuesToLabels.put(val.getInsertString(), labels.get(i));
            }
            if (currentCIValue == null && val.getInsertString().equals(currentValue)) {
              currentCIValue = val;
            }
          }
        }
        
        
      }

      // Reset selection. By default the first item is selected.
      comboBox.setSelectedItem(null);
      // Select the current value
      if (currentCIValue != null) {
        comboBox.setSelectedItem(currentCIValue);
      } else {
        comboBox.setSelectedItem(currentValue);
      }

      // Because the editor is reused, a focus lost event comes at an unexpected 
      // moment: after the previous edit was used and when the second edit has begun.
      // As a result it will stop the second edit. So we add the listeners later.
      // Much like EXM-10823.
      UiUtil.invokeLater(new Runnable() {
        @Override
        public void run() {
          comboBox.getEditor().getEditorComponent().addFocusListener(focusListener);
          comboBox.getEditor().getEditorComponent().addKeyListener(keyListener);
          comboBox.addKeyListener(keyListener);
          
          comboBox.addItemListener(itemListener);
          if (comboBox.isEditable()) {
            ((JTextComponent) comboBox.getEditor().getEditorComponent()).getDocument().addDocumentListener(
                documentListener);
          }
        }
      });

      // Reinstate the fix form JActionComboBox.
      comboBox.setPreferredSize(new java.awt.Dimension(comboBox.getPreferredSize()));

      // The popUp size computation takes into account the size of the component.
      // We must make sure it doesn't use the size of the previous edit event.
      comboBox.resetLastComputedPopupWidth();
      
      InplaceEditorUtil.relayout(comboBox, context);
    } finally {
      comboBox.enableLookup(true);
    }
  }
  
  /**
   * @see ro.sync.ecss.extensions.api.editor.InplaceEditor#getEditorComponent(ro.sync.ecss.extensions.api.editor.AuthorInplaceContext, ro.sync.exml.view.graphics.Rectangle, ro.sync.exml.view.graphics.Point)
   */
  @Override
  public Object getEditorComponent(
      AuthorInplaceContext context, Rectangle allocation, ro.sync.exml.view.graphics.Point mouseLocation) {
    prepareEditor(context);
    
    return comboBox;
  }

  /**
   * @return Returns the comboBox.
   */
  public AttrValueComboBoxEditor getComboBox() {
    return comboBox;
  }
  
  /**
   * @see ro.sync.ecss.extensions.api.editor.InplaceEditor#getValue()
   */
  @Override
  public Object getValue() {
    //Commit
    String val = null;
    Object item = null;
    if (comboBox.isEditable()) {
      item = comboBox.getEditor().getItem();
    }
    
    if (item == null) {
      item = comboBox.getSelectedItem();
    }
    
    if (item instanceof CIValue) {
      val = ((CIValue) item).getInsertString();
    } else if (item instanceof String) {
      val = (String) item;
    }
    
    return val;
  }
  
  /**
   * @see ro.sync.ecss.extensions.api.editor.InplaceEditor#requestFocus()
   */
  @Override
  public void requestFocus() {
    if (comboBox.isEditable()) {
      comboBox.getEditor().getEditorComponent().requestFocus();
    } else {
      comboBox.requestFocus();
    }
  }
  
  /**
   * @see ro.sync.ecss.extensions.api.editor.InplaceEditor#stopEditing()
   */
  @Override
  public void stopEditing() {
    stopEditing(true, false);
  }
  
  /**
   * Stops editing.
   * 
   * @param requestFocus <code>true</code> to request focus inside the author component.
   */
  private void stopEditing(boolean requestFocus, boolean onEnter) {
    editingStopped = true;
    
    if (onEnter) {
      fireNextEditLocationRequested();
    } else {
      fireEditingStopped(new EditingEvent((String) getValue(), requestFocus));
    }
  }
  
  /**
   * @see ro.sync.ecss.extensions.api.editor.InplaceEditor#cancelEditing()
   */
  @Override
  public void cancelEditing() {
    editingStopped = true;
    
    fireEditingCanceled();
  }

  /**
   * @see ro.sync.ecss.extensions.api.Extension#getDescription()
   */
  @Override
  public String getDescription() {
    return "A combo box editor";
  }

  /**
   * @see ro.sync.ecss.extensions.api.editor.InplaceEditor#getScrollRectangle()
   */
  @Override
  public Rectangle getScrollRectangle() {
    Rectangle r = null;
    if (comboBox.isEditable()) {
      JTextComponent editorComp = (JTextComponent) comboBox.getEditor().getEditorComponent();
      try {
        java.awt.Rectangle caretRect = editorComp.modelToView(editorComp.getCaretPosition());
        Point convertPoint = UiUtil.convertPoint(editorComp, new Point(caretRect.x, caretRect.y), comboBox);
        if (convertPoint != null) {
          r = new Rectangle(convertPoint.x, convertPoint.y, caretRect.width, caretRect.height);
        }
      } catch (BadLocationException e) {
        logger.error(e, e);
      }
    }
    return r;
  }
}