/*
 *  The Syncro Soft SRL License
 *
 *  Copyright (c) 1998-2012 Syncro Soft SRL, Romania.  All rights
 *  reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistribution of source or in binary form is allowed only with
 *  the prior written permission of Syncro Soft SRL.
 *
 *  2. Redistributions of source code must retain the above copyright
 *  notice, this list of conditions and the following disclaimer.
 *
 *  3. Redistributions in binary form must reproduce the above copyright
 *  notice, this list of conditions and the following disclaimer in
 *  the documentation and/or other materials provided with the
 *  distribution.
 *
 *  4. The end-user documentation included with the redistribution,
 *  if any, must include the following acknowledgment:
 *  "This product includes software developed by the
 *  Syncro Soft SRL (http://www.sync.ro/)."
 *  Alternately, this acknowledgment may appear in the software itself,
 *  if and wherever such third-party acknowledgments normally appear.
 *
 *  5. The names "Oxygen" and "Syncro Soft SRL" must
 *  not be used to endorse or promote products derived from this
 *  software without prior written permission. For written
 *  permission, please contact support@oxygenxml.com.
 *
 *  6. Products derived from this software may not be called "Oxygen",
 *  nor may "Oxygen" appear in their name, without prior written
 *  permission of the Syncro Soft SRL.
 *
 *  THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 *  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *  OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *  DISCLAIMED.  IN NO EVENT SHALL THE SYNCRO SOFT SRL OR
 *  ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 *  USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 *  ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 *  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 *  OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */
package ro.sync.ecss.component.editor;

import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.event.FocusAdapter;
import java.awt.event.FocusEvent;
import java.awt.event.KeyAdapter;
import java.awt.event.KeyEvent;
import java.sql.Date;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.List;

import javax.swing.event.DocumentEvent;
import javax.swing.event.DocumentListener;
import javax.xml.bind.DatatypeConverter;

import org.apache.log4j.Logger;

import ro.sync.annotations.api.API;
import ro.sync.annotations.api.APIType;
import ro.sync.annotations.api.SourceType;
import ro.sync.ecss.extensions.api.CursorType;
import ro.sync.ecss.extensions.api.editor.AbstractInplaceEditor;
import ro.sync.ecss.extensions.api.editor.AuthorInplaceContext;
import ro.sync.ecss.extensions.api.editor.EditingEvent;
import ro.sync.ecss.extensions.api.editor.InplaceEditorArgumentKeys;
import ro.sync.ecss.extensions.api.editor.InplaceRenderer;
import ro.sync.ecss.extensions.api.editor.RendererLayoutInfo;
import ro.sync.ecss.extensions.commons.editor.InplaceEditorUtil;
import ro.sync.exml.view.graphics.Font;
import ro.sync.exml.view.graphics.Point;
import ro.sync.exml.view.graphics.Rectangle;
import ro.sync.ui.ListFocusTraversalPolicy;
import ro.sync.ui.calendar.DateEditor;
import ro.sync.ui.calendar.JDateChooser;

/**
 * Date picker form control.
 * 
 * @author alex_jitianu
 * @author costi
 */
@API(type=APIType.INTERNAL, src=SourceType.PUBLIC)
public class DatePickerEditor extends AbstractInplaceEditor implements InplaceRenderer, LoggableInplaceRenderer {
  /**
   * Logger for logging.
   */
  private static final Logger logger = Logger.getLogger(DatePickerEditor.class.getName());
  /**
   * The type of format that must be used to serialize the committed date. 
   * One of FORMAT_TYPE_ constants.
   */
  private String formatType = DatePickerEditorUtil.DATE_FORMAT_TYPE_CUSTOM;
  /**
   * A JAVA date format that was specified inside the CSS.
   */
  private String customFormat;
  /**
   * Date chooser used to present and edit a date.
   */
  private JDateChooser dateChooser;
  /**
   * The text field that presents the date inside the date chooser.
   */
  private DateEditor dateEditor;
  /**
   * The editor default font.
   */
  private java.awt.Font defaultFont;
  
  /**
   * The error message for wrong date/time format.
   */
  private String errorMessage;
  
  /**
   * Gets notified when the date has been edited so that we may notify the listeners. 
   */
  private DocumentListener documentListener = new DocumentListener() {
    @Override
    public void removeUpdate(DocumentEvent e) {
      fireEditingOccured();
    }
    @Override
    public void insertUpdate(DocumentEvent e) {
      fireEditingOccured();
    }
    @Override
    public void changedUpdate(DocumentEvent e) {
      fireEditingOccured();
    }
  };
  private KeyAdapter keyListener;
  /**
   * The color to render a valid date.
   */
  private Color defaultValidDateColor;
  /**
   * <code>true</code> if the date input should be validated.
   */
  private Boolean validateInput;


  /**
   * Constructor.
   */
  public DatePickerEditor() {
    this.dateEditor = new DateEditor(null);
    dateChooser = new JDateChooser(null, null, null, dateEditor, null);
    
    // The text field is the only focusable component.
    List<Component> comps = new ArrayList<Component>(1);
    comps.add(dateEditor);
    dateChooser.setFocusTraversalPolicy(new ListFocusTraversalPolicy(
        dateChooser, 
        comps, 
        dateEditor, 
        false));
    dateChooser.setFocusTraversalPolicyProvider(true);
    
    keyListener = new KeyAdapter() {
      @Override
      public void keyPressed(KeyEvent e) {
        if (e.getKeyCode() == KeyEvent.VK_ESCAPE) {
          e.consume();
          cancelEditing();
        } else if (e.getKeyCode() == KeyEvent.VK_ENTER) {
          e.consume();
          stopEditing(false, true);
        }
      }
    };
    
    defaultFont = dateChooser.getFont();
    defaultValidDateColor = dateEditor.getValidColor();
  }

  /**
   * @see ro.sync.ecss.extensions.api.Extension#getDescription()
   */
  @Override
  public String getDescription() {
    return "A date picker.";
  }

  /**
   * @see ro.sync.ecss.extensions.api.editor.InplaceEditor#getEditorComponent(ro.sync.ecss.extensions.api.editor.AuthorInplaceContext, ro.sync.exml.view.graphics.Rectangle, ro.sync.exml.view.graphics.Point)
   */
  @Override
  public Object getEditorComponent(AuthorInplaceContext context, Rectangle allocation,
      Point mouseInvocationLocation) {
    prepare(context, true);
    return dateChooser;
  }

  /**
   * @see ro.sync.ecss.extensions.api.editor.InplaceEditor#getScrollRectangle()
   */
  @Override
  public Rectangle getScrollRectangle() {
    return null;
  }

  /**
   * @see ro.sync.ecss.extensions.api.editor.InplaceEditor#requestFocus()
   */
  @Override
  public void requestFocus() {
    dateEditor.requestFocus();
  }

  /**
   * @see ro.sync.ecss.extensions.api.editor.InplaceEditor#getValue()
   */
  @Override
  public Object getValue() {
    String value = "";
    Date date = dateChooser.getDate();
    if (date != null) {
      if (formatType == DatePickerEditorUtil.DATE_FORMAT_TYPE_XS_DATE) {
        // Date should be converted as specified by the schema (xs:date).
        Calendar cal = Calendar.getInstance();
        cal.setTime(date);
        value = DatatypeConverter.printDate(cal);
      } else if (formatType == DatePickerEditorUtil.DATE_FORMAT_TYPE_XS_DATE_TIME) {
        Calendar cal = Calendar.getInstance();
        cal.setTime(date);
        value = DatatypeConverter.printDateTime(cal);
      } else {
        // In schema it is not specified as a xs:date; convert it using the CSS format.
        java.util.Date mydate = new java.util.Date(date.getTime());
        try {
          value = new SimpleDateFormat(customFormat).format(mydate);
        } catch (IllegalArgumentException e) {
          // The CSS format is not a valid Java date time format.
          if (logger.isDebugEnabled()) {
            logger.debug(e, e);
          }
        }
      }
    }
    return value;
  }

  private void stopEditing(boolean requestFocus, boolean onEnter) {
    dateChooser.hidePopup();
    dateChooser.removeKeyListener(keyListener);
    dateEditor.getDocument().removeDocumentListener(documentListener);
    
    if (onEnter) {
      fireNextEditLocationRequested();
    } else {
      fireEditingStopped(new EditingEvent((String)getValue(), requestFocus));
    }
  }
  
  /**
   * @see ro.sync.ecss.extensions.api.editor.InplaceEditor#stopEditing()
   */
  @Override
  public void stopEditing() {
    stopEditing(true, false);
  }

  /**
   * @see ro.sync.ecss.extensions.api.editor.InplaceEditor#cancelEditing()
   */
  @Override
  public void cancelEditing() {
    dateChooser.hidePopup();
    dateChooser.removeKeyListener(keyListener);
    dateEditor.getDocument().removeDocumentListener(documentListener);
    fireEditingCanceled();
  }

  /**
   * @see ro.sync.ecss.extensions.api.editor.InplaceRenderer#getRendererComponent(ro.sync.ecss.extensions.api.editor.AuthorInplaceContext)
   */
  @Override
  public Object getRendererComponent(AuthorInplaceContext context) {
    prepare(context, false);
    
    return dateChooser;
  }

  /**
   * @see ro.sync.ecss.extensions.api.editor.InplaceRenderer#getCursorType(ro.sync.ecss.extensions.api.editor.AuthorInplaceContext, int, int)
   */
  @Override
  public CursorType getCursorType(AuthorInplaceContext context, int x, int y) {
    return CursorType.CURSOR_NORMAL;
  }

  /**
   * @see ro.sync.ecss.extensions.api.editor.InplaceRenderer#getTooltipText(ro.sync.ecss.extensions.api.editor.AuthorInplaceContext, int, int)
   */
  @Override
  public String getTooltipText(AuthorInplaceContext context, int x, int y) {
    prepare(context, false);
    if (errorMessage != null) {
      return errorMessage;
    }
    return dateEditor.getText();
  }

  /**
   * @see ro.sync.ecss.extensions.api.editor.InplaceRenderer#getRenderingInfo(ro.sync.ecss.extensions.api.editor.AuthorInplaceContext)
   */
  @Override
  public RendererLayoutInfo getRenderingInfo(AuthorInplaceContext context) {
    prepare(context, false);
    
    final Dimension dateEditorPreferredSize = dateEditor.getPreferredSize();
    return new RendererLayoutInfo(
        dateEditor.getBaseline(dateEditorPreferredSize.width, dateEditorPreferredSize.height),
        InplaceEditorUtil.getPreferredSize(dateChooser, context));
  }
  
  /**
   * Prepare the editor for usage.
   * 
   * @param context Context information. 
   * @param forEditing <code>true</code> if we prepare for editing. <code>false</code>
   * if we prepare for rendering a value.
   */
  private void prepare(AuthorInplaceContext context, boolean forEditing) {
    errorMessage = null;
    Font font = (Font) context.getArguments().get(InplaceEditorArgumentKeys.FONT);
    if (font != null) {
      dateChooser.setFont(new java.awt.Font(font.getName(), font.getStyle(), font.getSize()));
    } else {
      dateChooser.setFont(defaultFont);
    }
    
    ro.sync.exml.view.graphics.Color color =
        (ro.sync.exml.view.graphics.Color) context.getArguments().get(InplaceEditorArgumentKeys.PROPERTY_COLOR);
    if (color != null) {
      dateEditor.setValidColor(new Color(color.getRed(), color.getGreen(), color.getBlue(), color.getAlpha()));
    } else {
      dateEditor.setValidColor(defaultValidDateColor);
    }

    dateChooser.setOpaque(false);
    
    // EXM-27418: Check input only if the 'validateInput' property is set.
    String validate = (String) context.getArguments().get(InplaceEditorArgumentKeys.PROPERTY_VALIDATE_INPUT);
    validateInput = validate == null ? null : Boolean.valueOf(validate);
    if (validateInput == null || validateInput.booleanValue()) {
      dateEditor.setValidateInput(true);
    } else {
      dateEditor.setValidateInput(false);
    }
    
    formatType = DatePickerEditorUtil.getFormatType(context);
    
    customFormat = (String) context.getArguments().get(InplaceEditorArgumentKeys.PROPERTY_FORMAT);
    String datePattern = ((SimpleDateFormat) DateFormat.getDateInstance(DateFormat.MEDIUM, dateEditor.getLocale())).toPattern();
    if (customFormat == null) {
      customFormat = datePattern;
    }
    
    String initialValue = (String) context.getArguments().get(InplaceEditorArgumentKeys.INITIAL_VALUE);
    Long dateInMills = null;
    // Convert the initial value according to schema or format property value.
    if (formatType == DatePickerEditorUtil.DATE_FORMAT_TYPE_CUSTOM) {
      // Convert it according to the custom format value from CSS.
      if (initialValue != null) {
        try {
          java.util.Date date = new SimpleDateFormat(customFormat).parse(initialValue);
          dateInMills = date.getTime();
        } catch (Exception e) {
          // The CSS format is not a valid Java date time format.
          errorMessage = e.getMessage() + ". Expected \"" + customFormat + "\" format.";
          if (logger.isDebugEnabled()) {
            logger.debug(e, e);
          }
        }
      }
      dateEditor.setDateFormatString(customFormat);
    } else if (formatType == DatePickerEditorUtil.DATE_FORMAT_TYPE_XS_DATE) {
      // Convert it according to the schema date type.
      if (initialValue != null) {
        try {
          Calendar date = DatatypeConverter.parseDate(initialValue);
          dateInMills = date.getTimeInMillis();
        } catch (IllegalArgumentException e) {
          errorMessage = "Invalid date: " + e.getMessage() + ". Expected the xs:date format.";
        }
      }
      dateEditor.setDateFormatString(datePattern);
    } else if (formatType == DatePickerEditorUtil.DATE_FORMAT_TYPE_XS_DATE_TIME) {
      // Convert it according to the schema date time type.
      if (initialValue != null) {
        try {
          Calendar date = DatatypeConverter.parseDateTime(initialValue);
          dateInMills = date.getTimeInMillis();
        } catch (IllegalArgumentException e) {
          errorMessage = "Invalid date-time: " + e.getMessage() + ". Expected the xs:dateTime format.";
        }
      }

      String timePattern = ((SimpleDateFormat) DateFormat.getDateTimeInstance(DateFormat.MEDIUM, DateFormat.SHORT)).toPattern();
      dateEditor.setDateFormatString(timePattern);
    }
    if (dateInMills != null) {
      dateChooser.setDate(new Date(dateInMills));
    } else {
      dateChooser.setDate(null);
    }
    
    if (forEditing) {
      dateEditor.getDocument().addDocumentListener(documentListener);
      dateEditor.addKeyListener(keyListener);

      dateEditor.addFocusListener(new FocusAdapter() {
        @Override
        public void focusGained(FocusEvent e) {
          dateChooser.hidePopup();
        }
      });
    }
    
    if (errorMessage != null && initialValue != null) {
      dateEditor.setText(initialValue);
    }
    
    dateEditor.addFocusListener(new FocusAdapter() {
      @Override
      public void focusLost(FocusEvent e) {
        if (!dateChooser.isPopupVisible()) {
          // On a focus lost not triggered by the date chooser pop up, stop the editing.
          fireEditingStopped(new EditingEvent((String)getValue(), false));
        }
      }
    });
  }
  
  /**
   * @see ro.sync.ecss.extensions.api.editor.InplaceRenderer#getCursorType(int, int)
   */
  @Override
  public CursorType getCursorType(int x, int y) {
    return null;
  }

  /**
   * @see ro.sync.ecss.component.editor.LoggableInplaceRenderer#dump(ro.sync.ecss.extensions.api.editor.AuthorInplaceContext, java.lang.StringBuilder)
   */
  @Override
  public void dump(AuthorInplaceContext inplaceContext, StringBuilder builder) {
    prepare(inplaceContext, false);
    builder.append("value:'").append(dateEditor.getText()).append("'");
    builder.append("color:'").append(dateEditor.getForeground()).append("'");
  }
}