/*
 *  The Syncro Soft SRL License
 *
 *  Copyright (c) 1998-2012 Syncro Soft SRL, Romania.  All rights
 *  reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistribution of source or in binary form is allowed only with
 *  the prior written permission of Syncro Soft SRL.
 *
 *  2. Redistributions of source code must retain the above copyright
 *  notice, this list of conditions and the following disclaimer.
 *
 *  3. Redistributions in binary form must reproduce the above copyright
 *  notice, this list of conditions and the following disclaimer in
 *  the documentation and/or other materials provided with the
 *  distribution.
 *
 *  4. The end-user documentation included with the redistribution,
 *  if any, must include the following acknowledgment:
 *  "This product includes software developed by the
 *  Syncro Soft SRL (http://www.sync.ro/)."
 *  Alternately, this acknowledgment may appear in the software itself,
 *  if and wherever such third-party acknowledgments normally appear.
 *
 *  5. The names "Oxygen" and "Syncro Soft SRL" must
 *  not be used to endorse or promote products derived from this
 *  software without prior written permission. For written
 *  permission, please contact support@oxygenxml.com.
 *
 *  6. Products derived from this software may not be called "Oxygen",
 *  nor may "Oxygen" appear in their name, without prior written
 *  permission of the Syncro Soft SRL.
 *
 *  THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 *  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *  OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *  DISCLAIMED.  IN NO EVENT SHALL THE SYNCRO SOFT SRL OR
 *  ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 *  USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 *  ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 *  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 *  OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */
package ro.sync.ecss.component.editor;

import java.awt.Color;
import java.awt.Component;
import java.awt.event.ActionEvent;
import java.awt.event.FocusAdapter;
import java.awt.event.FocusEvent;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.awt.event.KeyAdapter;
import java.awt.event.KeyEvent;
import java.awt.event.MouseEvent;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;
import java.util.List;

import javax.swing.Action;
import javax.swing.JComboBox;
import javax.swing.SwingUtilities;
import javax.swing.event.DocumentEvent;
import javax.swing.event.DocumentListener;
import javax.swing.event.PopupMenuEvent;
import javax.swing.event.PopupMenuListener;
import javax.swing.text.BadLocationException;
import javax.swing.text.JTextComponent;

import org.apache.log4j.Logger;

import ro.sync.annotations.api.API;
import ro.sync.annotations.api.APIType;
import ro.sync.annotations.api.SourceType;
import ro.sync.ecss.extensions.api.AuthorAccess;
import ro.sync.ecss.extensions.api.CursorType;
import ro.sync.ecss.extensions.api.editor.AbstractInplaceEditor;
import ro.sync.ecss.extensions.api.editor.AuthorInplaceContext;
import ro.sync.ecss.extensions.api.editor.EditingEvent;
import ro.sync.ecss.extensions.api.editor.InplaceEditorArgumentKeys;
import ro.sync.ecss.extensions.api.editor.InplaceRenderer;
import ro.sync.ecss.extensions.api.editor.RendererLayoutInfo;
import ro.sync.ecss.extensions.commons.editor.InplaceEditorUtil;
import ro.sync.exml.EXMLResourceBoundle;
import ro.sync.exml.MessageBundle;
import ro.sync.exml.Tags;
import ro.sync.exml.view.graphics.Dimension;
import ro.sync.exml.view.graphics.Font;
import ro.sync.exml.view.graphics.Point;
import ro.sync.exml.view.graphics.Rectangle;
import ro.sync.ui.ActionWrapper;
import ro.sync.ui.BaseURLProvider;
import ro.sync.ui.InputUrlPanel;
import ro.sync.ui.ListFocusTraversalPolicy;
import ro.sync.ui.MenuButtonPanel;
import ro.sync.ui.UiUtil;
import ro.sync.ui.application.ApplicationComboBox;
import ro.sync.ui.toolbar.ToolBar;
import ro.sync.util.SimpleFileFilter;

import com.jidesoft.swing.DefaultSplitButtonModel;
import com.jidesoft.swing.JideSplitButton;

/**
 * Inplace editor used to select an URL. It uses the heavily customized InputURLPanel.
 * 
 * @author alex_jitianu
 * @author iulian_velea
 */
@API(type=APIType.INTERNAL, src=SourceType.PUBLIC)
public class InputURLEditor extends AbstractInplaceEditor implements InplaceRenderer, LoggableInplaceRenderer {
  /**
   * Logger for logging.
   */
  private static final Logger logger = Logger.getLogger(InputURLEditor.class.getName());
  
  /** 
   * The messages resource bundle. 
   */
  protected static final MessageBundle messages = EXMLResourceBoundle.getResourceBundleInstance();
  
  /**
   * <code>true</code> if the focus lost event should be ignored. <code>false</code>
   * to stop the editing when the focus is lost.
   */
  private boolean skipFocusLost;
  
  /**
   * We need to extend the InputURLPanel to be able to override the default behavior.
   * For example when the focus is lost from the editor there are some situations the event should be ignored.
   * 
   * @author alex_jitianu
   */
  private class RendererInputUrlPanel extends InputUrlPanel {
    
    /**
     * Internal Constructor.
     * 
     * @param filter The File Filter
     * @param initialComboContent content of the URL combo.
     * @param showEditorVariables <code>true</code> for showing the Editor Variables button. <code>false</code> otherwise.
     * @param persistenceKey The key that will be used to save entries from this URL panel as persistent between sessions.
     * Null to inhibit persistence.
     */
    public RendererInputUrlPanel(
        SimpleFileFilter filter, 
        String[] initialComboContent, 
        boolean showEditorVariables, 
        String persistenceKey) {
      super(filter, initialComboContent, showEditorVariables, persistenceKey);
      
      // It makes sure the focus is in the editor. It's useful when we start the 
      // editing session with a click on the popUp showing arrow then we close the 
      // popUp with a similar click or ESC.
      browseButtonGroup.getPopupMenu().addPopupMenuListener(new PopupMenuListener() {
        @Override
        public void popupMenuWillBecomeVisible(PopupMenuEvent e) {}
        @Override
        public void popupMenuWillBecomeInvisible(final PopupMenuEvent e) {
          // EXM-27338 The popUp is hidden first thing. We do the focus check 
          // later so that is a dialog was presented we wont do anything.
          UiUtil.invokeLater(new Runnable() {
            @Override
            public void run() {
              if (!skipFocusLost) {
                getUrlCombo().requestFocus();
              }
            }
          });
        }
        @Override
        public void popupMenuCanceled(PopupMenuEvent e) {
          popupMenuWillBecomeInvisible(e);
        }
      });
      
      final JTextComponent editor = (JTextComponent) getUrlCombo().getEditor().getEditorComponent();
      setFocusTraversalPolicyProvider(true);
      
      List<Component> comps = new ArrayList<Component>();
      comps.add(editor);
      comps.add(browseButtonGroup);
      setFocusTraversalPolicy(new ListFocusTraversalPolicy(
          this, 
          comps, 
          comps.isEmpty() ? null : comps.get(0), 
          false));
    }
    
    /**
     * @return The button with all the browse actions.
     */
    private MenuButtonPanel getBrowseButtonGroup() {
      return browseButtonGroup;
    }
    
    /**
     * @see ro.sync.ui.InputUrlPanel#customizeContributedActions(java.util.List)
     */
    @Override
    protected void customizeContributedActions(List<Action> browseActionsList) {
      super.customizeContributedActions(browseActionsList);
      
      List<Action> originals = new ArrayList<Action>(browseActionsList);
      browseActionsList.clear();
      for (Action action : originals) {
        browseActionsList.add(new ActionWrapper(action) {
          @Override
          public void actionPerformed(ActionEvent e) {
            try {
              // The browse action will present different dialogs so we must not stop editing on focus lost.
              skipFocusLost = true;
              super.actionPerformed(e);
            } finally {
              skipFocusLost = false;
              getUrlCombo().requestFocus();
            }
          }
        });
      }
    }
    
    @Override
    protected String preprocessUserSelection(URL url) {
      return makeRelativeURL(url);
    }
  }
  
  /**
   * The panel holding the URL input controls.
   */
  private RendererInputUrlPanel urlPanel = new RendererInputUrlPanel(null, null, false, null);

  /**
   * The listener notified about modifications in the URL editor.
   */
  private DocumentListener editingOccuredListener;

  /**
   * <code>true</code> when the editing is initiated bya  mouse event and there is no need to artificially request the focus 
   * in the URL combo.
   */
  private boolean skipFocusRequest;
  
  /**
   * Default foreground color of the text field.
   */
  private final Color defaultForeground;

  /**
   * The default font.
   */
  private final java.awt.Font defaultFont;
  /**
   * Handles the automatic commit when the user selects a new entry from the pop-up.
   */
  private ItemListener itemListener;
  /**
   * Utility access taken from the context.
   */
  private AuthorAccess authorAccess;
  
  /**
   * Constructor.
   */
  public InputURLEditor() {
    urlPanel.setOpaque(false);
    // We can add a label from CSS.
    urlPanel.setMessageLabelText(null);
    editingOccuredListener = new DocumentListener() {
      @Override
      public void insertUpdate(DocumentEvent e) {
        fireEditingOccured();
      }
      @Override
      public void removeUpdate(DocumentEvent e) {
        fireEditingOccured();
      }
      @Override
      public void changedUpdate(DocumentEvent e) {
        fireEditingOccured();
      }
    };
    
    itemListener = new ItemListener() {
      @Override
      public void itemStateChanged(ItemEvent e) {
        fireEditingOccured();
        
        if (e.getStateChange() == ItemEvent.SELECTED) {
          JComboBox comboBox = urlPanel.getUrlCombo();
          java.awt.Point location = comboBox.getLocation();
          fireCommitValue(new EditingEvent((String) getValue()));
          // On Metal if we reposition, the popUp remains at the old position.
          if (!location.equals(comboBox.getLocation())) {
            comboBox.hidePopup();
          }
        }
      }
    };
    
    defaultForeground = urlPanel.getUrlCombo().getForeground();
    defaultFont = urlPanel.getUrlCombo().getFont();
  }
  
  /**
   * Initialize the editor for the current context.
   * 
   * @param context Current context.
   * @param forEditing <code>true</code> if we are preparing for editing. <code>false</code>
   * if we are preparing for rendering.
   */
  private void prepare(final AuthorInplaceContext context, boolean forEditing) {
    // Reset the models so that the UI marks the last size as dirty. Otherwise
    // we will get an old size from cache.
    urlPanel.getUrlCombo().removeAllItems();
    urlPanel.setSelectedURL("");
    String text = (String) context.getArguments().get(InplaceEditorArgumentKeys.INITIAL_VALUE);
    ro.sync.exml.view.graphics.Color color =
        (ro.sync.exml.view.graphics.Color) context.getArguments().get(InplaceEditorArgumentKeys.PROPERTY_COLOR);
    Component editorComponent = urlPanel.getUrlCombo().getEditor().getEditorComponent();
    if (color != null) {
      editorComponent.setForeground(new Color(color.getRed(), color.getGreen(), color.getBlue(), color.getAlpha()));
    } else {
      urlPanel.getUrlCombo().getEditor().getEditorComponent().setForeground(defaultForeground);
    }
    
    if (!forEditing && text == null) {
      editorComponent.setForeground(Color.GRAY);
      text = (String) context.getArguments().get(InplaceEditorArgumentKeys.DEFAULT_VALUE);
    }
    
    if (text == null) {
      text = forEditing ? "" : " ";
    }
    
    urlPanel.setBaseURLProvider(new BaseURLProvider() {
      @Override
      public URL getBaseURL() {
        return context.getAuthorAccess().getEditorAccess().getEditorLocation();
      }
    });
    
    Font font = (Font) context.getArguments().get(InplaceEditorArgumentKeys.FONT);
    if (font != null) {
      java.awt.Font currentFont = new java.awt.Font(font.getName(), font.getStyle(), font.getSize());
      urlPanel.getUrlCombo().setFont(currentFont);
    } else {
      urlPanel.getUrlCombo().setFont(defaultFont);
    }
    
    urlPanel.setSelectedURL(text);
    
    if (forEditing) {
      urlPanel.updateOpenedEditorsList();
    }
    
    // There are some fixes in our DropAcceptingHiddenPasswordComboBox and super classes
    // that interfere with the size policy of the form control.
    // We need to ask the combo about a preferred dimension so before that we 
    // must reset all fixes.
    urlPanel.getUrlCombo().setPreferredSize(null);
    urlPanel.getUrlCombo().setMinimumSize(null);
    // As stated in HiddenPasswordComboBox : On java 1.5 the combos are resized when a long text is inserted.
    // So we impose a width of 300.
    java.awt.Dimension preferredSize = new java.awt.Dimension(300, urlPanel.getUrlCombo().getPreferredSize().height);
    
    urlPanel.getUrlCombo().setPreferredSize(preferredSize);
    urlPanel.getUrlCombo().setMinimumSize(preferredSize);
    
    Dimension size = InplaceEditorUtil.getPreferredSize(urlPanel, context);
    // Unless the text field has a size it will not be able to respond to certain
    // calls like modelToView().
    urlPanel.setSize(size.width, size.height);
    urlPanel.doLayout();
    
    JComboBox comboBox = urlPanel.getUrlCombo();
    // Unless the text field has a size it will not be able to respond to certain
    // calls like modelToView().
    comboBox.doLayout();
    InplaceEditorUtil.setCaretAtEnd((JTextComponent) comboBox.getEditor().getEditorComponent(), context);
    
    if (forEditing) {
      // The editor is not reused so we don't have to bother with clean up.
      final JTextComponent editor = (JTextComponent) urlPanel.getUrlCombo().getEditor().getEditorComponent();
      final FocusAdapter focusListener = new FocusAdapter() {
        @Override
        public void focusLost(FocusEvent e) {
          if (!urlPanel.getBrowseButtonGroup().isPopupMenuVisible() 
              && !skipFocusLost 
              && (
                  e.getOppositeComponent() == null 
                  || !SwingUtilities.isDescendingFrom(e.getOppositeComponent(), urlPanel))) {
            // The focus has left the panel and is not generated by any browse action.
            if (logger.isDebugEnabled()) {
              logger.debug("Focus lost, stop editing. Opp comp: " + e.getOppositeComponent());
            }
            
            fireEditingStopped(new EditingEvent((String) getValue(), false));
          } 
        }
      };
      editor.addFocusListener(focusListener);
      
      editor.getDocument().addDocumentListener(editingOccuredListener);
      
      urlPanel.getUrlCombo().addItemListener(itemListener);
      
      editor.addKeyListener(new KeyAdapter() {
        @Override
        public void keyPressed(KeyEvent e) {
          if (!urlPanel.getUrlCombo().isPopupVisible()) {
            if (e.getKeyCode() == KeyEvent.VK_ESCAPE) {
              // ESC must cancel the edit.
              e.consume();
              // A focus lost event will come so must not process the event.
              editor.removeFocusListener(focusListener);
              fireEditingCanceled();
            } else if (e.getKeyCode() == KeyEvent.VK_ENTER) {
              // An ENTER commits the value.
              e.consume();
              // A focus lost event will come so must not process the event.
              editor.removeFocusListener(focusListener);
              // The editing occurred event is on a timer so 
              fireEditingOccured();
              fireNextEditLocationRequested();
            }
          }
        }
      });
    }
    
    if (!forEditing) {
      // Make sure the combo has no highlight.
      MouseEvent me = new MouseEvent(
          comboBox, 
          MouseEvent.MOUSE_EXITED, 
          System.currentTimeMillis(), 
          0, 
          0, 0, 
          1, 
          false);
      comboBox.dispatchEvent(me);
      
      // make sure the button has no highlight effect on it.
      JideSplitButton browseButtonGroup = urlPanel.getBrowseButtonGroup();
      DefaultSplitButtonModel model = (DefaultSplitButtonModel) browseButtonGroup.getModel();
      model.setButtonRollover(false);
      model.setButtonSelected(false);
    
      Point relPos = context.getRelativeMouseLocation();
      if (relPos != null) {
        int x = relPos.x;
        int y = relPos.y;
        Component componentAt = urlPanel.getComponentAt(x, y);
        if (componentAt instanceof JComboBox) {
          componentAt = comboBox.getComponentAt(x - componentAt.getX(), y - componentAt.getY());
          if (componentAt != null) {
            // The component under the mouse:
            // - for a editable combo can be either the button or the text field.
            // - for a read-only combo it can only be the combo.
            // Either way, the rollover effect is performed on a mouse entered event
            // in com.sun.java.swing.plaf.windows.WindowsComboBoxUI.
            me = new MouseEvent(
                componentAt, 
                MouseEvent.MOUSE_ENTERED, 
                System.currentTimeMillis(), 
                0, 
                relPos.x - componentAt.getX(), relPos.y - componentAt.getY(), 
                1, 
                false);
            componentAt.dispatchEvent(me);
          }
        } else if (componentAt instanceof ToolBar) {
          model.setButtonRollover(true);
          model.setButtonSelected(true);
        }
      }
    }
  }

  /**
   * @see ro.sync.ecss.extensions.api.editor.InplaceEditor#getEditorComponent(ro.sync.ecss.extensions.api.editor.AuthorInplaceContext, ro.sync.exml.view.graphics.Rectangle, ro.sync.exml.view.graphics.Point)
   */
  @Override
  public Object getEditorComponent(AuthorInplaceContext context, Rectangle allocation,
      Point mouseInvocationLocation) {
    prepare(context, true);
    
    authorAccess = context.getAuthorAccess();
    
    // We will let the mouse event set the focus in the right control.
    skipFocusRequest = mouseInvocationLocation != null;
    return urlPanel;
  }

  /**
   * @see ro.sync.ecss.extensions.api.editor.InplaceEditor#getScrollRectangle()
   */
  @Override
  public Rectangle getScrollRectangle() {
    Rectangle r = null;
    JComboBox comboBox = urlPanel.getUrlCombo();
    if (comboBox.isEditable()) {
      JTextComponent editorComp = (JTextComponent) comboBox.getEditor().getEditorComponent();
      try {
        java.awt.Rectangle caretRect = editorComp.modelToView(editorComp.getCaretPosition());
        if (caretRect != null) {
          java.awt.Point convertPoint = UiUtil.convertPoint(editorComp, new java.awt.Point(caretRect.x, caretRect.y), comboBox);
          if (convertPoint != null) {
            r = new Rectangle(convertPoint.x, convertPoint.y, caretRect.width, caretRect.height);
          }
        }
      } catch (BadLocationException e) {
        logger.error(e, e);
      }
    }
    
    return r;
  }

  /**
   * @see ro.sync.ecss.extensions.api.editor.InplaceEditor#requestFocus()
   */
  @Override
  public void requestFocus() {
    if (!skipFocusRequest) {
      urlPanel.getUrlCombo().requestFocus();
    }
  }

  /**
   * @see ro.sync.ecss.extensions.api.editor.InplaceEditor#getValue()
   */
  @Override
  public Object getValue() {
    String value = null;
    try {
      URL selectedURL = urlPanel.getSelectedURL();
      value = makeRelativeURL(selectedURL);
    } catch (MalformedURLException e) {
      value = urlPanel.getNotExpandedSelectedURL();
    }
    
    if (value == null) {
      // An empty value will remove the attribute.
      value = "";
    }
    
    return value;
  }

  /**
   * Make the given URL relative to the base URL.
   * 
   * @param selectedURL URL to make relative.
   * 
   * @return The relative URL.
   */
  private String makeRelativeURL(URL selectedURL) {
    String value = null;
    if (selectedURL != null) {
      selectedURL = authorAccess.getUtilAccess().removeUserCredentials(selectedURL);
      value = authorAccess.getUtilAccess().makeRelative(authorAccess.getEditorAccess().getEditorLocation(), selectedURL);
    }
    return value;
  }

  /**
   * @see ro.sync.ecss.extensions.api.editor.InplaceEditor#stopEditing()
   */
  @Override
  public void stopEditing() {
    fireEditingStopped(new EditingEvent((String) getValue(), true));
  }

  /**
   * @see ro.sync.ecss.extensions.api.editor.InplaceEditor#cancelEditing()
   */
  @Override
  public void cancelEditing() {
    fireEditingCanceled();
  }

  /**
   * @see ro.sync.ecss.extensions.api.Extension#getDescription()
   */
  @Override
  public String getDescription() {
    return "An input URL panel form control.";
  }

  /**
   * @see ro.sync.ecss.extensions.api.editor.InplaceRenderer#getRendererComponent(ro.sync.ecss.extensions.api.editor.AuthorInplaceContext)
   */
  @Override
  public Object getRendererComponent(AuthorInplaceContext context) {
    prepare(context, false);
    
    return urlPanel;
  }

  /**
   * @see ro.sync.ecss.extensions.api.editor.InplaceRenderer#getCursorType(ro.sync.ecss.extensions.api.editor.AuthorInplaceContext, int, int)
   */
  @Override
  public CursorType getCursorType(AuthorInplaceContext context, int x, int y) {
    return null;
  }
  
  /**
   * @see ro.sync.ecss.extensions.api.editor.InplaceRenderer#getCursorType(int, int)
   */
  @Override
  public CursorType getCursorType(int x, int y) {
    return null;
  }

  /**
   * @see ro.sync.ecss.extensions.api.editor.InplaceRenderer#getTooltipText(ro.sync.ecss.extensions.api.editor.AuthorInplaceContext, int, int)
   */
  @Override
  public String getTooltipText(AuthorInplaceContext context, int x, int y) {
    prepare(context, false);
    
    String tooltip = messages.getString(Tags.BROWSE);
    Component componentAt = urlPanel.getComponentAt(x, y);
    if (componentAt instanceof JComboBox) {
      JComboBox comboBox = (JComboBox) componentAt;
      Object selectedItem = comboBox.getSelectedItem();
      if (selectedItem != null) {
        tooltip = selectedItem.toString();
      }
    }
    
    return tooltip;
  }

  /**
   * @see ro.sync.ecss.extensions.api.editor.InplaceRenderer#getRenderingInfo(ro.sync.ecss.extensions.api.editor.AuthorInplaceContext)
   */
  @Override
  public RendererLayoutInfo getRenderingInfo(AuthorInplaceContext context) {
    prepare(context, false);
    ApplicationComboBox urlCombo = urlPanel.getUrlCombo();
    // The combo has no insets inside the parent panel so we can use the baseline as it is.
    int baseline = urlCombo.getBaseline(urlCombo.getPreferredSize().width, urlCombo.getPreferredSize().height);
    
    RendererLayoutInfo info = new RendererLayoutInfo(
        baseline, 
        // On prepare we have already imposed a size for the panel.
        new Dimension(urlPanel.getSize().width, urlPanel.getSize().height));
    
    return info;
  }
  
  /**
   * @see ro.sync.ecss.component.editor.LoggableInplaceRenderer#dump(ro.sync.ecss.extensions.api.editor.AuthorInplaceContext, java.lang.StringBuilder)
   */
  @Override
  public void dump(AuthorInplaceContext inplaceContext, StringBuilder builder) {
    prepare(inplaceContext, false);
    
    builder.append(" value: " + urlPanel.getNotExpandedSelectedURL());
    builder.append(" color: " + urlPanel.getUrlCombo().getEditor().getEditorComponent().getForeground());
  }
}
