/*
 *  The Syncro Soft SRL License
 *
 *  Copyright (c) 1998-2012 Syncro Soft SRL, Romania.  All rights
 *  reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistribution of source or in binary form is allowed only with
 *  the prior written permission of Syncro Soft SRL.
 *
 *  2. Redistributions of source code must retain the above copyright
 *  notice, this list of conditions and the following disclaimer.
 *
 *  3. Redistributions in binary form must reproduce the above copyright
 *  notice, this list of conditions and the following disclaimer in
 *  the documentation and/or other materials provided with the
 *  distribution.
 *
 *  4. The end-user documentation included with the redistribution,
 *  if any, must include the following acknowledgment:
 *  "This product includes software developed by the
 *  Syncro Soft SRL (http://www.sync.ro/)."
 *  Alternately, this acknowledgment may appear in the software itself,
 *  if and wherever such third-party acknowledgments normally appear.
 *
 *  5. The names "Oxygen" and "Syncro Soft SRL" must
 *  not be used to endorse or promote products derived from this
 *  software without prior written permission. For written
 *  permission, please contact support@oxygenxml.com.
 *
 *  6. Products derived from this software may not be called "Oxygen",
 *  nor may "Oxygen" appear in their name, without prior written
 *  permission of the Syncro Soft SRL.
 *
 *  THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 *  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *  OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *  DISCLAIMED.  IN NO EVENT SHALL THE SYNCRO SOFT SRL OR
 *  ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 *  USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 *  ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 *  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 *  OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */
package ro.sync.ecss.component.editor;

import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.FocusAdapter;
import java.awt.event.FocusEvent;
import java.awt.event.FocusListener;
import java.awt.event.KeyAdapter;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;
import java.util.StringTokenizer;

import javax.swing.JCheckBox;
import javax.swing.JPanel;
import javax.swing.JToolTip;
import javax.swing.border.Border;

import org.apache.log4j.Logger;

import ro.sync.annotations.api.API;
import ro.sync.annotations.api.APIType;
import ro.sync.annotations.api.SourceType;
import ro.sync.contentcompletion.xml.CIValue;
import ro.sync.ecss.extensions.api.editor.AbstractInplaceEditor;
import ro.sync.ecss.extensions.api.editor.AuthorInplaceContext;
import ro.sync.ecss.extensions.api.editor.EditingEvent;
import ro.sync.ecss.extensions.api.editor.InplaceEditorArgumentKeys;
import ro.sync.ecss.extensions.api.editor.InplaceEditorCSSConstants;
import ro.sync.exml.EXMLResourceBoundle;
import ro.sync.exml.LinkLabel;
import ro.sync.exml.MessageBundle;
import ro.sync.exml.Tags;
import ro.sync.exml.view.graphics.Font;
import ro.sync.exml.view.graphics.Point;
import ro.sync.exml.view.graphics.Rectangle;
import ro.sync.ui.ListFocusTraversalPolicy;
import ro.sync.ui.application.ApplicationBorderFactory;
import ro.sync.ui.application.ApplicationScrollPane;
import ro.sync.ui.tooltip.MultiLineToolTip;


/**
 * A panel with checkboxes that can be used to render boolean values but also lists
 * of values.
 * 
 * @author alex_jitianu
 */
@API(type=APIType.INTERNAL, src=SourceType.PUBLIC)
public class PopupCheckBoxEditor extends AbstractInplaceEditor {
  /** 
   * The messages resource bundle. 
   */
  protected static final MessageBundle messages = EXMLResourceBoundle.getResourceBundleInstance();
  /**
   * Logger for logging.
   */
  private static final Logger logger = Logger.getLogger(PopupCheckBoxEditor.class.getName());
  /**
   * The size of the empty border added on the panel.
   */
  private static final int BORDER_SIZE = 5;
  /**
   * The ascending values comparator.
   */
  private PopupCheckboxComparator ASCENDING_VALUES_COMPARATOR;
  
  /**
   * The descending values comparator.
   */
  Comparator<CIValue> DESCENDING_VALUES_COMPARATOR;

  /**
   * The panel with the check-boxes.
   */
  private JPanel panel;
  /**
   * The checkboxes added inside the panel.
   */
  private List<JCheckBox> checks;
  /**
   * The values of property {@link InplaceEditorCSSConstants#PROPERTY_VALUES}.
   */
  private List<CIValue> values;
  /**
   * The value of property {@link InplaceEditorCSSConstants#PROPERTY_SEPARATOR}.
   */
  private String separator;
  /**
   * If there is a number of rows imposed we will use a scroll pane.
   */
  private ApplicationScrollPane scrollPane;
  /**
   * The labels of property {@link InplaceEditorCSSConstants#PROPERTY_LABELS}.
   */
  private List<String> labels;
  
  /**
   * Constructor.
   */
  public PopupCheckBoxEditor() {
    ASCENDING_VALUES_COMPARATOR = new PopupCheckboxComparator();
    DESCENDING_VALUES_COMPARATOR = new Comparator<CIValue>() {
      @Override
      public int compare(CIValue o1, CIValue o2) {
        return - ASCENDING_VALUES_COMPARATOR.compare(o1, o2);
      }
    };
  }
  
  /**
   * Prepare the renderer.
   * 
   * @param context In-place editing information.
   */
  private void prepare(AuthorInplaceContext context) {
    KeyListener keyListener = new KeyAdapter() {
      @Override
      public void keyPressed(KeyEvent e) {
        // On ESC cancel the editing.
        if (e.getKeyCode() == KeyEvent.VK_ESCAPE) {
          e.consume();
          cancelEditing();
        } else if (e.getKeyCode() == KeyEvent.VK_ENTER) {
          // On ENTER commit the value.
          e.consume();
          stopEditing(false, true);
        }
      }
    };
    
    FocusListener focusListener = new FocusAdapter() {
      @Override
      public void focusLost(FocusEvent e) {
        if (!checks.contains(e.getOppositeComponent())) {
          // Stop editing when the focus is no longer in one of the checkboxes.
          stopEditing(e.getOppositeComponent() == null, false);
        }
      }
      @Override
      public void focusGained(final FocusEvent e) {
        // We might have a scroll so lets make sure it's visible.
        panel.scrollRectToVisible(e.getComponent().getBounds());
      }
    };
    
    ActionListener actionListener = new ActionListener() {
      @Override
      public void actionPerformed(ActionEvent e) {
        // Mark the document as dirty each time an checkbox is changed.
        fireEditingOccured();
      }
    };
    
    panel = new JPanel(new GridBagLayout());
    GridBagConstraints c = new GridBagConstraints();
    c.gridx = 0;
    c.gridy = 0;
    c.gridwidth = 1;
    c.gridheight = 1;
    c.fill = GridBagConstraints.NONE;
    c.anchor = GridBagConstraints.WEST;
    
    panel.setBorder(ApplicationBorderFactory.createEmptyBorder());
    checks = new ArrayList<JCheckBox>();

    values = (List<CIValue>) context.getArguments().get(InplaceEditorArgumentKeys.PROPERTY_VALUES);
    labels = (List<String>) context.getArguments().get(InplaceEditorArgumentKeys.PROPERTY_LABELS);
    separator = (String) context.getArguments().get(InplaceEditorArgumentKeys.PROPERTY_SEPARATOR);
    // If no separator is specified, SPACE is used.
    if (separator == null) {
      separator = " ";
    }

    String initialValue = (String) context.getArguments().get(InplaceEditorArgumentKeys.INITIAL_VALUE);
    if (logger.isDebugEnabled()) {
      logger.debug("initialValue " + initialValue);
    }
    List<String> initialValues = new ArrayList<String>();
    if (initialValue != null) {
      // Obtain the initial values from the property.
      StringTokenizer stringTokenizer = new StringTokenizer(initialValue, separator);
      while (stringTokenizer.hasMoreTokens()) {
        String token = stringTokenizer.nextToken();
        initialValues.add(token.trim());
      }
    }
    
    String defaultValue = (String) context.getArguments().get(InplaceEditorArgumentKeys.DEFAULT_VALUE);
    if (logger.isDebugEnabled()) {
      logger.debug("defaultValue " + defaultValue);
    }
    
    List<String> defaultValues = new ArrayList<String>();
    if (defaultValue != null) {
      // Obtain the default values from the property.
      StringTokenizer stringTokenizer = new StringTokenizer(defaultValue, separator);
      while (stringTokenizer.hasMoreTokens()) {
        String token = stringTokenizer.nextToken();
        defaultValues.add(token.trim());
      }
    }
    
    ro.sync.exml.view.graphics.Color color =
        (ro.sync.exml.view.graphics.Color) context.getArguments().get(InplaceEditorArgumentKeys.PROPERTY_COLOR);
    Color imposedColor = null;
    if (color != null) {
      imposedColor = new Color(color.getRed(), color.getGreen(), color.getBlue(), color.getAlpha());
    }
    
    //Maybe we have to sort the rendered values.
    Object obj = context.getArguments().get(InplaceEditorArgumentKeys.PROPERTY_EDITOR_SORT);
    ASCENDING_VALUES_COMPARATOR.setValues(values);
    ASCENDING_VALUES_COMPARATOR.setLabels(labels);
    if(InplaceEditorArgumentKeys.SORT_ASCENDING.equals(obj)) {
      Collections.sort(values, ASCENDING_VALUES_COMPARATOR);
      Collections.sort(labels, PopupCheckBoxUtil.ASCENDING_LABELS_COMPARATOR);
    } else if(InplaceEditorArgumentKeys.SORT_DESCENDING.equals(obj)) {
      Collections.sort(values, DESCENDING_VALUES_COMPARATOR);
      Collections.sort(labels, PopupCheckBoxUtil.DESCENDING_LABELS_COMPARATOR);
    }
    
    Font font = (Font) context.getArguments().get(InplaceEditorArgumentKeys.FONT);
    // Create checkboxes for all values.
    for (int i = 0; i < values.size(); i++) {
      CIValue ciValue = values.get(i);
      String text = ciValue.getInsertString();
      if (labels != null && !labels.isEmpty()) {
        if (text.length() > 0) {
          // We have a label to present.
          text = labels.get(i);
        } else {
          labels.remove(i);
        }
      }

      // No need for empty value in a multiple popup list
      if (text.length() > 0) {
        JCheckBox checkBox = new JCheckBox(text) {
          @Override
          public JToolTip createToolTip() {
            MultiLineToolTip tooltip = new MultiLineToolTip();
            tooltip.setComponent(this);
            return tooltip;
          }
        };
        if (font != null) {
          checkBox.setFont(new java.awt.Font(font.getName(), java.awt.Font.PLAIN, font.getSize()));
        }

        if (imposedColor != null) {
          checkBox.setForeground(imposedColor);
        }

        if (initialValues.contains(ciValue.getInsertString())
            || (initialValue == null && defaultValues.contains(ciValue.getInsertString()))) {
          checkBox.setSelected(true);
        }

        // Add required listeners.
        checkBox.addKeyListener(keyListener);
        checkBox.addFocusListener(focusListener);
        checkBox.addActionListener(actionListener);
        checkBox.setToolTipText(ciValue.getAnnotation());

        panel.add(checkBox, c);
        c.gridy ++;
        checks.add(checkBox);
      } else {
        values.remove(i);
        i--;
      }
    }
    
    // Add a "Close" button to the buttom of the panel.
    LinkLabel closeBtn = new LinkLabel(messages.getString(Tags.CLOSE)) {
      @Override
      protected void performAction() {
        // Commit the value when "Close" is pressed.
        stopEditing();
      }
    };
    closeBtn.setFocusable(false);
    
    c.anchor = GridBagConstraints.EAST;
    panel.add(closeBtn, c);
    
    int rows = PopupCheckBoxUtil.getRowsProperty(context);
    Border emptyBorder =
        ApplicationBorderFactory.createEmptyBorder(BORDER_SIZE, BORDER_SIZE, BORDER_SIZE, BORDER_SIZE);

    if (rows > 0 && checks.size() > rows) {
      // We need a scroll pane.
      scrollPane = new ApplicationScrollPane(
          panel, 
          ApplicationScrollPane.VERTICAL_SCROLLBAR_AS_NEEDED,
          ApplicationScrollPane.HORIZONTAL_SCROLLBAR_NEVER);
      scrollPane.getVerticalScrollBar().setBlockIncrement(5);
      scrollPane.getVerticalScrollBar().setUnitIncrement(5);
      
      List<Component> comps = new ArrayList<Component>(checks);
      scrollPane.setFocusTraversalPolicy(new ListFocusTraversalPolicy(
          scrollPane, 
          comps, 
          comps.isEmpty() ? null : comps.get(0), false));
      scrollPane.setFocusTraversalPolicyProvider(true);
      
      panel.setBorder(emptyBorder);
    } else {
      panel.setBorder(
          ApplicationBorderFactory.createCompoundBorder(
              ApplicationBorderFactory.createLineBorder(Color.BLACK, 1),
              emptyBorder));
    }
  }

  /**
   * @see ro.sync.ecss.extensions.api.Extension#getDescription()
   */
  @Override
  public String getDescription() {
    return "A renderer that uses checkboxes to allow multiple selection.";
  }

  //////////////// EDITOR METHODS ///////////////////
  /**
   * @see ro.sync.ecss.extensions.api.editor.InplaceEditor#getEditorComponent(ro.sync.ecss.extensions.api.editor.AuthorInplaceContext, ro.sync.exml.view.graphics.Rectangle, ro.sync.exml.view.graphics.Point)
   */
  @Override
  public Object getEditorComponent(AuthorInplaceContext context, Rectangle allocation, Point mouseLocation) {
    prepare(context);
    Dimension preferredSize = null;
    if (scrollPane != null && !checks.isEmpty()) {
      // We have a scroll pane.
      Dimension checkSize = checks.get(0).getPreferredSize();
      preferredSize = panel.getPreferredSize();
      int rows = PopupCheckBoxUtil.getRowsProperty(context);
      preferredSize.height = checkSize.height * rows + 2 * BORDER_SIZE;
      preferredSize.width += scrollPane.getVerticalScrollBar().getPreferredSize().width;
    } else {
      preferredSize = panel.getPreferredSize();
    }
    // We want to present the editor relative to the mouse location.
    if (mouseLocation != null) {
      allocation.x = mouseLocation.x;
    }
    
    allocation.y += allocation.height;
    allocation.width = preferredSize.width;
    allocation.height = preferredSize.height;
    
    return scrollPane != null ? scrollPane : panel;
  }

  /**
   * @see ro.sync.ecss.extensions.api.editor.InplaceEditor#requestFocus()
   */
  @Override
  public void requestFocus() {
    if (!checks.isEmpty()) {
      checks.get(0).requestFocus();
    }
  }

  /**
   * @see ro.sync.ecss.extensions.api.editor.InplaceEditor#getValue()
   */
  @Override
  public Object getValue() {
    StringBuilder valuesStr = new StringBuilder();
    // Compose the result value.
    for (int i = 0; i < checks.size(); i++) {
      JCheckBox checkBox = checks.get(i);
      if (checkBox.isSelected()) {
        String value = values.get(i).getInsertString();
        if (valuesStr.length() > 0) {
          valuesStr.append(separator);
        }
        
        valuesStr.append(value);
      }
    }
    
    return valuesStr.toString();
  }

  /**
   * @see ro.sync.ecss.extensions.api.editor.InplaceEditor#stopEditing()
   */
  @Override
  public void stopEditing() {
    stopEditing(true, false);
  }
  
  /**
   * Stops editing.
   * 
   * @param requestFocus <code>true</code> to request focus inside the author component.
   */
  private void stopEditing(boolean requestFocus, boolean onEnter) {
    if (onEnter) {
      fireNextEditLocationRequested();
    } else {
      fireEditingStopped(new EditingEvent((String) getValue(), requestFocus));
    }
  }

  /**
   * @see ro.sync.ecss.extensions.api.editor.InplaceEditor#cancelEditing()
   */
  @Override
  public void cancelEditing() {
    fireEditingCanceled();
  }

  /**
   * @see ro.sync.ecss.extensions.api.editor.InplaceEditor#getScrollRectangle()
   */
  @Override
  public Rectangle getScrollRectangle() {
    return null;
  }
}