/*
 *  The Syncro Soft SRL License
 *
 *  Copyright (c) 1998-2012 Syncro Soft SRL, Romania.  All rights
 *  reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistribution of source or in binary form is allowed only with
 *  the prior written permission of Syncro Soft SRL.
 *
 *  2. Redistributions of source code must retain the above copyright
 *  notice, this list of conditions and the following disclaimer.
 *
 *  3. Redistributions in binary form must reproduce the above copyright
 *  notice, this list of conditions and the following disclaimer in
 *  the documentation and/or other materials provided with the
 *  distribution.
 *
 *  4. The end-user documentation included with the redistribution,
 *  if any, must include the following acknowledgment:
 *  "This product includes software developed by the
 *  Syncro Soft SRL (http://www.sync.ro/)."
 *  Alternately, this acknowledgment may appear in the software itself,
 *  if and wherever such third-party acknowledgments normally appear.
 *
 *  5. The names "Oxygen" and "Syncro Soft SRL" must
 *  not be used to endorse or promote products derived from this
 *  software without prior written permission. For written
 *  permission, please contact support@oxygenxml.com.
 *
 *  6. Products derived from this software may not be called "Oxygen",
 *  nor may "Oxygen" appear in their name, without prior written
 *  permission of the Syncro Soft SRL.
 *
 *  THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 *  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *  OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *  DISCLAIMED.  IN NO EVENT SHALL THE SYNCRO SOFT SRL OR
 *  ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 *  USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 *  ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 *  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 *  OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */
package ro.sync.ecss.extensions.api.link;

import java.util.List;

import ro.sync.annotations.api.API;
import ro.sync.annotations.api.APIType;
import ro.sync.annotations.api.SourceType;
import ro.sync.contentcompletion.xml.Context;

/**
 * Recognizer for ID declaration and references in attribute values.
 * The recognizer may be used when search references and declaration for a particular ID.
 * 
 * @author radu_pisoi
 */
@API(type=APIType.EXTENDABLE, src=SourceType.PUBLIC)
public abstract class IDTypeRecognizer {
  
  /**
   * Mode used when locate an ID. It is used to locate declaration.
   */
  public static final short MODE_LOCATE_DECLARATIONS = 1;
  
  /**
   * Mode used when locate an ID. It is used to locate references.
   */
  public static final short MODE_LOCATE_REFERENCES = 2;
  
  /**
   * Detect the ID declaration or reference for the provided attribute context and offset. 
   * The offset is relative to the attribute value.
   * 
   * @param systemID The systemID of the resource that specifies the attribute.
   * @param context The element content to detect the ID. 
   * The top element from the context element stack represents the parent element.
   * @param attrName The attribute name.
   * @param attrNs The attribute namespace.
   * @param attributeValue The attribute value.
   * @param offset The offset that is relative to the attribute value. It is zero based. 
   * If it is -1 and the attribute type is IDREFS then all the IDs should be returned. 
   * 
   * @return The ID identifier for the given attribute and offset. 
   * If the offset is -1 and the attribute type is IDREFS then all the IDs should be returned. 
   * Can be null if no ID was detected.
   * 
   * @throws CannotRecognizeIDException Exception that can be thrown when an ID cannot be identified in the given context. 
   */

  public abstract List<IDTypeIdentifier> detectIDType(
      String systemID,
      Context context,
      String attrName, 
      String attrNs, 
      String attributeValue,
      int offset) throws CannotRecognizeIDException;

  /**
   * Detect if the given ID is located in the specified attribute. 
   * If an attribute declaration or reference was identified then compute it's location relative to the attribute value.
   * 
   * Usually the method is used for attributes with IDREFS type to detect the internal ID references.<br />
   * By example, for the content 'id1 id2 id3' the detection should have the following results: <br/>
   * - if the offset is 0 the detected ID should be id1; <br/>
   * - if the offset is 3 the detected ID should be id1; <br/>
   * - if the offset is 4 the detected ID should be id2; <br/>
   * 
   * @param systemID The systemID of the resource that specifies the attribute.
   * @param context The element content to detect the ID. 
   * The top element from the context element stack represents the parent element.
   * @param attrName The attribute name.
   * @param attrNs The attribute namespace.
   * @param attributeValue The attribute value.
   * @param idIdentifier The ID identifier.
   * @param mode The detection mode that is represented as a bitwise mask. 
   * Supported modes are {@link IDTypeRecognizer#MODE_LOCATE_REFERENCES} and {@link IDTypeRecognizer#MODE_LOCATE_DECLARATIONS}.
   * 
   * @return The location of the ID [startOffset, endOffset) in the given attribute value,
   * by example if the attribute value is 'idRef1 idRef2' and we are looking for 'idRef1' 
   * the method should return [0, 6). The endOffset is exclusive. 
   * Null value can be returned if the ID was not located.  
   */
  public abstract int[] locateIDType(
      String systemID,
      Context context,
      String attrName, 
      String attrNs, 
      String attributeValue,
      IDTypeIdentifier idIdentifier,
      short mode);
  
  /**
   * If <code>false</code> then disable the default recognition of the IDs based on the current associated schema. 
   * Otherwise the IDs declaration and references will be detected for document with DTD, XML Schema or 
   * RelaxNG schemas.  
   * 
   * @return True to disable the recognition of the IDs based on the current associated schema.
   */
  public abstract boolean isDefaultIDTypeRecognitionAvailable();
  
  /**
   * If true then ID type recognition is available. If this method return false also the default ID type 
   * recognition will be disabled.
   * 
   * @return False to disable the recognition of the IDs.
   */
  public abstract boolean isIDTypeRecognitionAvailable();
}
