package ro.sync.ecss.extensions.api.webapp.cc;

import java.util.List;

import javax.swing.text.BadLocationException;

import ro.sync.annotations.api.API;
import ro.sync.annotations.api.APIType;
import ro.sync.annotations.api.SourceType;
import ro.sync.ecss.extensions.api.AuthorSelectionAndCaretModel;

/**
 * This class offers support for actions with content completion such as: insert
 * element, surround with tags and rename element.
 * 
 * For every action we have two methods: one that returns a list of proposals
 * to be presented to the user and another one that performs the action based
 * on the user choice. 
 * 
 * Users of the API will typically show a dialog with the proposals and after 
 * the user selects one of the call the second method to finish the action. 
 * 
 * @since 15.1
 */
@API(type=APIType.NOT_EXTENDABLE, src=SourceType.PUBLIC)
public interface ContentCompletionManager {

	/**
	 * Returns the proposals for element insertions at the current offset.
	 * 
	 * @param selectionModel The selection and caret model of the document.
	 * 
	 * @return The proposals for element insertions at the current offset.
	 * 
	 * @throws BadLocationException
	 */
	public abstract List<CCItemProxy> getProposedElementsForInsert(AuthorSelectionAndCaretModel selectionModel)
			throws BadLocationException;

	/**
	 * Executes the insert element action.
	 * 
	 * @param ccItem The content completion item chosen by the user.
	 * @param selectionModel The selection and caret model of the document. 
	 * 
	 * @throws BadLocationException
	 * @throws ItemNotFoundException
	 */
	public abstract void executeInsert(CCItemProxy ccItem, AuthorSelectionAndCaretModel selectionModel)
			throws BadLocationException, ItemNotFoundException;

	/**
	 * Returns the proposals for renaming.
	 * 
	 * @param selectionModel The selection and caret model of the document.
	 * 
	 * @return The proposed elements for rename.
	 * 
	 * @throws BadLocationException
	 */
	public abstract List<CCItemProxy> getProposedElementsForRename(AuthorSelectionAndCaretModel selectionModel) throws BadLocationException;

	/**
	 * Executes the rename operation.
	 * 
	 * @param selectionModel The selection and caret model of the document.
	 * @param ccItem The content completion item chosen for rename.
	 * 
	 * @throws BadLocationException
	 */
	public abstract void executeRename(AuthorSelectionAndCaretModel selectionModel,
			CCItemProxy ccItem) throws BadLocationException;

	/**
	 * Returns the proposals for a tag to surround the current selection with.
	 * 
	 * @param selectionModel The selection and caret model of the document. 
	 * 
	 * @return The list of proposals.
	 * 
	 * @throws BadLocationException
	 */
	public abstract List<CCItemProxy> getProposedElementsForSurround(AuthorSelectionAndCaretModel selectionModel)
			throws BadLocationException;

	/**
	 * Surrounds the current selection with an element that has the specified
	 * render string and type as chosen by the user/
	 * 
   * @param ccItem The content completion item chosen by the user.
	 * @param selectionModel The selection and caret model of the document.
	 * 
	 * @throws BadLocationException
	 * @throws ItemNotFoundException
	 */
	public abstract void executeSurround(CCItemProxy ccItem, AuthorSelectionAndCaretModel selectionModel)
			throws BadLocationException, ItemNotFoundException;
	
	
	/**
	 * Returns the block elements around the caret position that can be split.
	 * 
	 * @param selectionModel The selection and caret model of the document.
	 * 
	 * @return The list of block elements that can be split.
	 * 
	 * @throws BadLocationException
	 */
  public abstract List<CCItemProxy> getProposedElementsToSplit(
      AuthorSelectionAndCaretModel selectionModel) throws BadLocationException;

  /**
   * Splits the element chosen by the user.
   * 
   * @param ccItem The element chosen by the user.
   * @param selectionModel The selection model of the document.
   * 
   * @throws BadLocationException
   * @throws ItemNotFoundException
   */
  public abstract void executeSplit(CCItemProxy ccItem, AuthorSelectionAndCaretModel selectionModel)
      throws BadLocationException, ItemNotFoundException;
}