/*
 *  The Syncro Soft SRL License
 *
 *  Copyright (c) 1998-2009 Syncro Soft SRL, Romania.  All rights
 *  reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistribution of source or in binary form is allowed only with
 *  the prior written permission of Syncro Soft SRL.
 *
 *  2. Redistributions of source code must retain the above copyright
 *  notice, this list of conditions and the following disclaimer.
 *
 *  3. Redistributions in binary form must reproduce the above copyright
 *  notice, this list of conditions and the following disclaimer in
 *  the documentation and/or other materials provided with the
 *  distribution.
 *
 *  4. The end-user documentation included with the redistribution,
 *  if any, must include the following acknowledgment:
 *  "This product includes software developed by the
 *  Syncro Soft SRL (http://www.sync.ro/)."
 *  Alternately, this acknowledgment may appear in the software itself,
 *  if and wherever such third-party acknowledgments normally appear.
 *
 *  5. The names "Oxygen" and "Syncro Soft SRL" must
 *  not be used to endorse or promote products derived from this
 *  software without prior written permission. For written
 *  permission, please contact support@oxygenxml.com.
 *
 *  6. Products derived from this software may not be called "Oxygen",
 *  nor may "Oxygen" appear in their name, without prior written
 *  permission of the Syncro Soft SRL.
 *
 *  THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 *  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *  OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *  DISCLAIMED.  IN NO EVENT SHALL THE SYNCRO SOFT SRL OR
 *  ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 *  USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 *  ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 *  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 *  OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */
package ro.sync.ecss.extensions.dita.conref;

import java.io.IOException;

import org.apache.xerces.impl.Constants;
import org.xml.sax.ContentHandler;
import org.xml.sax.DTDHandler;
import org.xml.sax.EntityResolver;
import org.xml.sax.ErrorHandler;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.SAXNotRecognizedException;
import org.xml.sax.SAXNotSupportedException;
import org.xml.sax.XMLReader;
import org.xml.sax.ext.LexicalHandler;

import ro.sync.annotations.api.API;
import ro.sync.annotations.api.APIType;
import ro.sync.annotations.api.SourceType;

/**
 * Delegating XML Reader used to parse DITA 'conref' references.
 */
@API(type=APIType.INTERNAL, src=SourceType.PUBLIC)
public class DITAXMLReaderWrapper implements XMLReader {

  /**
   * The wrapped XML Reader.
   */
  private XMLReader xmlReader;
  
  /**
   * The filtering handler.
   */
  private DITAFilteringContentHandler filteringContentHandler;

  /**
   * Constructor.
   * 
   * @param xmlReader The wrapped XML Reader.
   * @param topicPath The topic IDs path. If null, the first encountered topic will be used.
   * @param endTopicPath If a "conrefend" is specified, this is the end range path
   * @param isKeyReference <code>true</code> if the reference is a key reference.
   *  
   * @throws SAXNotSupportedException 
   * @throws SAXNotRecognizedException 
   */
  public DITAXMLReaderWrapper(XMLReader xmlReader, String[] topicPath, String[] endTopicPath, boolean isKeyReference) 
    throws SAXNotRecognizedException, SAXNotSupportedException {
    this.xmlReader = xmlReader;
    this.filteringContentHandler = new DITAFilteringContentHandler(topicPath, endTopicPath, isKeyReference);
    xmlReader.setContentHandler(filteringContentHandler);
    // Disable XInclude feature
    xmlReader.setFeature(
        Constants.XERCES_FEATURE_PREFIX + Constants.XINCLUDE_FEATURE, false);
    xmlReader.setProperty(Constants.SAX_PROPERTY_PREFIX + Constants.LEXICAL_HANDLER_PROPERTY, 
        filteringContentHandler);
  }
  
  /**
   * @see org.xml.sax.XMLReader#getContentHandler()
   */
  public ContentHandler getContentHandler() {
    return filteringContentHandler;
  }

  /**
   * @see org.xml.sax.XMLReader#getDTDHandler()
   */
  public DTDHandler getDTDHandler() {
    return xmlReader.getDTDHandler();
  }

  /**
   * @see org.xml.sax.XMLReader#getEntityResolver()
   */
  public EntityResolver getEntityResolver() {
    return xmlReader.getEntityResolver();
  }

  /**
   * @see org.xml.sax.XMLReader#getErrorHandler()
   */
  public ErrorHandler getErrorHandler() {
    return xmlReader.getErrorHandler();
  }

  /**
   * @see org.xml.sax.XMLReader#getFeature(java.lang.String)
   */
  public boolean getFeature(String name) throws SAXNotRecognizedException, SAXNotSupportedException {
    return xmlReader.getFeature(name);
  }

  /**
   * @see org.xml.sax.XMLReader#getProperty(java.lang.String)
   */
  public Object getProperty(String name) throws SAXNotRecognizedException, SAXNotSupportedException {
    if (name.equals(Constants.SAX_PROPERTY_PREFIX + Constants.LEXICAL_HANDLER_PROPERTY)) {
      return filteringContentHandler;
    } else {
      return xmlReader.getProperty(name);
    }
  }

  /**
   * @see org.xml.sax.XMLReader#parse(org.xml.sax.InputSource)
   */
  public void parse(InputSource input) throws IOException, SAXException {
    xmlReader.parse(input);
  }

  /**
   * @see org.xml.sax.XMLReader#parse(java.lang.String)
   */
  public void parse(String systemId) throws IOException, SAXException {
    xmlReader.parse(systemId);
  }

  /**
   * @see org.xml.sax.XMLReader#setContentHandler(org.xml.sax.ContentHandler)
   */
  public void setContentHandler(ContentHandler handler) {
    filteringContentHandler.setContentHandler(handler);
  }

  /**
   * @see org.xml.sax.XMLReader#setDTDHandler(org.xml.sax.DTDHandler)
   */
  public void setDTDHandler(DTDHandler handler) {
    xmlReader.setDTDHandler(handler);
  }

  /**
   * @see org.xml.sax.XMLReader#setEntityResolver(org.xml.sax.EntityResolver)
   */
  public void setEntityResolver(EntityResolver resolver) {
    xmlReader.setEntityResolver(resolver);
  }

  /**
   * @see org.xml.sax.XMLReader#setErrorHandler(org.xml.sax.ErrorHandler)
   */
  public void setErrorHandler(ErrorHandler handler) {
    xmlReader.setErrorHandler(handler);
  }

  /**
   * @see org.xml.sax.XMLReader#setFeature(java.lang.String, boolean)
   */
  public void setFeature(String name, boolean value) throws SAXNotRecognizedException,
      SAXNotSupportedException {
    xmlReader.setFeature(name, value);
  }

  /**
   * @see org.xml.sax.XMLReader#setProperty(java.lang.String, java.lang.Object)
   */
  public void setProperty(String name, Object value) throws SAXNotRecognizedException,
      SAXNotSupportedException {
    if (name.equals(Constants.SAX_PROPERTY_PREFIX + Constants.LEXICAL_HANDLER_PROPERTY)) {
      filteringContentHandler.setLexicalHandler((LexicalHandler) value);
    } else {
      xmlReader.setProperty(name, value);
    }
  }
}