/*
*  The Syncro Soft SRL License
 *
 *  Copyright (c) 1998-2012 Syncro Soft SRL, Romania.  All rights
 *  reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistribution of source or in binary form is allowed only with
 *  the prior written permission of Syncro Soft SRL.
 *
 *  2. Redistributions of source code must retain the above copyright
 *  notice, this list of conditions and the following disclaimer.
 *
 *  3. Redistributions in binary form must reproduce the above copyright
 *  notice, this list of conditions and the following disclaimer in
 *  the documentation and/or other materials provided with the
 *  distribution.
 *
 *  4. The end-user documentation included with the redistribution,
 *  if any, must include the following acknowledgment:
 *  "This product includes software developed by the
 *  Syncro Soft SRL (http://www.sync.ro/)."
 *  Alternately, this acknowledgment may appear in the software itself,
 *  if and wherever such third-party acknowledgments normally appear.
 *
 *  5. The names "Oxygen" and "Syncro Soft SRL" must
 *  not be used to endorse or promote products derived from this
 *  software without prior written permission. For written
 *  permission, please contact support@oxygenxml.com.
 *
 *  6. Products derived from this software may not be called "Oxygen",
 *  nor may "Oxygen" appear in their name, without prior written
 *  permission of the Syncro Soft SRL.
 *
 *  THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 *  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *  OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *  DISCLAIMED.  IN NO EVENT SHALL THE SYNCRO SOFT SRL OR
 *  ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 *  USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 *  ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 *  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 *  OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */
package ro.sync.ecss.extensions.dita.link;

import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.Set;
import java.util.WeakHashMap;

import org.apache.log4j.Logger;

import ro.sync.annotations.api.API;
import ro.sync.annotations.api.APIType;
import ro.sync.annotations.api.SourceType;
import ro.sync.ecss.dita.ContextKeyManager;
import ro.sync.ecss.dita.DITAAccess;
import ro.sync.ecss.dita.Reference;
import ro.sync.ecss.dita.reference.keyref.KeyInfo;
import ro.sync.ecss.extensions.api.AuthorAccess;
import ro.sync.ecss.extensions.api.AuthorListenerAdapter;
import ro.sync.ecss.extensions.api.link.LinkTextResolver;
import ro.sync.ecss.extensions.api.node.AttrValue;
import ro.sync.ecss.extensions.api.node.AuthorElement;
import ro.sync.ecss.extensions.api.node.AuthorNode;
import ro.sync.util.URLUtil;

/**
 * Can resolve DITA references to another topic made through the href attribute on elements of classes: 
 * map/topicref ,  topic/xref and  topic/link .  
 * 
 * It also resolves key references provided that the ditamap is opened in DITA Map Manager."
 * @since 14.2
 */
@API(type=APIType.EXTENDABLE, src=SourceType.PUBLIC)
public class DitaLinkTextResolver extends LinkTextResolver {
  /**
   * Logger for logging.
   */
  private static final Logger logger = Logger.getLogger(DitaLinkTextResolver.class.getName());

  /**
   * The <code>keyref</code> keyref.
   */
  private static final String ATTRIBUTE_KEYREF = "keyref";
  /**
   * Href attribute.
   */
  private static final String ATTRIBUTE_HREF = "href";
  
  /**
   * Cached info about a resolved DITA link content. 
   */
  private static class ResolvedLinkInfo {
    /**
     * Resolved content.
     */
    final String content;
    /**
     * Location where the link pointed at.
     */
    final String definitionLocation;
    /**
     * Constructor.
     * 
     * @param content Resolved content.
     * @param definitionLocation Location where the link was defined.
     */
    public ResolvedLinkInfo(String content, String definitionLocation) {
      this.content = content;
      this.definitionLocation = definitionLocation;
    }
    
    @Override
    public String toString() {
      return content + ", location: " + definitionLocation;
    }
  }
  
  /**
   * Cached resolved DITA keyrefs. Maps a key name to a text content of the key.
   */
  private final HashMap<String, ResolvedLinkInfo> keyrefs = new HashMap<String, ResolvedLinkInfo>();
  
  /**
   * Cached resolved DITA hrefs. Maps a href to the title that should be presented for the
   * referred topic.
   */
  private final HashMap<String, ResolvedLinkInfo> hrefs = new HashMap<String, ResolvedLinkInfo>();
  /**
   * Keeps a track of all the elements that have used keys from a given URL.
   * 
   * The value is a WeakHashMap in which only the keys are used. The WeakHashMap
   * uses weak references an more than that it does a clean up of the garbage collected
   * keys. 
   */
  private final HashMap<String, WeakHashMap<AuthorNode, Object>> ditaLinkElements =
      new HashMap<String, WeakHashMap<AuthorNode, Object>>();
  
  /**
   * For a map/topicref the attributes locktitle and navtitle control the expanded text.
   * So when they change we must also refresh.
   */
  private AuthorListenerAdapter authorListenerAdapter = new AuthorListenerAdapter() {
    @Override
    public void attributeChanged(ro.sync.ecss.extensions.api.AttributeChangedEvent e) {
      AuthorElement element = (AuthorElement) e.getOwnerAuthorNode();
      String attrName = e.getAttributeName();
      
      if (resolvesTopicrefs) {
        // If the resolver was never for topicrefs used there is no point in refreshing. 
        // oxy_link_text() was encountered on topicref styles. Changing these two 
        // attributes will give a different text for the element so we'll refresh.
        AttrValue classValue = element.getAttribute("class");
        if (classValue != null 
            && classValue.getValue().contains(" map/topicref ")) {
          boolean depends = "locktitle".equals(attrName) ||
              "navtitle".equals(attrName);
          if (depends) {
            // Refresh the node so that it's styles are recomputed.
            authorAccess.getEditorAccess().refresh(element);
          }
        }    
      }
    }
  };

  private AuthorAccess authorAccess;
  /**
   * <code>true</code> if the resolver has encountered topicrefs.
   */
  private boolean resolvesTopicrefs;

  /**
   * The context-aware key manager.
   */
  private ContextKeyManager keyManager;
  
  /**
   * Constructor.
   */
  public DitaLinkTextResolver() {
    this.keyManager = ContextKeyManager.getDefault();
  }

  /**
   * Constructor.
   * 
   * @param keyManager The context-aware key manager.
   */
  public DitaLinkTextResolver(ContextKeyManager keyManager) {
    this.keyManager = keyManager;
  }

  /**
   * Get the text of the reference.
   * 
   * @param node Author node.
   * 
   * @return The link text.
   */
  @Override
  public String resolveReference(AuthorNode node) {
    ResolvedLinkInfo resolvedInfo = null;

    if (node.getType() == AuthorNode.NODE_TYPE_ELEMENT) {
      AuthorElement element = (AuthorElement) node;

      AttrValue attrValue = element.getAttribute(ATTRIBUTE_KEYREF);
      if(attrValue != null 
          && attrValue.getValue() != null 
          //If key contains a "/", it references a subelement, so do not bother looking for metadata.
          && !attrValue.getValue().contains("/")) {
        // The text for each keyref is resolved differently based the class attribute of the 
        // element with the "keyref" attribute.
        String keyrefCacheKey = buildKeyrefCacheKey(element, attrValue.getValue());
        resolvedInfo = keyrefs.get(keyrefCacheKey);

        if (resolvedInfo == null) {
          // This key is not present in the cache. Resolve it now.
          LinkedHashMap<String, KeyInfo> keys = DITAAccess.getKeys(node.getXMLBaseURL(), keyManager);
          KeyInfo keyInfo = keys != null ? keys.get(attrValue.getValue()) : null;
          if (logger.isDebugEnabled()) {
            logger.debug("Resolve using keyref: " + node + " to " + keyInfo);
          }
          if(keyInfo != null) {
            if(keyInfo.getMetaContentProvider() != null) {
              AttrValue classValue = element.getAttribute("class");
              if(classValue != null) {
                String linkContent = keyInfo.getMetaContentProvider().getContent(classValue.getValue());
                if (linkContent != null) {
                  resolvedInfo = new ResolvedLinkInfo(linkContent, URLUtil.removeQueryOrAnchorFromName(keyInfo.getDefinitionLocation().toExternalForm()));
                }
              }
            } 
            
            if (resolvedInfo == null
                && Boolean.TRUE.equals(keyInfo.getLocktitle())
                && keyInfo.getNavTitle() != null) {
              resolvedInfo = new ResolvedLinkInfo(keyInfo.getNavTitle(), keyInfo.getHrefLocation().toExternalForm());
            }
            
            if (resolvedInfo == null && keyInfo.getHrefLocation() != null) {
              // If we don't have content we will try to detect the title by following
              // the href attribute.
              String absoluteLocation = keyInfo.getHrefLocation().toExternalForm();
              try {
                String targetInfo = DITAAccess.computeLinkText(absoluteLocation, absoluteLocation);
                if (logger.isDebugEnabled()) {
                  logger.debug("targetInfo " + targetInfo);
                }
                if (targetInfo != null) {
                  resolvedInfo = new ResolvedLinkInfo(targetInfo, absoluteLocation);
                }
              } catch (Exception e) {
                // Avoid going after the resource again.
                resolvedInfo = new ResolvedLinkInfo(null, absoluteLocation);
              }            
            }
            
            keyrefs.put(keyrefCacheKey, resolvedInfo);
            if (logger.isDebugEnabled()) {
              logger.debug("Resolved " + attrValue.getValue() + " to " + resolvedInfo);
            }
          }
        } else {
          if (logger.isDebugEnabled()) {
            logger.debug("Cache hit! " + node);
          }
        }
      }

      if (resolvedInfo == null) {
        // If it's a topicref we should first check @locktitle and @navtitle
        AttrValue classValue = element.getAttribute("class");
        if (classValue != null 
            && classValue.getValue().contains(" map/topicref ")) {
          resolvesTopicrefs = true;
          // Check @locktitle.
          AttrValue locktitle = element.getAttribute("locktitle");
          boolean isTopicHead = classValue.getValue().contains(" mapgroup-d/topichead ");
          if (
              // EXM-28109 A topic head doesn't need the title to be locked.
              isTopicHead || 
              locktitle != null && "yes".equals(locktitle.getValue())) {
            // We must use @navtitle, if any.
            AttrValue navtitle = element.getAttribute("navtitle");
            if (navtitle != null) {
              resolvedInfo = new ResolvedLinkInfo(navtitle.getValue(), null);
            } 
          }
        }
        
        if (resolvedInfo == null) {
          attrValue = element.getAttribute(ATTRIBUTE_HREF);
          if (attrValue != null  
              && isHrefElement(element) 
              && node.getXMLBaseURL() != null 
              && isLocalReference(element)) {
            
            resolvedInfo = hrefs.get(attrValue.getValue());
            if (resolvedInfo == null) {
              if (logger.isDebugEnabled()) {
                logger.debug("Resolve using href: " + node);
              }
              try {
                String targetInfo = DITAAccess.computeLinkText(attrValue.getValue(), node.getXMLBaseURL().toExternalForm());
                if (targetInfo != null) {
                  Reference parsed = 
                      DITAAccess.parseDITAHref(node.getXMLBaseURL().toExternalForm(), attrValue.getValue());
                  String linkLocation = URLUtil.removeQueryOrAnchorFromName(parsed.getUri());
                  resolvedInfo = new ResolvedLinkInfo(targetInfo, linkLocation);
                  hrefs.put(attrValue.getValue(), resolvedInfo);
                }
              } catch (Exception e) {
                // Avoid going after the resource again.
                hrefs.put(attrValue.getValue(), new ResolvedLinkInfo(null, null));
              }
            } else {
              if (logger.isDebugEnabled()) {
                logger.debug("Cache hit! " + node);
              }
            }
          }
        }
      }

      if (resolvedInfo != null && resolvedInfo.definitionLocation != null) {
        WeakHashMap<AuthorNode, Object> dependencies = ditaLinkElements.get(resolvedInfo.definitionLocation);
        if (dependencies == null) {
          dependencies = new WeakHashMap<AuthorNode, Object>();

          // Keep track of the element to refresh it's styles when needed.
          ditaLinkElements.put(resolvedInfo.definitionLocation, dependencies);
        }

        dependencies.put(node, null);
      }
    }
    
    return resolvedInfo != null ? resolvedInfo.content : null;
  }
  
  /**
   * Checks if the reference is a local one and to a DITA file.
   * 
   * @param element Reference element.
   * 
   * @return <code>true</code> if this is a local reference.
   */
  private static boolean isLocalReference(AuthorElement element) {
    AttrValue formatAttr = element.getAttribute("format");
    String format = formatAttr != null ? formatAttr.getValue() : null;
    // Scope
    AttrValue scopeAttr = element.getAttribute("scope");
    String scope = scopeAttr != null ? scopeAttr.getValue() : null;
    boolean ditaFormat = format == null || "dita".equals(format) || "ditamap".equals(format);
    boolean localScope = scope == null ||  "local".equals(scope);
    
    return ditaFormat && localScope;
  }

  /**
   * Update the cache of references.
   * 
   * @param modifiedURLs The URLs that are modified.
   */
  @Override
  public void update(Set<String> modifiedURLs) {
    Set<String> keySet = new HashSet<String>(keyrefs.keySet());
    for (String key : keySet) {
      ResolvedLinkInfo resolvedKeyInfo = keyrefs.get(key);
      if (resolvedKeyInfo == null ||
          // The key wasn't resolved. Try to get it again.
          resolvedKeyInfo.definitionLocation == null ||
          modifiedURLs.contains(resolvedKeyInfo.definitionLocation)) {
        keyrefs.remove(key);
      }
    }
    
    keySet = new HashSet<String>(hrefs.keySet());
    for (String key : keySet) {
      ResolvedLinkInfo resolvedKeyInfo = hrefs.get(key);
      if (// The href wasn't resolved. Try to get it again.
          resolvedKeyInfo.definitionLocation == null ||
          modifiedURLs.contains(resolvedKeyInfo.definitionLocation)) {
        hrefs.remove(key);
      }
    }
    
    for (String url : modifiedURLs) {
      Map<AuthorNode, Object> nodes = ditaLinkElements.get(url);
      if (nodes != null) {
        Set<AuthorNode> toRefresh = nodes.keySet();
        for (AuthorNode authorNode : toRefresh) {
          authorAccess.getEditorAccess().refresh(authorNode);
        }
      }
    }
  }

  /**
   * Checks if this element is an element that supports an href attribute.
   * 
   * @param element
   * @return
   */
  private boolean isHrefElement(AuthorElement element) {
      // Check the class attribute.
      AttrValue classValue = element.getAttribute("class");
      boolean hasReferences = 
          classValue != null 
          && (classValue.getValue().contains(" map/topicref ")
                || classValue.getValue().contains(" topic/xref ")
                || classValue.getValue().contains(" topic/link "));
      
    return hasReferences;
  }

  /**
   * @see ro.sync.ecss.extensions.api.link.LinkTextResolver#refresh()
   */
  @Override
  public void refresh() {
    keyrefs.clear();
    hrefs.clear();
    Collection<WeakHashMap<AuthorNode, Object>> values = ditaLinkElements.values();
    for (Iterator iterator = values.iterator(); iterator.hasNext();) {
      WeakHashMap<AuthorNode, Object> weakHashMap = (WeakHashMap<AuthorNode, Object>) iterator
          .next();
      Set<AuthorNode> keySet = weakHashMap.keySet();
      for (AuthorNode authorNode : keySet) {
        authorAccess.getEditorAccess().refresh(authorNode);
      }
      
    }
    ditaLinkElements.clear();
  }

  /**
   * @see ro.sync.ecss.extensions.api.link.LinkTextResolver#refreshNodeReferences(ro.sync.ecss.extensions.api.node.AuthorNode)
   */
  @Override
  public void refreshNodeReferences(AuthorNode node) {
    if (node.getType() == AuthorNode.NODE_TYPE_ELEMENT) {
      AuthorElement element = (AuthorElement) node;
      AttrValue attrValue = element.getAttribute(ATTRIBUTE_KEYREF);
      if(attrValue != null 
          && attrValue.getValue() != null 
          //If key contains a "/", it references a subelement, so do not bother looking for metadata.
          && !attrValue.getValue().contains("/")) {
        keyrefs.remove(buildKeyrefCacheKey(element, attrValue.getValue()));
      }

      attrValue = element.getAttribute(ATTRIBUTE_HREF);
      if (attrValue != null  
          && isHrefElement(element) 
          && node.getXMLBaseURL() != null 
          && isLocalReference(element)) {
        hrefs.remove(attrValue.getValue());
      }
      
      // Call refresh so that the reference is computed again.
      authorAccess.getEditorAccess().refresh(node);
    }
  }
  
  /**
   * Build the cache key for a keyRef with the given name that appears on 
   * the given node.
   *
   * @param node The node on which the keyRef appears.
   * @param keyName The name of the keyRef.
   * 
   * @return The cache key.
   */
  public String buildKeyrefCacheKey(AuthorElement node, String keyName) {
    return keyName + "\"" + node.getAttribute("class").getValue();
  }
  
  /**
   * @see ro.sync.ecss.extensions.api.link.LinkTextResolver#clearReferencesCache()
   */
  @Override
  public void clearReferencesCache() {
    keyrefs.clear();
    hrefs.clear();
    ditaLinkElements.clear();
  }

  /**
   * @see ro.sync.ecss.extensions.api.AuthorExtensionStateListener#activated(ro.sync.ecss.extensions.api.AuthorAccess)
   */
  @Override
  public void activated(AuthorAccess authorAccess) {
    this.authorAccess = authorAccess;
    authorAccess.getDocumentController().addAuthorListener(authorListenerAdapter);
  }

  /**
   * @see ro.sync.ecss.extensions.api.AuthorExtensionStateListener#deactivated(ro.sync.ecss.extensions.api.AuthorAccess)
   */
  @Override
  public void deactivated(AuthorAccess authorAccess) {
    authorAccess.getDocumentController().removeAuthorListener(authorListenerAdapter);
  }
}
