/*
 *  The Syncro Soft SRL License
 *
 *  Copyright (c) 1998-2009 Syncro Soft SRL, Romania.  All rights
 *  reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistribution of source or in binary form is allowed only with
 *  the prior written permission of Syncro Soft SRL.
 *
 *  2. Redistributions of source code must retain the above copyright
 *  notice, this list of conditions and the following disclaimer.
 *
 *  3. Redistributions in binary form must reproduce the above copyright
 *  notice, this list of conditions and the following disclaimer in
 *  the documentation and/or other materials provided with the
 *  distribution.
 *
 *  4. The end-user documentation included with the redistribution,
 *  if any, must include the following acknowledgment:
 *  "This product includes software developed by the
 *  Syncro Soft SRL (http://www.sync.ro/)."
 *  Alternately, this acknowledgment may appear in the software itself,
 *  if and wherever such third-party acknowledgments normally appear.
 *
 *  5. The names "Oxygen" and "Syncro Soft SRL" must
 *  not be used to endorse or promote products derived from this
 *  software without prior written permission. For written
 *  permission, please contact support@oxygenxml.com.
 *
 *  6. Products derived from this software may not be called "Oxygen",
 *  nor may "Oxygen" appear in their name, without prior written
 *  permission of the Syncro Soft SRL.
 *
 *  THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 *  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *  OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *  DISCLAIMED.  IN NO EVENT SHALL THE SYNCRO SOFT SRL OR
 *  ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 *  SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 *  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 *  USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 *  ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 *  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 *  OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */
package ro.sync.ecss.extensions.docbook.table;

import javax.swing.text.BadLocationException;

import ro.sync.annotations.api.API;
import ro.sync.annotations.api.APIType;
import ro.sync.annotations.api.SourceType;
import ro.sync.ecss.extensions.api.ArgumentDescriptor;
import ro.sync.ecss.extensions.api.ArgumentsMap;
import ro.sync.ecss.extensions.api.AuthorAccess;
import ro.sync.ecss.extensions.api.AuthorOperationException;
import ro.sync.ecss.extensions.api.AuthorOperationStoppedByUserException;
import ro.sync.ecss.extensions.api.node.AuthorDocumentFragment;
import ro.sync.ecss.extensions.api.node.AuthorElement;
import ro.sync.ecss.extensions.api.node.AuthorNode;
import ro.sync.ecss.extensions.commons.table.operations.AbstractTableOperation;
import ro.sync.ecss.extensions.commons.table.operations.AuthorTableHelper;
import ro.sync.ecss.extensions.commons.table.operations.InsertTableOperationBase;
import ro.sync.ecss.extensions.commons.table.operations.TableCustomizerConstants;
import ro.sync.ecss.extensions.commons.table.operations.TableCustomizerConstants.ColumnWidthsType;
import ro.sync.ecss.extensions.commons.table.operations.TableInfo;
import ro.sync.ecss.extensions.commons.table.operations.TableOperationsUtil;
import ro.sync.ecss.extensions.commons.table.operations.cals.CALSDocumentTypeHelper;

/**
 * Operation used to insert a CALS Docbook table or 
 * an HTML table in a DocBook document. 
 */
@API(type=APIType.INTERNAL, src=SourceType.PUBLIC)
public class InsertTableOperation extends AbstractTableOperation implements InsertTableOperationBase {
  
  /**
   * Constructor.
   */
  public InsertTableOperation() {
    super(new CALSDocumentTypeHelper());
  }

  /**
   * Arguments of the operation.
   */
  private static final ArgumentDescriptor[] ARGUMENTS = new ArgumentDescriptor[] { NAMESPACE_ARGUMENT_DESCRIPTOR };
  
  /**
   * @see ro.sync.ecss.extensions.api.AuthorOperation#doOperation(ro.sync.ecss.extensions.api.AuthorAccess, ro.sync.ecss.extensions.api.ArgumentsMap)
   */
  @Override
  public void doOperation(AuthorAccess authorAccess, ArgumentsMap args)
      throws IllegalArgumentException, AuthorOperationException {
    Object namespaceObj =  args.getArgumentValue(NAMESPACE_ARGUMENT);
    insertTable(null, false, authorAccess, (String) namespaceObj, null);
  }

  /**
   * Create the XML fragment for the table to be inserted.
   * 
   * @param tableInfo       The object containing information about the table.
   * @param namespace       The namespace of the table element.
   * @param innerTable      <code>true</code> if inner table.
   * @param fragments       The cells fragments. 
   * @param cellsFragments  <code>true</code> if the fragments are cells.
   * @param authorAccess    The Author access.
   * @param tableHelper     The table helper.
   *
   * @return The string buffer containing the XML fragment representing the table element.
   *
   * @throws AuthorOperationException 
   */
  private StringBuilder getXMLFragment(TableInfo tableInfo, String namespace, 
      boolean innerTable, AuthorDocumentFragment[] fragments, boolean cellsFragments, 
      AuthorAccess authorAccess, AuthorTableHelper tableHelper) throws AuthorOperationException {
    // Create the table XML fragment
    StringBuilder tableXMLFragment = new StringBuilder();
    if(innerTable) {
      //Add an entry table.
      tableXMLFragment.append("<entrytbl ");
      if (tableInfo.getTitle() != null 
          //The title is available only for DB 4.
          && namespace == null) {
        // Title was specified, insert a table with title
        tableXMLFragment.append("title=\"").append(tableInfo.getTitle()).append("\" ");
      }
      tableXMLFragment.append("cols=\"" + tableInfo.getColumnsNumber() + "\" ");
    } else {
      // Table element
      if (tableInfo.getTitle() != null) {
        // Title was specified, insert a table with title
        tableXMLFragment.append("<table ");
      } else {
        // Title was not specified, insert an informaltable        
        tableXMLFragment.append("<informaltable ");
      }
    }
    
    // Add the namespace.
    if (namespace != null) {
      tableXMLFragment.append("xmlns=\"").append(namespace).append("\" ");
    }
    
    if(!innerTable) {
      // Add the frame.
      if (tableInfo.getFrame() != null) {
        // EXM-23110 Add the frame attribute only if its value is different than "<unspecified>"
        tableXMLFragment.append("frame=\"" + tableInfo.getFrame() + "\"");
      }
    }
    
    //End the table tag
    tableXMLFragment.append(">");
    
    if (tableInfo.getTableModel() == TableInfo.TABLE_MODEL_CUSTOM
        || tableInfo.getTableModel() == TableInfo.TABLE_MODEL_CALS 
        || innerTable) {
      // CALS table model for DocBook v.4 or DocBook v.5
      addCALSXMLFragment(tableXMLFragment, tableInfo, namespace, innerTable, 
          fragments, cellsFragments, authorAccess, tableHelper);
    } else {
      // HTML table model for DocBook v.4 or DocBook v.5
      addHTMLXMLFragment(tableXMLFragment, tableInfo, namespace, 
          fragments, cellsFragments, authorAccess, tableHelper);
    }
    if(innerTable) {
      //Inner table
      tableXMLFragment.append("</entrytbl>");
    } else {
      if (tableInfo.getTitle() != null) {
        tableXMLFragment.append("</table>");
      } else {
        tableXMLFragment.append("</informaltable>");
      }
    }
    return tableXMLFragment;
  }
  
  /**
   * Adds to the XML fragment the elements (tgroup, colspec, tbody, thead, etc.) specific to
   * the CALS table model.
   *  
   * @param tableXMLFragment  The string buffer where to append CALS elements.
   * @param tableInfo         The table information.
   * @param namespace         The table element namespace.
   * @param fragments         The cells fragments.
   * @param cellsFragments    <code>true</code> if the fragments are cells.
   * @param authorAccess      The Author acccess.
   * @param tableHelper       The table helper
   *
   * @throws AuthorOperationException 
   */
  private void addCALSXMLFragment(StringBuilder tableXMLFragment, TableInfo tableInfo, 
      String namespace, boolean innerTable, AuthorDocumentFragment[] fragments, 
      boolean cellsFragments, AuthorAccess authorAccess, AuthorTableHelper tableHelper) throws AuthorOperationException {
    // Title element
    if(!innerTable) {
      if (tableInfo.getTitle() != null) {
        tableXMLFragment.append("<title>" + tableInfo.getTitle() + "</title>");
      }
    }
    
    if(!innerTable) {
      // Add 'tgroup', A wrapper for the main content of a table
      tableXMLFragment.append("<tgroup cols=\"" + tableInfo.getColumnsNumber() + "\">");
    }
    
    // Add column specifications
    addCALSColSpecs(tableXMLFragment, tableInfo);
    
    // Add table header
    addCALSTableHeader(tableXMLFragment, tableInfo, fragments, cellsFragments, 
        authorAccess, tableHelper, namespace);
    
    if(! innerTable) {
      // Add table footer
      addCALSTableFooter(tableXMLFragment, tableInfo, fragments, cellsFragments, 
          authorAccess, tableHelper, namespace);
    }
    
    // Add table body
    addCALSTableBody(tableXMLFragment, tableInfo, fragments, cellsFragments, 
        authorAccess, tableHelper, namespace);
    
    if(!innerTable) {
      tableXMLFragment.append("</tgroup>");
    }
  }

  /**
   * Add table body for CALS model.
   *
   * @param tableXMLFragment  The string buffer where to append CALS elements.
   * @param tableInfo         The table information.
   * @param fragments         The cell fragments.
   * @param cellFragments     <code>true</code> if the fragments are cells.
   * @param authorAccess      The Author access.
   * @param tableHelper       The table helper.
   * @param namespace         The table element namespace.
   * 
   * @throws AuthorOperationException 
   */
  private void addCALSTableBody(StringBuilder tableXMLFragment, TableInfo tableInfo, 
      AuthorDocumentFragment[] fragments, boolean cellsFragments, 
      AuthorAccess authorAccess, AuthorTableHelper tableHelper, String namespace) throws AuthorOperationException {
    tableXMLFragment.append("<tbody>");
    for (int i = 0; i < tableInfo.getRowsNumber(); i++) {
      tableXMLFragment.append("<row>");
      for (int j = 0; j < tableInfo.getColumnsNumber(); j++) {
        if (fragments != null) {
          int index = i;
          if (tableInfo.isGenerateFooter()) {
            index++;
          }
          if (tableInfo.isGenerateHeader()) {
            index++;
          }
          String cellXMLFragment = TableOperationsUtil.createCellXMLFragment(
              authorAccess, 
              fragments, 
              cellsFragments, 
              "entry", 
              index, 
              namespace, 
              tableHelper);
          tableXMLFragment.append(cellXMLFragment);
        } else {
          tableXMLFragment.append("<entry></entry>"); 
        }
      }
      tableXMLFragment.append("</row>"); 
    }
    tableXMLFragment.append("</tbody>");
  }

  /**
   * Add table footer for CALS model.
   *
   * @throws AuthorOperationException 
   */
  private void addCALSTableFooter(StringBuilder tableXMLFragment, TableInfo tableInfo,
      AuthorDocumentFragment[] fragments, boolean cellsFragments, AuthorAccess authorAccess, 
      AuthorTableHelper tableHelper, String namespace) throws AuthorOperationException {
    if (tableInfo.isGenerateFooter()) {
      tableXMLFragment.append("<tfoot><row>");
      for (int i = 1; i <= tableInfo.getColumnsNumber(); i++) {
        if (i == 1 && fragments != null && fragments.length > 0) {
          int index = tableInfo.isGenerateHeader() ? 1 : 0;
          String cellXMLFragment = TableOperationsUtil.createCellXMLFragment(
              authorAccess, 
              fragments, 
              cellsFragments, 
              "entry", 
              index, 
              namespace, 
              tableHelper);
          tableXMLFragment.append(cellXMLFragment);
        } else {
          tableXMLFragment.append("<entry></entry>");
        }
      }
      tableXMLFragment.append("</row></tfoot>");
    }
  }

  /**
   * Add table header for CALS model.
   *
   * @throws AuthorOperationException 
   */
  private void addCALSTableHeader(StringBuilder tableXMLFragment, TableInfo tableInfo, 
      AuthorDocumentFragment[] fragments, boolean cellsFragments, AuthorAccess authorAccess, 
      AuthorTableHelper tableHelper, String namespace) throws AuthorOperationException {
    if (tableInfo.isGenerateHeader()) {
      tableXMLFragment.append("<thead><row>");
      for (int i = 1; i <= tableInfo.getColumnsNumber(); i++) {
        if (i == 1 && fragments != null && fragments.length > 0) {
          String cellXMLFragment = TableOperationsUtil.createCellXMLFragment(
              authorAccess, 
              fragments, 
              cellsFragments, 
              "entry", 
              0, 
              namespace, 
              tableHelper);
          tableXMLFragment.append(cellXMLFragment);
        } else {
          tableXMLFragment.append("<entry></entry>");
        }
      }
      tableXMLFragment.append("</row></thead>");
    }
  }

  /**
   * Add column specification to the table XML fragment.
   */
  private void addCALSColSpecs(StringBuilder tableXMLFragment, TableInfo tableInfo) {
    ColumnWidthsType columnsWidthsType = tableInfo.getColumnsWidthsType();
    String colWidth = null;
    if (columnsWidthsType == ColumnWidthsType.PROPORTIONAL_COL_WIDTHS) {
      // Proportional widths
      colWidth = TableCustomizerConstants.REL_COL_WIDTH_DEFAULT_VALUE;
    } else if (columnsWidthsType == ColumnWidthsType.FIXED_COL_WIDTHS) {
      // Fixed widths
      colWidth = TableCustomizerConstants.FIXED_COL_WIDTH_DEFAULT_VALUE;
    }
    for (int i = 1; i <= tableInfo.getColumnsNumber(); i++) {
      tableXMLFragment.append("<colspec colname=\"c" + i + "\" colnum=\"" + i + "\"");
      if (colWidth != null) {
        tableXMLFragment.append(" colwidth=\"" + colWidth + "\"");
      }
      tableXMLFragment.append("/>");
    }
  }

  /**
   * The operation will display a dialog for choosing table attributes.
   * 
   * @see ro.sync.ecss.extensions.api.AuthorOperation#getArguments()
   */
  @Override
  public ArgumentDescriptor[] getArguments() {
    return ARGUMENTS;
  }

  /**
   * @see ro.sync.ecss.extensions.api.Extension#getDescription()
   */
  @Override
  public String getDescription() {
    return "Insert a docbook table";
  }

  /**
   * Adds to the given XML fragment the elements (caption, tbody, thead, etc.) 
   * specific to the HTML table model.
   *  
   * @param tableXMLFragment  The string buffer where to append HTML elements.
   * @param tableInfo         The table information.
   * @param namespace         The table element namespace.
   * @param fragments         The cells fragments.
   * @param cellsFragments    <code>true</code> if the fragments are cells.
   * @param authorAccess      The Author acccess.
   * @param tableHelper       The table helper.
   *
   * @throws AuthorOperationException 
   */
  private void addHTMLXMLFragment(StringBuilder tableXMLFragment, TableInfo tableInfo, 
      String namespace, AuthorDocumentFragment[] fragments, boolean cellsFragments, 
      AuthorAccess authorAccess, AuthorTableHelper tableHelper) throws AuthorOperationException {    
    if (tableInfo.getTitle() != null) {
      tableXMLFragment.append("<caption>" + tableInfo.getTitle() + "</caption>");
    }
    
    // Add table columns specification
    addHTMLTableCols(tableXMLFragment, tableInfo);
    
    // Add table header
    addHTMLTableHeader(tableXMLFragment, tableInfo, fragments, cellsFragments, 
        authorAccess, tableHelper, namespace);

    // Add table footer
    addHTMLTableFooter(tableXMLFragment, tableInfo, fragments, cellsFragments, 
        authorAccess, tableHelper, namespace);

    // Add table body
    addHTMLTableBody(tableXMLFragment, tableInfo, fragments, cellsFragments, 
        authorAccess, tableHelper, namespace);
  }
  
  /**
   * Add table columns specification.
   */
  private void addHTMLTableCols(StringBuilder tableXMLFragment, TableInfo tableInfo) {
    ColumnWidthsType columnsWidthsType = tableInfo.getColumnsWidthsType();
    if (columnsWidthsType != ColumnWidthsType.DYNAMIC_COL_WIDTHS) {
      String colWidth = null;
      if (columnsWidthsType == ColumnWidthsType.PROPORTIONAL_COL_WIDTHS) {
        float proportionalWidth = (float)100 / tableInfo.getColumnsNumber();
        // Proportional widths
        colWidth = (int)(Math.round(proportionalWidth * 100.0) / 100.0) + "%";
      } else {
        // Fixed widths
        colWidth = TableCustomizerConstants.FIXED_COL_WIDTH_DEFAULT_VALUE;
      }
      for (int i = 1; i <= tableInfo.getColumnsNumber(); i++) {
        tableXMLFragment.append("<col width=\"" + colWidth + "\"/>");
      }
    }
  }

  /**
   * Add table header for HTML model.
   *
   * @throws AuthorOperationException 
   */
  private static void addHTMLTableHeader(StringBuilder tableXMLFragment, 
      TableInfo tableInfo, AuthorDocumentFragment[] fragments, boolean cellsFragments, 
      AuthorAccess authorAccess, AuthorTableHelper tableHelper, String namespace) throws AuthorOperationException {
    if (tableInfo.isGenerateHeader()) {
      tableXMLFragment.append("<thead><tr>");
      for (int i = 1; i <= tableInfo.getColumnsNumber(); i++) {
        if (i == 1 && fragments != null && fragments.length > 0) {
          String cellXMLFragment = TableOperationsUtil.createCellXMLFragment(
              authorAccess, 
              fragments, 
              cellsFragments, 
              "th", 
              0, 
              namespace, 
              tableHelper);
          tableXMLFragment.append(cellXMLFragment);
        } else {
          tableXMLFragment.append("<th></th>");
        }
      }
      tableXMLFragment.append("</tr></thead>");
    }
  }
  
  /**
   * Add table body for HTML model.
   * 
   * @param tableXMLFragment  The string buffer where to append HTML elements.
   * @param tableInfo         The table information.
   * @param fragments         The cell fragments.
   * @param cellFragments     <code>true</code> if the fragments are cells.
   * @param authorAccess      The Author access.
   * @param tableHelper       The table helper.
   * @param namespace         The table element namespace.
   * 
   * @throws AuthorOperationException
   * 
   * @throws AuthorOperationException 
   */
  private void addHTMLTableBody(StringBuilder tableXMLFragment, TableInfo tableInfo,
      AuthorDocumentFragment[] fragments, boolean cellsFragments, AuthorAccess authorAccess, 
      AuthorTableHelper tableHelper, String namespace) throws AuthorOperationException {
    tableXMLFragment.append("<tbody>");
    for (int i = 0; i < tableInfo.getRowsNumber(); i++) {
      tableXMLFragment.append("<tr>");
      for (int j = 0; j < tableInfo.getColumnsNumber(); j++) {
        if (fragments != null) {
          int index = i;
          if (tableInfo.isGenerateFooter()) {
            index++;
          }
          if (tableInfo.isGenerateHeader()) {
            index++;
          }
          String cellXMLFragment = TableOperationsUtil.createCellXMLFragment(
              authorAccess, 
              fragments, 
              cellsFragments, 
              "td", 
              index, 
              namespace, 
              tableHelper);
          tableXMLFragment.append(cellXMLFragment);
        } else {
          tableXMLFragment.append("<td></td>");
        }
      }
      tableXMLFragment.append("</tr>"); 
    }
    tableXMLFragment.append("</tbody>");
  }

  /**
   * Add table footer for HTML model.
   *
   * @throws AuthorOperationException 
   */
  private static void addHTMLTableFooter(StringBuilder tableXMLFragment, 
      TableInfo tableInfo, AuthorDocumentFragment[] fragments, boolean cellsFragments, 
      AuthorAccess authorAccess, AuthorTableHelper tableHelper, String namespace) throws AuthorOperationException {
    if (tableInfo.isGenerateFooter()) {
      tableXMLFragment.append("<tfoot><tr>");
      for (int i = 1; i <= tableInfo.getColumnsNumber(); i++) {
        if (i == 1 && fragments != null && fragments.length > 0) {
          int index = tableInfo.isGenerateHeader() ? 1 : 0;
          String cellXMLFragment = TableOperationsUtil.createCellXMLFragment(
              authorAccess, 
              fragments, 
              cellsFragments, 
              "td", 
              index, 
              namespace, 
              tableHelper);
          tableXMLFragment.append(cellXMLFragment);
        } else {
          tableXMLFragment.append("<td></td>");
        }
      }
      tableXMLFragment.append("</tr></tfoot>");
    }
  }

  /**
   * @see ro.sync.ecss.extensions.commons.table.operations.InsertTableOperationBase#insertTable(ro.sync.ecss.extensions.api.node.AuthorDocumentFragment[], boolean, ro.sync.ecss.extensions.api.AuthorAccess, java.lang.String, ro.sync.ecss.extensions.commons.table.operations.AuthorTableHelper)
   */
  @Override
  public void insertTable(AuthorDocumentFragment[] fragments, boolean cellsFragments,
      AuthorAccess authorAccess, String namespace, AuthorTableHelper tableHelper)
      throws AuthorOperationException {
    int rowsCount = 0;
    int columnsCount = 0;
    if (fragments != null) {
      rowsCount = fragments.length;
      columnsCount = 1;
    }
    //EXM-10013 Take into account the inner table model
    int caretOffset = authorAccess.getEditorAccess().getCaretOffset();
    boolean addInnerTable = false;
    AuthorNode node;
    try {
      node = authorAccess.getDocumentController().getNodeAtOffset(caretOffset);
      //Maybe we have to insert an "entrytbl"
      if(node.getType() == AuthorNode.NODE_TYPE_ELEMENT) {
        AuthorElement elem = (AuthorElement) node;
        if("row".equals(elem.getLocalName())) {
          //Add an inner table
          addInnerTable = true;
        }
      }
    } catch (BadLocationException e) {
      //
    }

    // Show the 'Insert table' dialog
    TableInfo tableInfo = null;
    if (authorAccess.getWorkspaceAccess().isStandalone()) {
      if(addInnerTable) {
        //Inner table
        tableInfo = SADocbookInnerTableCustomizer.getInstance().customizeTable(
            authorAccess, rowsCount, columnsCount);
      } else {
        tableInfo = SADocbookTableCustomizer.getInstance().customizeTable(
            authorAccess, rowsCount, columnsCount);
      }
    } else {
      if(addInnerTable) {
        //Inner table
        tableInfo = ECDocbookInnerTableCustomizer.getInstance().customizeTable(
            authorAccess, rowsCount, columnsCount);
      } else {
        tableInfo = ECDocbookTableCustomizer.getInstance().customizeTable(
            authorAccess, rowsCount, columnsCount);
      }
    }
    if (tableInfo != null) {
      StringBuilder tableXMLFragment = getXMLFragment(
          tableInfo, namespace, addInnerTable, fragments, cellsFragments, authorAccess, tableHelper);

      // Insert the table schema aware(EXM-16077)
      authorAccess.getDocumentController().insertXMLFragmentSchemaAware(
          tableXMLFragment.toString(), caretOffset);
    } else {
      // User canceled the operation 
      throw new AuthorOperationStoppedByUserException("Cancelled by user");
    }
  }
}